/*
 *  tsl2561.c - Linux kernel modules for light to digital convertor
 *
 *  Copyright (C) 2008 Jonathan Cameron <jic23@cam.ac.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Some portions based upon the tsl2550 driver.
 *
 * This driver could probably be adapted easily to talk to the tsl2560 (smbus)
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/i2c.h>


#define TSL2561_CONTROL_REGISTER 0x00
#define TSL2561_TIMING_REGISTER 0x01
#define TSL2561_THRESHLOW_LOW_REGISTER 0x02
#define TSL2561_THRESHLOW_HIGH_REGISTER 0x03
#define TSL2561_THRESHHIGH_LOW_REGISTER 0x04
#define TSL2561_THRESHHIGH_HIGH_REGISTER 0x05
#define TSL2561_INTERRUPT_CONTROL_REGISTER 0x06

#define TSL2561_INTERRUPT_CONT_REG_INT_OFF 0x00
#define TSL2561_INTERRUPT_CONT_REG_INT_LEVEL 0x08
#define TSL2561_INTERRUPT_CONT_REG_INT_SMBUS 0x10
#define TSL2561_INTERRUPT_CONT_REG_INT_TEST 0x18

/*FIXME implement interrupt handling */

#define TSL2561_ID_REGISTER 0x0A

#define TSL2561_DATA_0_LOW 0x0C
#define TSL2561_DATA_0_HIGH 0x0D
#define TSL2561_DATA_1_LOW 0x0E
#define TSL2561_DATA_1_HIGH 0x0F

/* Control Register Values */
#define TSL2561_CONT_REG_PWR_ON 0x03
#define TSL2561_CONT_REG_PWR_OFF 0x00

struct TSL2561_command {
	unsigned int address : 4;
	unsigned int block : 1;
	unsigned int word : 1;
	unsigned int clear : 1;
	unsigned int cmd : 1;
};

static unsigned short normal_i2c[] = { 0x29, 0x39, 0x49, I2C_CLIENT_END };
I2C_CLIENT_INSMOD;


static const struct i2c_device_id tsl2561_id[] = {
	{ "tsl2561", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, tsl2561_id);

static ssize_t tsl2561_scan(struct device *dev, struct device_attribute *attr, char* buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	int ret, len = 0;
	struct TSL2561_command Command = {
		.cmd =  1,
		.clear = 0,
		.word = 1,
		.block = 0,
		.address = TSL2561_DATA_0_LOW,
	};

	/* Use word reading */
	ret = i2c_smbus_read_word_data(client, *(char*)(&Command));
	len += sprintf(buf + len, "%u ", ret);

	Command.address = TSL2561_DATA_1_LOW;
	ret = i2c_smbus_read_word_data(client, *(char*)(&Command));
	len += sprintf(buf + len, "%u ", ret);

	return len;
}

static DEVICE_ATTR(scan, S_IRUGO, tsl2561_scan, NULL);

static int tsl2561_initialize(struct i2c_client *client)
{
	int err;
	struct TSL2561_command Command = {
		.cmd =  1,
		.clear = 0,
		.word = 0,
		.block = 0,
		.address = TSL2561_CONTROL_REGISTER,
	};
	/*FIXME - use union rather than the cast here?*/
	err = i2c_smbus_write_byte_data(client, *(char*)(&Command),
					TSL2561_CONT_REG_PWR_ON);
	if(err < 0)
		return err;

	Command.address = TSL2561_INTERRUPT_CONTROL_REGISTER;
	err = i2c_smbus_write_byte_data(client, *(char*)(&Command),
					TSL2561_INTERRUPT_CONT_REG_INT_TEST);
	if (err < 0)
		return err;
 
	return 0;

}

static int tsl2561_powerdown(struct i2c_client *client)
{
	int err;
	struct TSL2561_command Command = {
		.cmd =  1,
		.clear = 0,
		.word = 0,
		.block = 0,
		.address = TSL2561_CONTROL_REGISTER,
	};

	err = i2c_smbus_write_byte_data(client, *(char*)(&Command),
					TSL2561_CONT_REG_PWR_OFF);
	if(err < 0)
		return err;
	return 0;
}
static int __devinit tsl2561_probe(struct i2c_client *client,
				   const struct i2c_device_id *id)
{
	int ret;

	/* add functionality checks */

	/* Intialize the chip */
	ret = tsl2561_initialize(client);
	if(ret < 0)
		return ret;

	device_create_file(&(client->dev), &dev_attr_scan);
	return 0;

}

static int __devexit tsl2561_remove(struct i2c_client *client)
{
	device_remove_file(&(client->dev), &dev_attr_scan);
	return tsl2561_powerdown(client);
}


static struct i2c_driver tsl2561_driver = {
	.driver = {
		.name = "tsl2561",
	},
	.probe = tsl2561_probe,
	.remove = __devexit_p(tsl2561_remove),
	.id_table  = tsl2561_id,
};

static __init int tsl2561_init(void)
{
	return i2c_add_driver(&tsl2561_driver);
}

static __exit void tsl2561_exit(void)
{
	i2c_del_driver(&tsl2561_driver);
}

MODULE_AUTHOR("Jonathan Cameron <jic23@cam.ac.uk>");
MODULE_DESCRIPTION("TSL2561 light sensor driver");
MODULE_LICENSE("GPL");

module_init(tsl2561_init);
module_exit(tsl2561_exit);
