/*
 * lis3l02dq.c	support STMicroelectronics LISD02DQ
 *		3d 2g Linear Accelerometers via SPI
 *
 * Copyright (c) 2007 Jonathan Cameron <jic23@cam.ac.uk>
 *
 * Loosely based upon tle62x0.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Settings:
 * Latch on interrupt generation enabled as it simplifies when to reenable
 * the interrupt.
 * 16 bit left justified mode used.
 *
 * Not implemented as yet
 */

#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/spi/spi.h>

#include <linux/sysfs.h>
#include <linux/list.h>
#include <linux/rtc.h>
#include <linux/industrialio/iio.h>
#include <linux/industrialio/sysfs.h>
#include <linux/industrialio/ring_sw.h>
#include <linux/industrialio/accel.h>

#include "lis3l02dq.h"


int lis3l02dq_spi_read_reg_8(struct device *dev,
				    u8 reg_address,
				    u8 *val)
{
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct lis3l02dq_state *st = indio_dev->dev_data;
	u8 txbuf = LIS3L02DQ_READ_REG(reg_address);

	return spi_write_then_read(st->us, &txbuf, 1, val, 1);
}

static int lis3l02dq_spi_write_reg_8(struct device *dev,
				     u8 reg_address,
				     u8 *val)
{
	struct spi_message msg;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct lis3l02dq_state *st = indio_dev->dev_data;

	u8 tx_buf[2] = {
		LIS3L02DQ_WRITE_REG(reg_address), *val
	};
	struct spi_transfer xfer = {
		.tx_buf = tx_buf,
		.bits_per_word = 8,
		.len = 2,
		.cs_change = 1,
	};
	printk("%d is the value inside the write function \n", *val);
	spi_message_init(&msg);
	spi_message_add_tail(&xfer, &msg);

	return spi_sync(st->us, &msg);
}


/* Relies on the MSB being one higher adress than the LSB */
static int lis3l02dq_spi_write_reg_s16(struct device *dev,
				       u8 lower_reg_address,
				       s16 value)
{
	struct spi_message msg;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct lis3l02dq_state *st = indio_dev->dev_data;
	u8 tx_buf[4] = {
		LIS3L02DQ_WRITE_REG(lower_reg_address), value & 0xFF,
		LIS3L02DQ_WRITE_REG(lower_reg_address+1), (value >> 8) & 0xFF,
	};
	struct spi_transfer xfers [] = { {
			.tx_buf = tx_buf,
			.bits_per_word = 8,
			.len = 2,
			.cs_change = 1,
		}, {
			.tx_buf = tx_buf + 2,
			.bits_per_word = 8,
			.len = 2,
			.cs_change = 1,
		},
	};

	spi_message_init(&msg);
	spi_message_add_tail(&xfers[0], &msg);
	spi_message_add_tail(&xfers[1], &msg);

	return spi_sync(st->us, &msg);
}

static int lis3l02dq_spi_read_reg_s16(struct device *dev,
				      u8 lower_reg_address,
				      s16 *val)
{
	struct spi_message msg;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct lis3l02dq_state *st = indio_dev->dev_data;
	int ret;
	u8 tx_buf[4] = {
		LIS3L02DQ_READ_REG(lower_reg_address),0,
		LIS3L02DQ_READ_REG(lower_reg_address+1),0,
	};
	u8 rx_buf[4];
	struct spi_transfer xfers [] = { {
			.tx_buf = tx_buf,
			.rx_buf = rx_buf,
			.bits_per_word = 8,
			.len = 2,
			.cs_change = 1,
		}, {
			.tx_buf = tx_buf + 2,
			.rx_buf = rx_buf + 2,
			.bits_per_word = 8,
			.len = 2,
			.cs_change = 1,

		},
	};

	spi_message_init(&msg);
	spi_message_add_tail(&xfers[0], &msg);
	spi_message_add_tail(&xfers[1], &msg);
	ret = spi_sync(st->us, &msg);
	if (ret) {
		dev_err(&st->us->dev, "problem when reading 16 bit register");
		goto error_ret;
	}
	*val = rx_buf[1] | (rx_buf[3] << 8);

error_ret:
	return ret;

}

static ssize_t lis3l02dq_read_signed(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	int ret;
	s8 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = lis3l02dq_spi_read_reg_8(dev, this_attr->address, (u8 *)&val);
	if (ret)
		goto err_ret;

	return sprintf(buf, "%d\n", val);

err_ret:
	return ret;
}

static ssize_t lis3l02dq_read_unsigned(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	int ret;
	u8 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = lis3l02dq_spi_read_reg_8(dev, this_attr->address, &val);
	if (ret) {
		dev_err(dev, "problem reading an unsigned 8 bit value");
		goto err_ret;
	}

	return sprintf(buf, "%d\n", val);
err_ret:
	return val;
}
/* Used for offsets etc so no need for lock. */
static ssize_t lis3l02dq_write_signed(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf,
				      size_t len)
{
	long valin;
	s8 val;
	int ret;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = strict_strtol(buf, 10, &valin);
	if (ret)
		goto err_ret;
	val = valin;
	ret = lis3l02dq_spi_write_reg_8(dev, this_attr->address, (u8 *)&val);
	if (ret)
		goto err_ret;

	return len;

err_ret:
	return ret;
}
/* Used for gains etc so no need for lock. */
static ssize_t lis3l02dq_write_unsigned(struct device *dev,
					struct device_attribute *attr,
					const char *buf,
					size_t len)
{
	int ret;
	ulong valin;
	u8 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = strict_strtoul(buf, 10, &valin);
	if (ret)
		goto err_ret;
	val = valin;
	ret = lis3l02dq_spi_write_reg_8(dev, this_attr->address, &val);
	if (ret)
		goto err_ret;

	return len;

err_ret:
	return ret;
}

static ssize_t lis3l02dq_read_16bit_signed(struct device *dev,
					   struct device_attribute *attr,
					   char *buf)
{
	int ret;
	s16 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = lis3l02dq_spi_read_reg_s16(dev, this_attr->address, &val);
	if (ret < 0) {
		dev_err(dev, "problem reading a signed 16 bit value from chip");
		return ret;
	}

	return sprintf(buf, "%d\n", val);
}


/* Prevents mode switching when running */
static ssize_t lis3l02dq_read_accel(struct device *dev,
				    struct device_attribute *attr,
				    char *buf)
{

	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	ssize_t returnval;

	mutex_lock(&indio_dev->mlock);
	//careful with this mode switch.
	if (indio_dev->currentmode == INDIO_RING_TRIGGERED)
		returnval = lis3l02dq_read_accel_from_ring(dev, attr, buf);
	else
		returnval =  lis3l02dq_read_16bit_signed(dev, attr, buf);
	mutex_unlock(&indio_dev->mlock);

	return returnval;
}

static ssize_t lis3l02dq_write_16bit_signed(struct device *dev,
					    struct device_attribute *attr,
					    const char *buf,
					    size_t len)
{
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	int ret;
	long val;

	ret = strict_strtol(buf, 10, &val);
	if (ret)
		return ret;
	mutex_lock(&indio_dev->mlock);
	ret = lis3l02dq_spi_write_reg_s16(dev, this_attr->address, val);
	mutex_unlock(&indio_dev->mlock);
	if (ret)
		return ret;

	return len;
}
//move over to the const one?
static ssize_t lis3l02dq_read_av_freq(struct device *dev,
				      struct device_attribute *attr,
				      char *buf)
{
	return sprintf(buf, "280 560 1120 4480\n");
}

static ssize_t lis3l02dq_read_frequency(struct device *dev,
					struct device_attribute *attr,
					char *buf)
{
	int ret, len;
	s8 t;
	ret = lis3l02dq_spi_read_reg_8(dev,
				       LIS3L02DQ_REG_CTRL_1_ADDRESS,
				       (u8 *)&t);
	if (ret)
		goto error_ret;
	t &= LIS3L02DQ_DEC_MASK;
	if (t == LIS3L02DQ_REG_CTRL_1_DF_128)
		len = sprintf(buf, "280");
	else if (t == LIS3L02DQ_REG_CTRL_1_DF_64)
		len = sprintf(buf, "560");
	else if (t == LIS3L02DQ_REG_CTRL_1_DF_32)
		len = sprintf(buf, "1120");
	else
		len = sprintf(buf, "4480");
	len += sprintf(buf+len, "\n");

	return len;

error_ret:
	return ret;
}

static ssize_t lis3l02dq_write_frequency(struct device *dev,
					 struct device_attribute *attr,
					 const char *buf,
					 size_t len)
{
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	long val;
	int ret;
	u8 t;

	ret = strict_strtol(buf, 10, &val);
	if (ret)
		return ret;

	mutex_lock(&indio_dev->mlock);
	ret = lis3l02dq_spi_read_reg_8(dev,
				       LIS3L02DQ_REG_CTRL_1_ADDRESS,
				       &t);
	if (ret)
		goto error_ret;
	/* Wipe the bits clean */
	t &= ~LIS3L02DQ_DEC_MASK;
	switch (val) {
	case 280:
		t |= LIS3L02DQ_REG_CTRL_1_DF_128;
		break;
	case 560:
		t |= LIS3L02DQ_REG_CTRL_1_DF_64;
		break;
	case 1120:
		t |= LIS3L02DQ_REG_CTRL_1_DF_32;
		break;
	case 4480:
		t |= LIS3L02DQ_REG_CTRL_1_DF_8;
		break;
	default:
		ret = -EINVAL;
		goto error_ret;
	};

	ret = lis3l02dq_spi_write_reg_8(dev,
					LIS3L02DQ_REG_CTRL_1_ADDRESS,
					&t);

	if (ret)
		goto error_ret;
	mutex_unlock(&indio_dev->mlock);
	return len;

error_ret:
	mutex_unlock(&indio_dev->mlock);
	return ret;

}

static int lis3l02dq_initial_setup(struct lis3l02dq_state *st)
{
	int ret;
	u8 val, valtest;

	st->us->mode = SPI_MODE_3;
	//hunch. Does the spidev change on this?
	spi_setup(st->us);

	val = LIS3L02DQ_DEFAULT_CTRL1;

	/* Write suitable defaults to ctrl1 */
	ret = lis3l02dq_spi_write_reg_8(&st->us->dev,
					LIS3L02DQ_REG_CTRL_1_ADDRESS,
					&val);
	if (ret) {
		dev_err(&st->us->dev, "problem with setup control register 1");
		goto err_ret;
	}
	/* Read back to check this has worked */
	ret = lis3l02dq_spi_read_reg_8(&st->us->dev,
				       LIS3L02DQ_REG_CTRL_1_ADDRESS,
				       &valtest);
	if (ret || (valtest != val)) {
		printk("device not playing ball\n");
		ret = -EINVAL;
		goto err_ret;
	}

	val = LIS3L02DQ_DEFAULT_CTRL2;
	ret = lis3l02dq_spi_write_reg_8(&st->us->dev,
					LIS3L02DQ_REG_CTRL_2_ADDRESS,
					&val);
	if (ret) {
		dev_err(&st->us->dev, "problem with setup control register 2");
		goto err_ret;
	}

	val = LIS3L02DQ_REG_WAKE_UP_CFG_LATCH_SRC;
	ret = lis3l02dq_spi_write_reg_8(&st->us->dev,
					LIS3L02DQ_REG_WAKE_UP_CFG_ADDRESS,
					&val);
	if (ret)
		dev_err(&st->us->dev, "problem with interrupt cfg register");
err_ret:
	return ret;
}



/* These are all a case of reading / writing directly to the chip */

static IIO_DEV_ATTR_ACCEL_X_OFFSET(S_IWUSR | S_IRUGO,
				   lis3l02dq_read_signed,
				   lis3l02dq_write_signed,
				   LIS3L02DQ_REG_OFFSET_X_ADDRESS);

static IIO_DEV_ATTR_ACCEL_Y_OFFSET(S_IWUSR | S_IRUGO,
				   lis3l02dq_read_signed,
				   lis3l02dq_write_signed,
				   LIS3L02DQ_REG_OFFSET_Y_ADDRESS);

static IIO_DEV_ATTR_ACCEL_Z_OFFSET(S_IWUSR | S_IRUGO,
				   lis3l02dq_read_signed,
				   lis3l02dq_write_signed,
				   LIS3L02DQ_REG_OFFSET_Z_ADDRESS);

static IIO_DEV_ATTR_ACCEL_X_GAIN(S_IWUSR | S_IRUGO,
				 lis3l02dq_read_unsigned,
				 lis3l02dq_write_unsigned,
				 LIS3L02DQ_REG_GAIN_X_ADDRESS);

static IIO_DEV_ATTR_ACCEL_Y_GAIN(S_IWUSR | S_IRUGO,
				 lis3l02dq_read_unsigned,
				 lis3l02dq_write_unsigned,
				 LIS3L02DQ_REG_GAIN_Y_ADDRESS);

static IIO_DEV_ATTR_ACCEL_Z_GAIN(S_IWUSR | S_IRUGO,
				 lis3l02dq_read_unsigned,
				 lis3l02dq_write_unsigned,
				 LIS3L02DQ_REG_GAIN_Z_ADDRESS);

static IIO_DEV_ATTR_ACCEL_THRESH(S_IWUSR | S_IRUGO,
				 lis3l02dq_read_16bit_signed,
				 lis3l02dq_write_16bit_signed,
				 LIS3L02DQ_REG_THS_L_ADDRESS);

/* Obviously the reading method for these will change depending on whether
   ring buffer capture is in use. Allow specification here of alternate
   function? Or take the approach used above of making this a driver issue
   rather than part of industrialio */



static IIO_DEV_ATTR_ACCEL_X(lis3l02dq_read_accel,
			    LIS3L02DQ_REG_OUT_X_L_ADDRESS);

static IIO_DEV_ATTR_ACCEL_Y(lis3l02dq_read_accel,
			    LIS3L02DQ_REG_OUT_Y_L_ADDRESS);

static IIO_DEV_ATTR_ACCEL_Z(lis3l02dq_read_accel,
			    LIS3L02DQ_REG_OUT_Z_L_ADDRESS);

static IIO_DEV_ATTR_SAMP_FREQ(S_IWUSR | S_IRUGO,
			      lis3l02dq_read_frequency,
			      lis3l02dq_write_frequency);

static IIO_DEV_ATTR_AVAIL_SAMP_FREQ(lis3l02dq_read_av_freq);



static ssize_t
lis3l02dq_read_interrupt_config(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	int len, ret, set;
	s8 val;
	struct iio_event_attr *this_attr = to_iio_event_attr(attr);

	ret = lis3l02dq_spi_read_reg_8(dev,
				       LIS3L02DQ_REG_WAKE_UP_CFG_ADDRESS,
				       (u8 *)&val);
	if (ret < 0)
		return ret;
	set = (val & this_attr->mask) ? 1 : 0;
	len = sprintf(buf, "%d\n", set);

	return len;
}

static ssize_t lis3l02dq_write_interrupt_config(struct device *dev,
						struct device_attribute *attr,
						const char *buf,
						size_t len)
{
	struct iio_event_attr *this_attr = to_iio_event_attr(attr);
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	int ret, currentlyset, addr, changed = 0;
	u8 valold, controlold;
	bool val;

	val = (buf[0] == '0') ? 0 : 1;

	mutex_lock(&indio_dev->mlock);
	/* read current value */
	ret = lis3l02dq_spi_read_reg_8(dev,
				       LIS3L02DQ_REG_WAKE_UP_CFG_ADDRESS,
				       &valold);
	if (ret)
		goto error_mutex_unlock;

	/* read current control */
	ret = lis3l02dq_spi_read_reg_8(dev,
				       LIS3L02DQ_REG_CTRL_2_ADDRESS,
				       &controlold);
	if (ret)
		goto error_mutex_unlock;
	currentlyset = valold & this_attr->mask ? 1 : 0;
	if (val == false && currentlyset) {
		valold &= ~this_attr->mask;
		changed = 1;
		ret = iio_remove_event_from_list(this_attr->listel);
		if (ret)
			goto error_mutex_unlock;
	} else if (val == true && !currentlyset) {
		changed = 1;
		valold |= this_attr->mask;
		/* move to a line spec rather than this? */
		ret = iio_add_event_to_list(this_attr->listel,
					    &indio_dev->interrupts[0]->ev_list);
		if (ret)
			goto error_mutex_unlock;
	}

	if (changed) {
		addr = LIS3L02DQ_REG_WAKE_UP_CFG_ADDRESS;
		ret = lis3l02dq_spi_write_reg_8(dev, addr, &valold);
		if (ret)
			goto error_mutex_unlock;
		/* This always enables the interrupt, even if we've remove the
		 * last thing using it. For this device we can use the reference
		 * count on the handler to tell us if anyone
		 * wants the interrupt */
		controlold = this_attr->listel->refcount ?
			(controlold | LIS3L02DQ_REG_CTRL_2_ENABLE_INTERRUPT):
			(controlold & ~LIS3L02DQ_REG_CTRL_2_ENABLE_INTERRUPT);
		addr = LIS3L02DQ_REG_CTRL_2_ADDRESS;
		ret = lis3l02dq_spi_write_reg_8(dev, addr, &controlold);
		if (ret)
			goto error_mutex_unlock;
	}
	mutex_unlock(&indio_dev->mlock);

	return len;

error_mutex_unlock:
	mutex_unlock(&indio_dev->mlock);
	return ret;
}


/* Caller responsible for locking as necessary. */
/* WHY LONG?*/
int __lis3l02dq_write_data_ready_config(struct device *dev,
					struct iio_event_handler_list *list,
					long state)
{
	int ret;
	u8 valold;
	bool currentlyset;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);

/* Get the current event mask register */
	ret = lis3l02dq_spi_read_reg_8(dev,
				       LIS3L02DQ_REG_CTRL_2_ADDRESS,
				       &valold);
	if (ret)
		goto error_ret;
/* Find out if data ready is already on */
	currentlyset
		= valold & LIS3L02DQ_REG_CTRL_2_ENABLE_DATA_READY_GENERATION;

/* Disable requested */
	if (state == 0 && currentlyset) {

		valold &= ~LIS3L02DQ_REG_CTRL_2_ENABLE_DATA_READY_GENERATION;
		/* The double right is to overcome what looks like a hardware bug */
		ret = lis3l02dq_spi_write_reg_8(dev, LIS3L02DQ_REG_CTRL_2_ADDRESS, &valold);
		if (ret)
			goto error_ret;
		ret = lis3l02dq_spi_write_reg_8(dev, LIS3L02DQ_REG_CTRL_2_ADDRESS, &valold);
		if (ret)
			goto error_ret;

		ret = iio_remove_event_from_list(list);
		if (ret)
			goto error_ret;

/* Enable requested */
	} else if (state && !currentlyset) {
		/* if not set, enable requested */
		valold |= LIS3L02DQ_REG_CTRL_2_ENABLE_DATA_READY_GENERATION;
		ret = iio_add_event_to_list(list, &indio_dev->interrupts[0]->ev_list);
		if (ret)
			goto error_ret;
		ret = lis3l02dq_spi_write_reg_8(dev, LIS3L02DQ_REG_CTRL_2_ADDRESS, &valold);
		if (ret)
			goto error_ret;
	}

	return 0;
error_ret:
	return ret;
}

static int lis3l02dq_thresh_handler_th(struct iio_dev *dev_info,
				       int index,
				       s64 timestamp,
				       int no_test)
{
	struct lis3l02dq_state *st = dev_info->dev_data;

	/* Stash the timestamp somewhere convenient for the bh */
	st->last_timestamp = timestamp;
	schedule_work(&st->work_cont_thresh.ws);

	return 0;
}


/* Unforunately it appears the interrupt won't clear unless you read from the
 * src register.
 */
static void lis3l02dq_thresh_handler_bh_no_check(struct work_struct *work_s)
{
	struct iio_work_cont *wc
		= container_of(work_s, struct iio_work_cont, ws_nocheck);
	struct lis3l02dq_state *st = wc->st;

	s8 t;

	lis3l02dq_spi_read_reg_8(st->indio_dev->dev,
				 LIS3L02DQ_REG_WAKE_UP_SRC_ADDRESS,
				 (u8 *)&t);

	if (t & LIS3L02DQ_REG_WAKE_UP_SRC_INTERRUPT_Z_HIGH)
		iio_push_event(st->indio_dev, 0,
			       IIO_EVENT_CODE_ACCEL_Z_HIGH,
			       st->last_timestamp);

	if (t & LIS3L02DQ_REG_WAKE_UP_SRC_INTERRUPT_Z_LOW)
		iio_push_event(st->indio_dev, 0,
			       IIO_EVENT_CODE_ACCEL_Z_LOW,
			       st->last_timestamp);

	if (t & LIS3L02DQ_REG_WAKE_UP_SRC_INTERRUPT_Y_HIGH)
		iio_push_event(st->indio_dev, 0,
			       IIO_EVENT_CODE_ACCEL_Y_HIGH,
			       st->last_timestamp);

	if (t & LIS3L02DQ_REG_WAKE_UP_SRC_INTERRUPT_Y_LOW)
		iio_push_event(st->indio_dev, 0,
			       IIO_EVENT_CODE_ACCEL_Y_LOW,
			       st->last_timestamp);

	if (t & LIS3L02DQ_REG_WAKE_UP_SRC_INTERRUPT_X_HIGH)
		iio_push_event(st->indio_dev, 0,
			       IIO_EVENT_CODE_ACCEL_X_HIGH,
			       st->last_timestamp);

	if (t & LIS3L02DQ_REG_WAKE_UP_SRC_INTERRUPT_X_LOW)
		iio_push_event(st->indio_dev, 0,
			       IIO_EVENT_CODE_ACCEL_X_LOW,
			       st->last_timestamp);

	enable_irq(st->us->irq);
	/* Ack (and allow for new interrupts? (not clear on data sheet ) )*/
	lis3l02dq_spi_read_reg_8(st->indio_dev->dev,
				 LIS3L02DQ_REG_WAKE_UP_ACK_ADDRESS,
				 (u8 *)&t);

	return;
}

/* A shared handler for a number of threshold types */
IIO_EVENT_SH(threshold, &lis3l02dq_thresh_handler_th);

IIO_EVENT_ATTR_ACCEL_X_HIGH_SH(iio_event_threshold,
			       lis3l02dq_read_interrupt_config,
			       lis3l02dq_write_interrupt_config,
			       LIS3L02DQ_REG_WAKE_UP_CFG_INTERRUPT_X_HIGH);

IIO_EVENT_ATTR_ACCEL_Y_HIGH_SH(iio_event_threshold,
			       lis3l02dq_read_interrupt_config,
			       lis3l02dq_write_interrupt_config,
			       LIS3L02DQ_REG_WAKE_UP_CFG_INTERRUPT_Y_HIGH);

IIO_EVENT_ATTR_ACCEL_Z_HIGH_SH(iio_event_threshold,
			       lis3l02dq_read_interrupt_config,
			       lis3l02dq_write_interrupt_config,
			       LIS3L02DQ_REG_WAKE_UP_CFG_INTERRUPT_Z_HIGH);

IIO_EVENT_ATTR_ACCEL_X_LOW_SH(iio_event_threshold,
			      lis3l02dq_read_interrupt_config,
			      lis3l02dq_write_interrupt_config,
			      LIS3L02DQ_REG_WAKE_UP_CFG_INTERRUPT_X_LOW);

IIO_EVENT_ATTR_ACCEL_Y_LOW_SH(iio_event_threshold,
			      lis3l02dq_read_interrupt_config,
			      lis3l02dq_write_interrupt_config,
			      LIS3L02DQ_REG_WAKE_UP_CFG_INTERRUPT_Y_LOW);

IIO_EVENT_ATTR_ACCEL_Z_LOW_SH(iio_event_threshold,
			      lis3l02dq_read_interrupt_config,
			      lis3l02dq_write_interrupt_config,
			      LIS3L02DQ_REG_WAKE_UP_CFG_INTERRUPT_Z_LOW);

static struct attribute *lis3l02dq_event_attributes[] = {
	&iio_event_attr_accel_x_high.dev_attr.attr,
	&iio_event_attr_accel_y_high.dev_attr.attr,
	&iio_event_attr_accel_z_high.dev_attr.attr,
	&iio_event_attr_accel_x_low.dev_attr.attr,
	&iio_event_attr_accel_y_low.dev_attr.attr,
	&iio_event_attr_accel_z_low.dev_attr.attr,

	NULL
};

static struct attribute_group lis3l02dq_event_attribute_group = {
	.attrs = lis3l02dq_event_attributes,
};

static struct attribute *lis3l02dq_attributes[] = {
	&iio_dev_attr_accel_x_offset.dev_attr.attr,
	&iio_dev_attr_accel_y_offset.dev_attr.attr,
	&iio_dev_attr_accel_z_offset.dev_attr.attr,
	&iio_dev_attr_accel_x_gain.dev_attr.attr,
	&iio_dev_attr_accel_y_gain.dev_attr.attr,
	&iio_dev_attr_accel_z_gain.dev_attr.attr,
	&iio_dev_attr_thresh.dev_attr.attr,
	&iio_dev_attr_accel_x.dev_attr.attr,
	&iio_dev_attr_accel_y.dev_attr.attr,
	&iio_dev_attr_accel_z.dev_attr.attr,
	&iio_dev_attr_sampling_frequency.dev_attr.attr,
	&iio_dev_attr_available_sampling_frequency.dev_attr.attr,
	NULL
};

static const struct attribute_group lis3l02dq_attribute_group = {
	.attrs = lis3l02dq_attributes,
};


static int lis3l02dq_scan_el_set_state(struct iio_scan_el *scan_el,
				       struct iio_dev *indio_dev,
				       bool state)
{
	u8 t;
	int ret;
	bool changed = false;

	ret = lis3l02dq_spi_read_reg_8(indio_dev->dev,
				       LIS3L02DQ_REG_CTRL_1_ADDRESS,
				       &t);
	if (ret)
		goto error_ret;

	switch(scan_el->label) {
	case LIS3L02DQ_REG_OUT_X_L_ADDRESS:
		if (!!(LIS3L02DQ_REG_CTRL_1_AXES_X_ENABLE & t) != state) {
			changed = true;
			if (state)
				t |= LIS3L02DQ_REG_CTRL_1_AXES_X_ENABLE;
			else
				t &= ~LIS3L02DQ_REG_CTRL_1_AXES_X_ENABLE;
		}
		break;
	case LIS3L02DQ_REG_OUT_Y_L_ADDRESS:
		if (!!(LIS3L02DQ_REG_CTRL_1_AXES_Y_ENABLE & t) != state) {
			changed = true;
			if (state)
				t |= LIS3L02DQ_REG_CTRL_1_AXES_Y_ENABLE;
			else
				t &= ~LIS3L02DQ_REG_CTRL_1_AXES_Y_ENABLE;
		}
		break;
	case LIS3L02DQ_REG_OUT_Z_L_ADDRESS:
		if (!!(LIS3L02DQ_REG_CTRL_1_AXES_Z_ENABLE & t) != state) {
			changed = true;
			if (state)
				t |= LIS3L02DQ_REG_CTRL_1_AXES_Z_ENABLE;
			else
				t &= ~LIS3L02DQ_REG_CTRL_1_AXES_Z_ENABLE;
		}
		break;
	default:
		goto error_ret;
	}
	if (changed)
		ret = lis3l02dq_spi_write_reg_8(indio_dev->dev,
						LIS3L02DQ_REG_CTRL_1_ADDRESS,
						&t);

error_ret:
	return ret;

}

static IIO_SCAN_EL_C(accel_x, 0, IIO_SIGNED(16), LIS3L02DQ_REG_OUT_X_L_ADDRESS, &lis3l02dq_scan_el_set_state);
static IIO_SCAN_EL_C(accel_y, 1, IIO_SIGNED(16), LIS3L02DQ_REG_OUT_Y_L_ADDRESS, &lis3l02dq_scan_el_set_state);
static IIO_SCAN_EL_C(accel_z, 2, IIO_SIGNED(16), LIS3L02DQ_REG_OUT_Z_L_ADDRESS, &lis3l02dq_scan_el_set_state);

static struct attribute *lis3l02dq_scan_el_attrs[] = {
	&iio_scan_el_accel_x.dev_attr.attr,
	&iio_scan_el_accel_y.dev_attr.attr,
	&iio_scan_el_accel_z.dev_attr.attr,
	NULL,
};

static struct attribute_group lis3l02dq_scan_el_group = {
	.attrs = lis3l02dq_scan_el_attrs,
	//MOVE NAME setting into iio_core.
	.name = "scan_elements",
};

static int __devinit lis3l02dq_probe(struct spi_device *spi)
{
	struct lis3l02dq_state *st;
	int ret;

	st = kzalloc(sizeof(struct lis3l02dq_state), GFP_KERNEL);
	if (st == NULL) {
		ret =  -ENOMEM;
		goto error_ret;
	}
	printk("Address of state %p\n", st);
	st->us = spi;


	/* setup the industrialio driver allocated elements */
	st->indio_dev = kzalloc(sizeof(struct iio_dev), GFP_KERNEL);
	if (st->indio_dev == NULL) {
		ret = -ENOMEM;
		goto error_free_st;
	}

	st->indio_dev->dev = &spi->dev;

	st->indio_dev->num_interrupt_lines = 1;
	st->indio_dev->event_attrs = &lis3l02dq_event_attribute_group;
	st->indio_dev->attrs = &lis3l02dq_attribute_group;
	st->indio_dev->dev_data = (void *)(st);
	st->indio_dev->driver_module = THIS_MODULE;
	st->indio_dev->modes = INDIO_DIRECT_MODE;
	st->indio_dev->scan_mask = iio_scan_el_accel_x.mask
		| iio_scan_el_accel_y.mask
		| iio_scan_el_accel_z.mask;
	st->indio_dev->scan_count = 3;
	st->indio_dev->scan_el_attrs = &lis3l02dq_scan_el_group;
	lis3l02dq_register_ring_funcs(st->indio_dev);

	ret = iio_device_register(st->indio_dev);
	if (ret)
		goto error_free_dev;

	/*FIXME no error checking */
	ret = lis3l02dq_probe_trigger(st->indio_dev);
	if (ret)
		goto error_unregister_dev;
	/* overly complex */
	if (spi->irq && gpio_is_valid(irq_to_gpio(spi->irq)) > 0) {
		INIT_WORK(&st->work_data_rdy_ring,
			  lis3l02dq_data_rdy_bh_to_ring);

		/* This is a little unusual, in that the device seems
		   to need a full read of the interrupt source reg before
		   the interrupt will reset.
		   Hence the two handlers are the same */
		iio_init_work_cont(&st->work_cont_thresh,
				   lis3l02dq_thresh_handler_bh_no_check,
				   lis3l02dq_thresh_handler_bh_no_check,
				   LIS3L02DQ_REG_WAKE_UP_SRC_ADDRESS,
				   0,
				   st);
		st->inter = 0;
		ret = iio_register_interrupt_line(spi->irq,
						  st->indio_dev,
						  0,
						  IRQF_TRIGGER_RISING,
						  "lis3l02dq");

		if (ret)
			goto error_remove_trigger;
	}

	/* Get the device into a sane initial state */
	ret = lis3l02dq_initial_setup(st);
	if (ret)
		goto error_unregister_line;
	return 0;
error_remove_trigger:
	lis3l02dq_remove_trigger(st->indio_dev);
error_unregister_line:
	if (st->indio_dev->modes & INDIO_RING_TRIGGERED)
		iio_unregister_interrupt_line(st->indio_dev, 0);
error_unregister_dev:
	iio_device_unregister(st->indio_dev);
error_free_dev:
	kfree(st->indio_dev);
error_free_st:
	kfree(st);
error_ret:
	return ret;
}

/* Power down the device */
static int lis3l02dq_stop_device(struct iio_dev *indio_dev)
{
	int ret;
	struct lis3l02dq_state *st = indio_dev->dev_data;
	u8 val = 0;
	mutex_lock(&indio_dev->mlock);
	ret = lis3l02dq_spi_write_reg_8(&st->us->dev,
					LIS3L02DQ_REG_CTRL_1_ADDRESS,
					&val);
	if (ret) {
		dev_err(&st->us->dev, "problem with turning device off: ctrl1");
		goto err_ret;
	}

	ret = lis3l02dq_spi_write_reg_8(&st->us->dev,
					LIS3L02DQ_REG_CTRL_2_ADDRESS,
					&val);
	if (ret)
		dev_err(&st->us->dev, "problem with turning device off: ctrl2");
err_ret:
	mutex_unlock(&indio_dev->mlock);
	return ret;
}

/* fixme, confirm ordering in this function */
static int lis3l02dq_remove(struct spi_device *spi)
{
	int ret;
	struct iio_dev *indio_dev = spi_get_drvdata(spi);
	struct lis3l02dq_state *st = indio_dev->dev_data;

	ret = lis3l02dq_stop_device(indio_dev);
	/* Make sure all bottom halfs of interrupts are done */
	flush_scheduled_work();
	if (ret)
		goto err_ret;
	/* Fixme slightly misleading test condition - even if valid */
	if (indio_dev->modes & INDIO_RING_TRIGGERED)
		iio_unregister_interrupt_line(indio_dev, 0);
	lis3l02dq_remove_trigger(indio_dev);
	iio_device_unregister(indio_dev);
	kfree(indio_dev);
	kfree(st);
	return 0;

err_ret:
	return ret;
}

static struct spi_driver lis3l02dq_driver = {
	.driver = {
		.name = "lis3l02dq",
		.owner = THIS_MODULE,
	},
	.probe = lis3l02dq_probe,
	.remove = __devexit_p(lis3l02dq_remove),
};

static __init int lis3l02dq_init(void)
{
	return spi_register_driver(&lis3l02dq_driver);
}
module_init(lis3l02dq_init);

static __exit void lis3l02dq_exit(void)
{
	spi_unregister_driver(&lis3l02dq_driver);
}
module_exit(lis3l02dq_exit);

MODULE_AUTHOR("Jonathan Cameron <jic23@cam.ac.uk>");
MODULE_DESCRIPTION("ST LIS3L02DQ Accelerometer SPI driver");
MODULE_LICENSE("GPL v2");
