
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/spi/spi.h>

#include <linux/sysfs.h>
#include <linux/list.h>
#include <linux/rtc.h>
#include <linux/industrialio/iio.h>
#include <linux/industrialio/sysfs.h>
#include <linux/industrialio/ring_sw.h>
#include <linux/industrialio/accel.h>
#include <linux/industrialio/trigger.h>
#include "lis3l02dq.h"

static u8 count_to_right(u16 full, u16 mask)
{
	u8 count = 0;
	while(mask){
		 mask >>= 1;
		if (mask & full)
			count++;
	}
	return count;
}

static inline u16 combine_8_to_16(u8 lower, u8 upper)
{
	u16 _lower = lower;
	u16 _upper = upper;
	return _lower | (_upper << 8);
}

static void lis3l02dq_poll_func_th(void *private_data)
{
	struct iio_dev *indio_dev = private_data;
	struct lis3l02dq_state *st = indio_dev->dev_data;
	schedule_work(&st->work_data_rdy_ring);
//	printk("th\n");
	st->inter = 1;
}
//running when no one is listening. Should not happen!
static int lis3l02dq_data_rdy_trig_poll(struct iio_dev *dev_info,
				       int index,
				       s64 timestamp,
				       int no_test)
{
	struct lis3l02dq_state *st = dev_info->dev_data;
	struct iio_trigger *trig = st->trig;

	/*notify the trigger of the event timestamp.*/
	trig->timestamp = timestamp;
	iio_trigger_poll(trig);

	return IRQ_HANDLED;
}

/* This is an event as it is a response to a physical interrupt */
IIO_EVENT_SH(data_rdy_trig, &lis3l02dq_data_rdy_trig_poll);

ssize_t lis3l02dq_read_accel_from_ring(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	struct iio_scan_el *el = NULL;
	int ret, len;

	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	s16* data;

	int i = 0;

	while (dev_info->scan_el_attrs->attrs[i]) {
		el = to_iio_scan_el((struct device_attribute *)
				    (dev_info->scan_el_attrs->attrs[i]));
		//label is in fact the address.
		if (el->label == this_attr->address)
			break;
		i++;
	}
	if (dev_info->scan_el_attrs->attrs[i] == NULL)
		return -EINVAL;
	/* If this element is in the scan mask */
	if (el->mask & dev_info->scan_mask) {
		data = kmalloc(sizeof(s16)*dev_info->scan_count + sizeof(s64),
			       GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		ret = dev_info->ring_access.read_last(dev_info->ring,
						      (u8 *)data);
		if (ret)
			return ret;
	} else
		return -EINVAL;

	len = sprintf(buf, "ring %d\n", data[count_to_right(dev_info->scan_mask,
							    el->mask)]);
	return len;

}

static const u8 read_all_tx_array[] =
{
	LIS3L02DQ_READ_REG(LIS3L02DQ_REG_OUT_X_L_ADDRESS),0,
	LIS3L02DQ_READ_REG(LIS3L02DQ_REG_OUT_X_H_ADDRESS),0,
	LIS3L02DQ_READ_REG(LIS3L02DQ_REG_OUT_Y_L_ADDRESS),0,
	LIS3L02DQ_READ_REG(LIS3L02DQ_REG_OUT_Y_H_ADDRESS),0,
	LIS3L02DQ_READ_REG(LIS3L02DQ_REG_OUT_Z_L_ADDRESS),0,
	LIS3L02DQ_READ_REG(LIS3L02DQ_REG_OUT_Z_H_ADDRESS),0,
};

int lis3l02dq_read_all(struct lis3l02dq_state *st, u8 *rx_array)
{
	struct spi_transfer *xfers;
	struct spi_message msg;
	int ret, i, j = 0;

	xfers = kzalloc((st->indio_dev->scan_count) * 2
			* sizeof(struct spi_transfer), GFP_KERNEL);

	if (!xfers)
		return -ENOMEM;

	for (i = 0; i < ARRAY_SIZE(read_all_tx_array)/4; i++) {
		if (st->indio_dev->scan_mask & ( 1 << i)) {
			/* lower byte */
			xfers[j].tx_buf = &read_all_tx_array[i*4];
			if (rx_array)
				xfers[j].rx_buf = &rx_array[i*4];
			xfers[j].bits_per_word = 8;
			xfers[j].len = 2;
			xfers[j].cs_change = 1;
			j++;
			/* upper byte */
			xfers[j].tx_buf = &read_all_tx_array[i*4 + 2];
			if (rx_array)
				xfers[j].rx_buf = &rx_array[i*4 + 2];
			xfers[j].bits_per_word = 8;
			xfers[j].len = 2;
			xfers[j].cs_change = 1;
			j++;
		}
	}
	/* After these are transmitted, the rx_buff should have
	 * values in alternate bytes
	 */
	spi_message_init(&msg);
	for (j = 0; j < st->indio_dev->scan_count * 2; j++)
		spi_message_add_tail(&xfers[j], &msg);
	ret = spi_sync(st->us, &msg);
	kfree(xfers);

	return ret;
}

/* Whilst this makes a lot of calls to iio_sw_ring functions - it is to device
 * specific to be rolled into the core.
 */
void lis3l02dq_data_rdy_bh_to_ring(struct work_struct *work_s)
{
	struct lis3l02dq_state *st
		= container_of(work_s, struct lis3l02dq_state,
			       work_data_rdy_ring);
	unsigned char *rx_array;
	int i;
	u16 *data;
//	printk("bh\n");
	data = kmalloc(sizeof(s16)*3+sizeof(s64), GFP_KERNEL);
	if (data == NULL) {
		dev_err(&st->us->dev, "memory alloc failed in ring bh");
		return;
	}
	rx_array = kmalloc(4 * st->indio_dev->scan_count, GFP_KERNEL);
	if (rx_array == NULL) {
		dev_err(&st->us->dev, "memory alloc failed in ring bh");
		return;
	}
	st->inter = 0;

	if (lis3l02dq_read_all(st, rx_array) >= 0) {
		for (i = 0; i < st->indio_dev->scan_count; i++)
			data[i]	= combine_8_to_16 (rx_array[i*4+1],
						   rx_array[i*4+3]);

		*((s64 *)(&data[i + 1])) = st->last_timestamp;

		st->indio_dev->ring_access.store_to(st->indio_dev->ring,
						    (u8 *)data,
						    st->last_timestamp);
	}
	/* so the data should now be in the rx array */
try_again:
	while (gpio_get_value(irq_to_gpio(st->us->irq)))
		if (lis3l02dq_read_all(st, rx_array) >= 0) {
			for (i = 0; i < st->indio_dev->scan_count; i++)
				data[i] = combine_8_to_16 (rx_array[i*4+1],
							   rx_array[i*4+3]);
			*((s64 *)(&data[i+1])) = 0;
			st->indio_dev->ring_access.store_to(st->indio_dev->ring,
							    (u8 *)data,
							    0);
		}
	/* push data into ring buffer before trying renable */
	enable_irq(st->us->irq);
	if (gpio_get_value(irq_to_gpio(st->us->irq)))
		if (st->inter == 0) {
			disable_irq_nosync(st->us->irq);
			goto try_again;
		}
	kfree(rx_array);

	return;
}

static int lis3l02dq_data_rdy_ring_preenable(struct iio_dev *indio_dev)
{
	int size;

	if(indio_dev->ring_access.set_bpd) {
		size = sizeof(s64) + indio_dev->scan_count*sizeof(s16);
		indio_dev->ring_access.set_bpd(indio_dev->ring, size);
	}

	return 0;
}

static int lis3l02dq_data_rdy_ring_postenable(struct iio_dev *indio_dev)
{
	if (indio_dev->trig)
		return iio_trigger_attach_poll_func(indio_dev->trig,
						    indio_dev->pollfunc);
	return 0;
}

static int lis3l02dq_data_rdy_ring_predisable(struct iio_dev *indio_dev)
{
	if (indio_dev->trig)
		return iio_trigger_dettach_poll_func(indio_dev->trig,
						     indio_dev->pollfunc);
		return 0;
}
/* disable the interrupt and acknowledge any that are already set */
static int lis3l02dq_data_rdy_trigger_set_state(struct iio_trigger *trig,
						bool state)
{
	struct lis3l02dq_state *st = trig->private_data;
	int ret = 0;
	u8 t;

	__lis3l02dq_write_data_ready_config(&st->us->dev,
					   &iio_event_data_rdy_trig,
					   state);

	//attempt to clear any cheeky interrupts
	//THE POINT OF THIS IS TO CLEAR ANY OUTSTANDING events.
	if (state == false) {
		ret = lis3l02dq_read_all(st, NULL);
	}


	lis3l02dq_spi_read_reg_8(st->indio_dev->dev,
				 LIS3L02DQ_REG_WAKE_UP_SRC_ADDRESS,
				 &t);
	printk("wakup stuff %d\n", t);
printk("if 1 then inte3rrupt high%d\n",gpio_get_value(irq_to_gpio(st->us->irq)));
	return ret;
}
DEVICE_ATTR(name, S_IRUGO, iio_trigger_read_name, NULL);

static struct attribute *lis3l02dq_trigger_attrs[] = {
	&dev_attr_name.attr,
	NULL,
};

static const struct attribute_group lis3l02dq_trigger_attr_group = {
	.attrs = lis3l02dq_trigger_attrs,
};

int lis3l02dq_probe_trigger(struct iio_dev *indio_dev)
{
	int ret;
	struct lis3l02dq_state *state = indio_dev->dev_data;

	state->trig = kzalloc(sizeof(*state->trig), GFP_KERNEL);
	if (!state->trig) {
		ret = -ENOMEM;
		goto error_ret;
	}
	iio_trigger_init(state->trig);
	state->trig->name = kmalloc(IIO_TRIGGER_NAME_LENGTH, GFP_KERNEL);
	if (!state->trig->name) {
		ret = -ENOMEM;
		goto error_free_trig;
	}
	snprintf((char *)state->trig->name,
		 IIO_TRIGGER_NAME_LENGTH,
		 "iio_dev%d_data_rdy", indio_dev->id);
	state->trig->owner = THIS_MODULE;
	state->trig->private_data = state;
	state->trig->set_trigger_state = &lis3l02dq_data_rdy_trigger_set_state;
	state->trig->control_attrs = &lis3l02dq_trigger_attr_group;
	ret = iio_trigger_register(state->trig, indio_dev->dev, 0);
	if (ret)
		goto error_free_trig_name;
	/* Set the default trigger for the device to be it's own data ready */
	indio_dev->trig = state->trig;

	return 0;

error_free_trig_name:
	kfree(state->trig->name);
error_free_trig:
	kfree(state->trig);
error_ret:
	return ret;
}

void lis3l02dq_remove_trigger(struct iio_dev *indio_dev)
{
	struct lis3l02dq_state *state = indio_dev->dev_data;

	iio_trigger_unregister(state->trig);
	kfree(state->trig->name);
	kfree(state->trig);
}
//FIXME: error codes now possible.
void lis3l02dq_register_ring_funcs(struct iio_dev *indio_dev)
{
	int ret;

	/* Effectively select the ring buffer implementation */
	iio_ring_sw_register_funcs(&indio_dev->ring_access);
	indio_dev->ring_preenable = &lis3l02dq_data_rdy_ring_preenable;
	indio_dev->ring_postenable = &lis3l02dq_data_rdy_ring_postenable;
	indio_dev->ring_predisable = &lis3l02dq_data_rdy_ring_predisable;

	/* must init this somewhere */
	INIT_LIST_HEAD(&iio_event_data_rdy_trig.list);

	indio_dev->pollfunc = kzalloc(sizeof(*indio_dev->pollfunc), GFP_KERNEL);
	if (indio_dev->pollfunc == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}
	indio_dev->pollfunc->poll_func_main = &lis3l02dq_poll_func_th;
	indio_dev->pollfunc->private_data = indio_dev;
	indio_dev->modes |= INDIO_RING_TRIGGERED;
error_ret:
	return;

}

int lis3l02dq_set_ring_length(struct iio_dev *indio_dev, int length)
{
	/* Set sensible defaults for the ring buffer */
	if (indio_dev->ring_access.set_length)
		return indio_dev->ring_access.set_length(indio_dev->ring, 500);
	return 0;
}


