
#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/spi/spi.h>
#include <linux/sysfs.h>
#include <linux/rtc.h>
#include <linux/industrialio/iio.h>
#include <linux/industrialio/sysfs.h>
#include <linux/industrialio/accel.h>
/* Shared kernel / userspace conversion functions */
#include <linux/industrialio/sca3000.h>
#include <linux/industrialio/ring_hw.h>
#include "sca3000.h"

/* FIXME - ioctl to not bother checking whether there is enough in the ring?
 * Afterall, if we are responding to an interrupt we have a minimum content
 * guaranteed*/
/* FIXME: allow for ring buffering of single / pairs of directions */

/* This inherently doesn't care which samples it's reading, merely the number */
int sca3000_rip_hw_rb(void *r, size_t count, u8 **data, int *dead_offset)
{
	struct iio_hw_ring_buffer *hw_ring = r;
	struct iio_dev *indio_dev = hw_ring->private;
	struct sca3000_state *st = indio_dev->dev_data;
	u8 *rx;
	int ret, num_available, num_read;
	int bytes_per_sample = 1;

	if (st->bps == 11)
		bytes_per_sample = 2;
	/* Check how much data is available!*/
	ret = sca3000_read_data(st->indio_dev->dev,
				SCA3000_REG_ADDR_BUF_COUNT,
				&rx, 1);
	if (ret)
		goto error_ret;
	else
		num_available = rx[1];
	/* num_available is the total number of samples available
	 * i.e. number of time points * number of channels.
	 */
	kfree(rx);
	if (count > num_available*bytes_per_sample)
		num_read = num_available*bytes_per_sample;
	else
		num_read = count - (count % (bytes_per_sample));

	*dead_offset = 1;
	ret = sca3000_read_data(st->indio_dev->dev,
				SCA3000_REG_ADDR_RING_OUT,
				data, num_read);
	if (ret)
		goto error_ret;
	return num_read;
error_ret:
	return ret;
}

/* Ring access functions */
static int sca3000_create_ring_buffer(void **r)
{
	*r = kzalloc(sizeof(struct iio_hw_ring_buffer),
		     GFP_KERNEL);
	if (*r == 0)
		return -ENOMEM;
	return 0;
}
static int sca3000_init_ring_buffer(void *r, void *iio_dev)
{
	struct iio_hw_ring_buffer *ring = r;

	ring->private = iio_dev;

	return 0;
}

static void sca3000_free_ring_buffer(void *r)
{
	kfree(r);
}

int sca3000_register_ring_access_and_init(struct iio_dev *indio_dev)
{
	struct iio_ring_access_funcs *ra = &indio_dev->ring_access;
	indio_dev->modes |= INDIO_RING_HARDWARE_BUFFER;
	ra->_create = &sca3000_create_ring_buffer;
	ra->_free = &sca3000_free_ring_buffer;
	ra->_init = &sca3000_init_ring_buffer;

	ra->rip_lots = &sca3000_rip_hw_rb;
	return 0;
}

static inline
int __sca3000_hw_ring_state_set(struct iio_dev *indio_dev, bool state)
{
	struct sca3000_state *st = indio_dev->dev_data;
	int ret;
	u8 *rx;

	mutex_lock(&st->lock);
	ret = sca3000_read_data(indio_dev->dev, SCA3000_REG_ADDR_MODE, &rx, 1);
	if (ret)
		goto error_ret;
	if (state)
		ret = sca3000_write_reg(indio_dev->dev,
					SCA3000_REG_ADDR_MODE,
					(rx[1] | SCA3000_RING_BUF_ENABLE));
	else
		ret = sca3000_write_reg(indio_dev->dev,
					SCA3000_REG_ADDR_MODE,
					(rx[1] & ~SCA3000_RING_BUF_ENABLE));
	kfree(rx);
error_ret:
	mutex_unlock(&st->lock);

	return ret;
}

static int sca3000_hw_ring_preenable(struct iio_dev *indio_dev)
{
	return __sca3000_hw_ring_state_set(indio_dev, 1);
}

static int sca3000_hw_ring_postdisable(struct iio_dev *indio_dev)
{
	return __sca3000_hw_ring_state_set(indio_dev, 0);
}
void sca3000_register_ring_funcs(struct iio_dev *indio_dev)
{
	indio_dev->ring_preenable = &sca3000_hw_ring_preenable;
	indio_dev->ring_postdisable = &sca3000_hw_ring_postdisable;
}

void sca3000_ring_int_process(u8 val, void *ring)
{
		if (val & SCA3000_INT_STATUS_THREE_QUARTERS)
		iio_push_or_escallate_ring_event(ring,
						IIO_EVENT_CODE_RING_75_FULL, 0);
	else if (val & SCA3000_INT_STATUS_HALF)
		iio_push_ring_event(ring,
				   IIO_EVENT_CODE_RING_50_FULL, 0);
}
