/*
 * Copyright (C) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * max1363_ring.c
 */

#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/workqueue.h>

#include <linux/device.h>
#include <linux/kernel.h>

#include <linux/sysfs.h>
#include <linux/list.h>
#include <linux/i2c.h>
#include <linux/rtc.h>

#include <linux/industrialio/iio.h>
#include <linux/industrialio/ring_sw.h>
#include <linux/industrialio/trigger.h>
#include <linux/industrialio/sysfs.h>

#include "max1363.h"

ssize_t max1363_scan_from_ring(struct device *dev,
			       struct device_attribute *attr,
			       char *buf)
{
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	struct max1363_data *info = dev_info->dev_data;
	int i, ret, len = 0;
	char *ring_data;

	ring_data = kmalloc(info->current_mode->numvals*2, GFP_KERNEL);
	if (ring_data == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}
	ret = dev_info->ring_access.read_last(dev_info->ring, ring_data);
	if (ret)
		goto error_free_ring_data;
	len += sprintf(buf+len, "ring ");
	for (i = 0; i < info->current_mode->numvals; i++)
		len += sprintf(buf + len, "%d ",
			       ((int)(ring_data[i*2 + 0] & 0x0F) << 8)
			       + ((int)(ring_data[i*2 + 1])));
	len += sprintf(buf + len, "\n");
	kfree(ring_data);

	return len;

error_free_ring_data:
	kfree(ring_data);
error_ret:
	return ret;
}


static int max1363_ring_preenable(struct iio_dev *indio_dev)
{
	struct max1363_data *data = indio_dev->dev_data;

	if(indio_dev->ring_access.set_bpd)
		indio_dev->ring_access.set_bpd(indio_dev->ring,
					       data->current_mode->numvals*2
					       + sizeof(s64));
	return 0;
}

static int max1363_ring_postenable(struct iio_dev *indio_dev)
{
	if (indio_dev->trig == NULL)
		return 0;
	return iio_trigger_attach_poll_func(indio_dev->trig,
					    indio_dev->pollfunc);
}

static int max1363_ring_predisable(struct iio_dev *indio_dev)
{
	if (indio_dev->trig)
		return iio_trigger_dettach_poll_func(indio_dev->trig,
						     indio_dev->pollfunc);
	else
		return 0;
}

/* As sampling only occurs on i2c comms occuring, leave the timestamping
 * until then.
 */
void max1363_poll_func_th(void *private_data)
{
	struct iio_dev *indio_dev = private_data;
	struct max1363_data *data = indio_dev->dev_data;

	schedule_work(&data->poll_work);

	return;
}

static void max1363_poll_bh_to_ring(struct work_struct *work_s)
{
	/* This shares a lot with a normal scan read - combine them? */
	struct max1363_data *data = container_of(work_s, struct max1363_data,
						 poll_work);
	struct iio_dev *indio_dev = data->indio_dev;
	struct iio_sw_ring_buffer *ring = indio_dev->ring;
	s64 time_ns;
	struct i2c_msg msg = {
		.addr = data->client->addr,
		.flags = I2C_M_RD,
		.len = data->current_mode->numvals*2,
		};
	__u8 *rxbuf;
	int ret;

	/* Ensure only one copy of this function running at a time */
	if (atomic_inc_return(&data->protect_ring) > 1)
		return;

	/* Monitor mode prevents reading */
	if (data->current_mode->numvals == 0)
		return;

	rxbuf = kmalloc(data->current_mode->numvals*2+sizeof(time_ns),
			GFP_KERNEL);
	if (rxbuf == NULL)
		return;
	msg.buf = rxbuf;
	ret = i2c_transfer(data->client->adapter, &msg, 1);
	if (ret == 1)
		ret = 0;
	else if (ret)
		goto done;
	time_ns = iio_get_time_ns();
	memcpy(rxbuf + data->current_mode->numvals*2,
	       &time_ns, sizeof(time_ns));
	indio_dev->ring_access.store_to(ring, rxbuf, time_ns);
done:
	kfree(rxbuf);

	atomic_dec(&data->protect_ring);
}

int max1363_register_ring_funcs_and_init(struct iio_dev *indio_dev)
{
	struct max1363_data *data = indio_dev->dev_data;
	int ret = 0;

	/* Effectively select the ring buffer implementation */
	iio_ring_sw_register_funcs(&data->indio_dev->ring_access);

	indio_dev->pollfunc = kzalloc(sizeof(*indio_dev->pollfunc), GFP_KERNEL);
	if (indio_dev->pollfunc == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}

	/* Configure the polling function called on trigger interrupts */
	indio_dev->pollfunc->poll_func_main = &max1363_poll_func_th;
	indio_dev->pollfunc->private_data = indio_dev;

	/* Ring buffer functions - here trigger setup related */
	indio_dev->ring_postenable = &max1363_ring_postenable;
	indio_dev->ring_preenable = &max1363_ring_preenable;
	indio_dev->ring_predisable = &max1363_ring_predisable;

	INIT_WORK(&data->poll_work, &max1363_poll_bh_to_ring);

	/* Flag that polled ring buffering is possible */
	indio_dev->modes |= INDIO_RING_TRIGGERED;

error_ret:
	return ret;
}

void max1363_ring_cleanup(struct iio_dev *indio_dev)
{
	if (indio_dev->pollfunc)
		kfree(indio_dev->pollfunc);
}

