/*
 * linux/drivers/industrialio/imu/adis16350.c
 * Copyright (C) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

/* Issues
 * Known erata on early devices
 * I have an 0741 so some of these effect me and others don't
 * Look at date code on bottom left
 * 0722 and earlier
 *
 * 1) Manual flash update fails 5% of the time - make sure
 *    to verify.
 *
 *  0805 and earlier
 * 2) Do not write to 0x2B and 0x2D as it'll kill the device.

 * 3) Status register only cleared by read of 0x3c not 0x3d.
 * 4) Autonull fails if rotation rate is > 75 degrees a second
 * 5) SLP_CNT provids option of placing device into low power sleep mode
 * (basically may not work - if needed this is a case of sending device back
 * for another one).
 */

/* To think about.
 *
 * - Multiple temp outputs, for consistency of interface across devices,
 *   supply both individual and some combination of them under temp.
 *
 * - Device supports going to sleep for set time periods (or until poked)
 *   Use this for anything?
 *
 * - Self tests?
 *
 * - Current assumption is that data ready is on line 1, events on line 2.
 *
 * - assume both alarms are not in the same mode (possible with diff thresh)
 *
 * - New means of controlling scan mode elements (and hence what goes into the
 * ring) Worth implementing for everything, or still allow approach used in
 * max1363 driver?
 */

#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/spi/spi.h>
#include <linux/delay.h>
#include <linux/log2.h>

#include <linux/sysfs.h>
#include <linux/list.h>
#include <linux/rtc.h>
#include <linux/industrialio/iio.h>
#include <linux/industrialio/sysfs.h>
#include <linux/industrialio/accel.h>
#include <linux/industrialio/gyro.h>
#include <linux/industrialio/adc.h>
#include <linux/industrialio/adis16350.h>
#include "adis16350.h"


/* Some decidely device specific elements */
#define IIO_DEV_ATTR_TEMP_GYRO_X(_show, _addr)				\
	IIO_DEVICE_ATTR(temp_gyro_x, S_IRUGO, _show, NULL, _addr)
#define IIO_DEV_ATTR_TEMP_GYRO_Y(_show, _addr)				\
	IIO_DEVICE_ATTR(temp_gyro_y, S_IRUGO, _show, NULL, _addr)
#define IIO_DEV_ATTR_TEMP_GYRO_Z(_show, _addr)				\
	IIO_DEVICE_ATTR(temp_gyro_z, S_IRUGO, _show, NULL, _addr)


/* The next 3 chuncks must have the same ordering */
static IIO_SCAN_EL(accel_x, 0, IIO_SIGNED(12), ADIS16350_X_ACCEL_ADDR);
static IIO_SCAN_EL(accel_y, 1, IIO_SIGNED(12), ADIS16350_Y_ACCEL_ADDR);
static IIO_SCAN_EL(accel_z, 2, IIO_SIGNED(12), ADIS16350_Z_ACCEL_ADDR);
static IIO_SCAN_EL(gyro_x, 3, IIO_SIGNED(14), ADIS16350_X_GYRO_ADDR);
static IIO_SCAN_EL(gyro_y, 4, IIO_SIGNED(14), ADIS16350_Y_GYRO_ADDR);
static IIO_SCAN_EL(gyro_z, 5, IIO_SIGNED(14), ADIS16350_Z_GYRO_ADDR);
static IIO_SCAN_EL(temp_x, 6, IIO_SIGNED(12), ADIS16350_X_TEMP_ADDR);
static IIO_SCAN_EL(temp_y, 7, IIO_SIGNED(12), ADIS16350_Y_TEMP_ADDR);
static IIO_SCAN_EL(temp_z, 8, IIO_SIGNED(12), ADIS16350_Z_TEMP_ADDR);
static IIO_SCAN_EL(adc_0, 9, IIO_UNSIGNED(12), ADIS16350_ADC_ADDR);
static IIO_SCAN_EL(power, 10, IIO_UNSIGNED(12), ADIS16350_POWER_SUPPLY_ADDR);

/* Endian hacks needed as we are going to be transfering
 * these within 16 bit transfers (whilst they are 8 bit)
 */
static const u16 read_all_tx_array[] = {
	be16_to_cpu(ADIS16350_READ(ADIS16350_X_ACCEL_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_Y_ACCEL_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_Z_ACCEL_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_X_GYRO_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_Y_GYRO_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_Z_GYRO_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_X_TEMP_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_Y_TEMP_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_Z_TEMP_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_ADC_ADDR)),
	be16_to_cpu(ADIS16350_READ(ADIS16350_POWER_SUPPLY_ADDR)),
};

static struct attribute *adis16350_scan_el_attrs[] = {
	&iio_scan_el_accel_x.dev_attr.attr,
	&iio_scan_el_accel_y.dev_attr.attr,
	&iio_scan_el_accel_z.dev_attr.attr,
	&iio_scan_el_gyro_x.dev_attr.attr,
	&iio_scan_el_gyro_y.dev_attr.attr,
	&iio_scan_el_gyro_z.dev_attr.attr,
	&iio_scan_el_temp_x.dev_attr.attr,
	&iio_scan_el_temp_y.dev_attr.attr,
	&iio_scan_el_temp_z.dev_attr.attr,
	&iio_scan_el_adc_0.dev_attr.attr,
	&iio_scan_el_power.dev_attr.attr,
	NULL,
};

static struct attribute_group adis16350_scan_el_group = {
	.attrs = adis16350_scan_el_attrs,
	.name = "scan_elements",
};

/* Device does interleaved reading. This is only
 * used in simple single register read cases.
 */
static int __adis16350_read_reg_16(struct device *dev,
				   u8 reg_address,
				   u16 *val)
{
	int ret;
	struct spi_message msg;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	u8 txbuf[2] = {ADIS16350_READ(reg_address)};
	u8 rxbuf[2];
	/* as the cs line is don't care between frames during a read
	 * they can be directl y combined?
	 */
	struct spi_transfer xfer []= {{
			.tx_buf = txbuf,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}, {
			.rx_buf = rxbuf,
			.bits_per_word = 16,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}};

	spi_message_init(&msg);
	spi_message_add_tail(&xfer[0], &msg);
	spi_message_add_tail(&xfer[1], &msg);

	ret = spi_sync(st->spi, &msg);
	if (ret)
		dev_err(&st->spi->dev, "problem with a 16bit read");
	*val = *(u16*)(rxbuf);
	printk("%d, %d\n", rxbuf[0], rxbuf[1]);
	return ret;
}

static int __adis16350_write_reg_8(struct device *dev,
				   u8 reg_address,
				   u8 val)
{
	struct spi_message msg;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	u8 txbuf[2];
	struct spi_transfer xfer = {
		.tx_buf = txbuf,
		.bits_per_word = 8,
		.len = 2,
		.speed_hz = st->max_speed_hz,
		.cs_change = 1,
	};

	txbuf[0] = ADIS16350_WRITE(reg_address);
	txbuf[1] = val;
	spi_message_init(&msg);
	spi_message_add_tail(&xfer, &msg);

	return spi_sync(st->spi, &msg);
}

static ssize_t adis16350_read_reg_u16(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	ssize_t len;
	int ret;
	u16 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = __adis16350_read_reg_16(dev,this_attr->address, &val);
	if (ret)
		return ret;
	len = sprintf(buf, "%u\n", val);

	return len;
}
/* Is there a better / standard way of doing this? */
static inline s16 __adis16350_sign_extend(s16 in, int bits)
{
	/* most of this should hopefuly optimize to relevant constant */
	s32 replacemask = 0;
	int i;
	for(i = bits; i < 16; i++)
		replacemask |= (1 << i);

	if(in & (1 << (bits - 1 )))
		return in | replacemask;
	else
		return in & ~replacemask;
}

/* Read values that need sign extending */
static ssize_t __adis16350_read_reg_signextend(struct device *dev,
						 struct device_attribute *attr,
						 char *buf,
						 int bits)
{

	int ret;
	s16 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = __adis16350_read_reg_16(dev,this_attr->address, &val);
	if (ret)
		return ret;
	/* Sign extend to 16 bit */
	val = __adis16350_sign_extend(val, bits);

	return sprintf(buf, "%d\n", val);
}

//static ssize_t adis16350_read_reg_s14(struct device *dev,
//				      struct device_attribute *attr, char *buf)
//{
//	return __adis16350_read_reg_signextend(dev, attr, buf, 14);
//}

static u8 count_to_right(u16 full, u16 mask)
{
	u8 count = 0;
	while(mask){
		 mask >>= 1;
		if (mask & full)
			count++;
	}
	return count;
}
/* FIXME: Can I roll this into the core - it's hardly device specific!*/
/* Add single element reading to the ring buffer */
static int __adis16350_read_reg_from_ring(s16 *val,
					  int label,
					  struct iio_dev *indio_dev)
{
	struct iio_scan_el *el;
	u16 *data;
	int ret;
	int i;
	for (i = 0; i < ARRAY_SIZE(adis16350_scan_el_attrs) - 1; i++) {
		el = to_iio_scan_el((struct device_attribute *)
				    (adis16350_scan_el_attrs[i]));
		if (el->label == label)
			break;
	}
	if (el->mask & indio_dev->scan_mask) {
		data = kmalloc(indio_dev->scan_count*sizeof(u16),
			       GFP_KERNEL);
		if (data == NULL) {
			ret = -ENOMEM; //wrong goto
			goto error_free_data;
		}
		ret = indio_dev->ring_access.read_last(indio_dev->ring, (u8 *)(data));
		if (ret)
			goto error_free_data;

		*val = data[count_to_right(indio_dev->scan_mask,
					  el->mask)];
		kfree(data);
		return 1;
	}
	return 0;

error_free_data:
	kfree(data);
	return ret;
}

static inline ssize_t __adis16350_read_reg_s_ringpos(struct device *dev,
					      struct device_attribute *attr,
					      char *buf,
					      int bits)
{
	ssize_t ret = 0;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	s16 val;
	mutex_lock(&indio_dev->mlock);
	if (indio_dev->currentmode == INDIO_RING_TRIGGERED)
		ret = __adis16350_read_reg_from_ring(&val, this_attr->address,
						     indio_dev);
	if (ret < 0)
		goto error_ret;
	/* only if needed read directly from the address */
	if (ret == 0) {
		ret = __adis16350_read_reg_16(dev, this_attr->address, &val);
		if (ret)
			goto error_ret;
	}
	/* Sign extend to 16 bit */
	val = __adis16350_sign_extend(val, 14);

	ret = sprintf(buf, "%d\n", val);
error_ret:
	mutex_unlock(&indio_dev->mlock);
	return ret;
}

static ssize_t adis16350_read_reg_s14_ringpos(struct device *dev,
					      struct device_attribute *attr,
					      char *buf)
{
	return __adis16350_read_reg_s_ringpos(dev, attr, buf, 14);
}

static ssize_t adis16350_read_reg_s12_ringpos(struct device *dev,
					      struct device_attribute *attr,
					      char *buf)
{
	return __adis16350_read_reg_s_ringpos(dev, attr, buf, 12);
}

static ssize_t adis16350_read_reg_s13(struct device *dev,
				      struct device_attribute *attr, char *buf)
{
	return __adis16350_read_reg_signextend(dev, attr, buf, 13);
}

static ssize_t adis16350_read_reg_s12(struct device *dev,
				      struct device_attribute *attr, char *buf)
{
	return __adis16350_read_reg_signextend(dev, attr, buf, 12);
}

static ssize_t adis16350_read_reg_u12(struct device *dev,
				      struct device_attribute *attr,
				      char *buf)
{
	ssize_t len;
	int ret;
	u16 val;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = __adis16350_read_reg_16(dev,this_attr->address, &val);
	if (ret)
		return ret;
	/* clean higher bits */
	val &= 0x0FFF;
	len = sprintf(buf, "%u\n", val);

	return len;
}

/* used to clean up some code when dealing with 2 part writes */

static ssize_t adis16350_write_reg_s16(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf,
				       size_t len)
{
	int ret;
	long value;
	s16 val;

	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	ret = strict_strtol(buf, 10, &value);
	if (ret)
		goto error_ret;
	val = value;
	ret = __adis16350_write_reg_8(dev, this_attr->address, val);
	if (ret)
		goto error_ret;
	ret = __adis16350_write_reg_8(dev, this_attr->address + 1, val >> 8);
	if (ret)
		goto error_ret;
	return len;

error_ret:
	return ret;

}
static IIO_DEVICE_ATTR(flash_write_count, S_IRUGO, adis16350_read_reg_u16,
		       NULL, ADIS16350_ENDURANCE_LSB_ADDR);

/* FIXME - add caps to these to prevent out of range attempts? */
static IIO_DEV_ATTR_ACCEL_X_OFFSET(S_IWUSR | S_IRUGO,
				   adis16350_read_reg_s12,
				   adis16350_write_reg_s16,
				   ADIS16350_X_ACCEL_OFF_LSB_ADDR);

static IIO_DEV_ATTR_ACCEL_Y_OFFSET(S_IWUSR | S_IRUGO,
				   adis16350_read_reg_s12,
				   adis16350_write_reg_s16,
				   ADIS16350_Y_ACCEL_OFF_LSB_ADDR);

static IIO_DEV_ATTR_ACCEL_Z_OFFSET(S_IWUSR | S_IRUGO,
				   adis16350_read_reg_s12,
				   adis16350_write_reg_s16,
				   ADIS16350_Z_ACCEL_OFF_LSB_ADDR);


static IIO_DEV_ATTR_GYRO_X_OFFSET(S_IWUSR | S_IRUGO,
				  adis16350_read_reg_s13,
				  adis16350_write_reg_s16,
				  ADIS16350_X_GYRO_OFF_LSB_ADDR);
static IIO_DEV_ATTR_GYRO_Y_OFFSET(S_IWUSR | S_IRUGO,
				  adis16350_read_reg_s13,
				  adis16350_write_reg_s16,
				  ADIS16350_Y_GYRO_OFF_LSB_ADDR);
static IIO_DEV_ATTR_GYRO_Z_OFFSET(S_IWUSR | S_IRUGO,
				  adis16350_read_reg_s13,
				  adis16350_write_reg_s16,
				  ADIS16350_Z_GYRO_OFF_LSB_ADDR);

/* make these deal with the ring when implemented */
static IIO_DEV_ATTR_ACCEL_X(adis16350_read_reg_s14_ringpos,
			    ADIS16350_X_ACCEL_ADDR);

static IIO_DEV_ATTR_ACCEL_Y(adis16350_read_reg_s14_ringpos,
			    ADIS16350_Y_ACCEL_ADDR);

static IIO_DEV_ATTR_ACCEL_Z(adis16350_read_reg_s14_ringpos,
			    ADIS16350_Z_ACCEL_ADDR);

static IIO_DEV_ATTR_GYRO_X(adis16350_read_reg_s14_ringpos,
			    ADIS16350_X_GYRO_ADDR);

static IIO_DEV_ATTR_GYRO_Y(adis16350_read_reg_s14_ringpos,
			    ADIS16350_Y_GYRO_ADDR);

static IIO_DEV_ATTR_GYRO_Z(adis16350_read_reg_s14_ringpos,
			    ADIS16350_Z_GYRO_ADDR);

static IIO_DEV_ATTR_TEMP_GYRO_X(adis16350_read_reg_s12_ringpos,
				ADIS16350_X_TEMP_ADDR);

static IIO_DEV_ATTR_TEMP_GYRO_Y(adis16350_read_reg_s12_ringpos,
				ADIS16350_Y_TEMP_ADDR);

static IIO_DEV_ATTR_TEMP_GYRO_Z(adis16350_read_reg_s12_ringpos,
				ADIS16350_Z_TEMP_ADDR);

static IIO_DEV_ATTR_ADC(0, adis16350_read_reg_u12,
			ADIS16350_ADC_ADDR);

static IIO_DEVICE_ATTR(supply_voltage, S_IRUGO, adis16350_read_reg_u12, NULL,
		       ADIS16350_POWER_SUPPLY_ADDR);

/* Not entirely happy with this approach, but the floating point
 * arithmetic needed to get all the frequencies would be pain.
 * For now the alternative time base is deliberately avoiding as
 * it results in the device switching to low power mode and changes
 * the possible spi frequencies etc.
*/
/* This device doesn't really fit this frame work - rather
 * complex way of specifying sampling frequencies.
 * Here I've elected to provide a useful subset.
 * Could also provide device specific controls to set precise frequencies? */


static const struct adis16350_sampling_freq adis16350_samp_freq_tb[] = {
	{"819.2", ADIS16350_BASE_610MICROSECS, 0x01},
	{"546.1", ADIS16350_BASE_610MICROSECS, 0x02},
	{"409.6", ADIS16350_BASE_610MICROSECS, 0x03},
	{"327.7", ADIS16350_BASE_610MICROSECS, 0x04},
	{"273.1", ADIS16350_BASE_610MICROSECS, 0x05},
	{"234.1", ADIS16350_BASE_610MICROSECS, 0x06},
	{"204.8", ADIS16350_BASE_610MICROSECS, 0x07},
	{"163.8", ADIS16350_BASE_610MICROSECS, 0x09},
	/* above this point the max spi frequency drops to 300kHz */
	{"136.5", ADIS16350_BASE_610MICROSECS, 11},
	{"102.4", ADIS16350_BASE_610MICROSECS, 16},
	{"52.9", ADIS16350_BASE_610MICROSECS, 31},
	{"25.6", ADIS16350_BASE_610MICROSECS, 63},
	{"1", ADIS16350_BASE_18921MICROSECS, 52},
};

static ssize_t adis16350_read_samp_freq(struct device *dev,
					struct device_attribute *attr,
					char *buf)
{
	int i, ret;
	u16 val;
	int len = 0;

	ret = __adis16350_read_reg_16(dev, ADIS16350_SAMP_PERIOD_LSB_ADDR,
				      &val);
	if (ret)
		goto error_ret;
	for (i = 0; i < ARRAY_SIZE(adis16350_samp_freq_tb); i++)
		if ((adis16350_samp_freq_tb[i].timebase == ((val >> 7) & 0x01))
		    && adis16350_samp_freq_tb[i].multiplier == (val & 0x7F)) {
			len += sprintf(buf, "%s\n",
				       adis16350_samp_freq_tb[i].name);
			break;
		}
	return len;
error_ret:
	return ret;
}

static ssize_t adis16350_write_samp_freq(struct device *dev,
					 struct device_attribute *attr,
					 const char *buf,
					 size_t len)
{
	int i, ret;
	u8 val;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	bool found = false;

	for (i = 0; i < ARRAY_SIZE(adis16350_samp_freq_tb); i++)
		if (strncmp(adis16350_samp_freq_tb[i].name, buf,
			    strlen(buf)-1) == 0
		    && (strlen(buf)-1
			== strlen(adis16350_samp_freq_tb[i].name))) {
			found = true;
			break;
		}
	if (!found) {
		ret = -EINVAL;
		goto error_ret;
	}
	val = adis16350_samp_freq_tb[i].timebase << 7
		| adis16350_samp_freq_tb[i].multiplier;

	ret = __adis16350_write_reg_8(dev, ADIS16350_SAMP_PERIOD_LSB_ADDR, val);
	if (ret)
		goto error_ret;
	/* obviously don't change this until after frequency command is sent*/
	if (val > 0x09)
		st->max_speed_hz = 200000;
	else
		st->max_speed_hz = 0;

	return len;

error_ret:
	return ret;
}

static IIO_DEV_ATTR_SAMP_FREQ(S_IWUSR | S_IRUGO, adis16350_read_samp_freq,
			      adis16350_write_samp_freq);

static ssize_t adis16350_command_issue(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf,
				       size_t len)
{
	int ret;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);

	if (buf[0] == 0 )
		return len;
	ret = __adis16350_write_reg_8(dev, ADIS16350_COMMAND_LSB_ADDR,
				      this_attr->address);
	if (ret)
		return ret;

	return len;
}

static IIO_DEVICE_ATTR(auto_null, S_IWUSR, NULL, adis16350_command_issue,
		       ADIS16350_COMMAND_AUTO_NULL);

static IIO_DEVICE_ATTR(auto_null_precision,
		       S_IWUSR, NULL, adis16350_command_issue,
		       ADIS16350_COMMAND_PREC_AUTO_NULL);

static IIO_DEVICE_ATTR(factory_reset,
		       S_IWUSR, NULL, adis16350_command_issue,
		       ADIS16350_COMMAND_FACTORY_RESTORE);

static IIO_DEVICE_ATTR(flash_store,
		       S_IWUSR, NULL, adis16350_command_issue,
		       ADIS16350_COMMAND_FLASH_UPDATE);


/* only one bit of the 'address' should be set. */
static ssize_t adis16350_msc_command_issue(struct device *dev,
					   struct device_attribute *attr,
					   const char *buf,
					   size_t len)
{
	int ret;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	u16 val = this_attr->address;

	if (buf[0] == 0 )
		return len;

	if (val >> 8)
		ret = __adis16350_write_reg_8(dev, ADIS16350_MSC_CTRL_MSB_ADDR,
					      val >> 8);
	else
		ret = __adis16350_write_reg_8(dev, ADIS16350_MSC_CTRL_LSB_ADDR,
					      val);
	if (ret)
		return ret;

	return len;
}

static ssize_t adis16350_misc_ctrl_bit_check(struct device *dev,
					     struct device_attribute *attr,
					     char *buf)
{
	int ret;
	u16 val;
	int len;
	struct iio_dev_attr *dev_attr = to_iio_dev_attr(attr);
	ret = __adis16350_read_reg_16(dev, ADIS16350_MSC_CTRL_LSB_ADDR, &val);
	if (ret)
		return ret;
	if (dev_attr->address & val)
		len = sprintf(buf, "1\n");
	else
		len = sprintf(buf, "0\n");

	return len;
}


static ssize_t adis16350_misc_ctrl_bit_set(struct device *dev,
					   struct device_attribute *attr,
					   const char *buf,
					   size_t len)
{
	int ret;
	struct iio_dev_attr *dev_attr = to_iio_dev_attr(attr);
	u8 mask = dev_attr->address;
	u16 current_val;

/* read current value */
	ret = __adis16350_read_reg_16(dev, ADIS16350_MSC_CTRL_LSB_ADDR,
				      &current_val);
	if (ret)
		goto error_ret;
/* need to change? */

	if (buf[0] == '0' && (current_val & mask))
		ret = __adis16350_write_reg_8(dev,
					      ADIS16350_MSC_CTRL_LSB_ADDR,
					      current_val & ~mask);
	else if (buf[0] != '0' && !(current_val & mask))
		ret = __adis16350_write_reg_8(dev,
					      ADIS16350_MSC_CTRL_LSB_ADDR,
					      current_val | mask);
	if (ret)
		goto error_ret;

	return len;
error_ret:
	return ret;

}

static IIO_DEVICE_ATTR(gyro_bias_comp,
		       S_IWUSR | S_IRUGO, adis16350_misc_ctrl_bit_check,
		       adis16350_misc_ctrl_bit_set,
		       ADIS16350_MSC_CTRL_LSB_GYRO_COMP_FOR_ACCEL);

static IIO_DEVICE_ATTR(accel_origin_align,
		       S_IWUSR | S_IRUGO, adis16350_misc_ctrl_bit_check,
		       adis16350_misc_ctrl_bit_set,
		       ADIS16350_MSC_CTRL_LSB_ACCEL_ORIGIN_ALIGN);

/* write value then latch */
static ssize_t adis16350_write_dac(struct device *dev,
				   struct device_attribute *attr,
				   const char *buf,
				   size_t len)
{
	int ret;
	ulong value;
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	u16 val;

	ret = strict_strtoul(buf, 10, &value);
	if (ret)
		goto error_ret;
	val = value &0x0FFF;

	ret = __adis16350_write_reg_8(dev, this_attr->address, val);
	if (ret)
		goto error_ret;
	ret = __adis16350_write_reg_8(dev, this_attr->address + 1, val >> 8);
	if (ret)
		goto error_ret;
	ret = __adis16350_write_reg_8(dev, ADIS16350_COMMAND_LSB_ADDR,
				      ADIS16350_COMMAND_DAC_DATA_LATCH);
	if (ret)
		goto error_ret;
	return len;

error_ret:
	return ret;
}

static IIO_DEVICE_ATTR(dac_0,
		       S_IWUSR | S_IRUGO, adis16350_read_reg_u12,
		       adis16350_write_dac, ADIS16350_DAC_LSB_ADDR);

static ssize_t adis16350_read_gyro_range(struct device *dev,
					 struct device_attribute *attr,
					 char *buf)
{
	int ret;
	u16 val;
	int len;

	ret = __adis16350_read_reg_16(dev, ADIS16350_SENSITIVITY_ADDR, &val);
	if (ret)
		return ret;
	switch (val & 0x7)
	{
	case ADIS16350_SENSITIVITY_300_DPS:
		len = sprintf(buf, "300\n");
		break;
	case ADIS16350_SENSITIVITY_150_DPS:
		len = sprintf(buf, "150\n");
		break;
	case ADIS16350_SENSITIVITY_75_DPS:
		len = sprintf(buf, "75\n");
		break;
	default:
		len = 0;
	}
	return len;
}

static ssize_t adis16350_write_gyro_range(struct device *dev,
					  struct device_attribute *attr,
					  const char *buf,
					  size_t len)
{
	int ret;
	ulong value;
	u8 val;
	ret = strict_strtol(buf, 10, &value);
	if (ret)
		goto error_ret;
	switch (value)
	{
	case 300:
		val = ADIS16350_SENSITIVITY_300_DPS;
		break;
	case 150:
		val = ADIS16350_SENSITIVITY_150_DPS;
		break;
	case 75:
		val = ADIS16350_SENSITIVITY_75_DPS;
		break;
	default:
		ret = -EINVAL;
		goto error_ret;
	}
	ret = __adis16350_write_reg_8(dev, ADIS16350_SENSITIVITY_ADDR, val);
	if (ret)
		goto error_ret;
	return len;
error_ret:
	return ret;

}
static IIO_DEV_ATTR_GYRO_RANGE(S_IWUSR | S_IRUGO, adis16350_read_gyro_range,
		       adis16350_write_gyro_range);

static IIO_CONST_ATTR_GYRO_RANGE_AVAILABLE("75 150 300");

/* this must be a power of 2 */
static ssize_t adis16350_write_filter_taps(struct device *dev,
					   struct device_attribute *attr,
					   const char *buf,
					   size_t len)
{
	int ret;
	ulong value;
	u8 val;
	ret = strict_strtol(buf, 10, &value);
	if (ret)
		goto error_ret;
	val = ilog2(value);
	/* note the docs refer to 7 steps but only list 6 */
	if (val > 7) {
		ret = -EINVAL;
		goto error_ret;
	}
	ret = __adis16350_write_reg_8(dev, ADIS16350_FILTERING_ADDR, val);
	if (ret)
		goto error_ret;
	return len;
error_ret:
	return ret;
}

static ssize_t adis16350_read_filter_taps(struct device *dev,
					  struct device_attribute *attr,
					  char *buf)
{
	int ret;
	u16 val;

	ret = __adis16350_read_reg_16(dev, ADIS16350_SENSITIVITY_ADDR, &val);
	if (ret)
		return ret;
	val &= 0x7;
	val = 1 << val;

	return sprintf(buf, "%u\n", val);
}
/* Pretty device specific */
static IIO_DEVICE_ATTR(filter_taps,
		       S_IWUSR | S_IRUGO, adis16350_read_filter_taps,
		       adis16350_write_filter_taps, 0);



/* need cs_change specified? */
static int adis16350_read_all(struct adis16350_state *st, u16 *rx_array)
{
	struct spi_transfer *xfers;
	int ret, i, j = 0;
	struct spi_message msg;


	xfers = kzalloc((st->indio_dev->scan_count + 1)
			*sizeof(struct spi_transfer),
			GFP_KERNEL);
//FIXME: Unhandled error 
	for (i = 0; i < ARRAY_SIZE(read_all_tx_array); i++)
		if(st->indio_dev->scan_mask & (1 << i)) {
			xfers[j].tx_buf = read_all_tx_array + i;
			xfers[j].bits_per_word = 16;
			xfers[j].len = 2;
			xfers[j].speed_hz = st->max_speed_hz;
			xfers[j+1].rx_buf = rx_array + j;
			j++;
		}
	xfers[j].bits_per_word = 16;
	xfers[j].len = 2;
	xfers[j].speed_hz = st->max_speed_hz;


	spi_message_init(&msg);
	for(j = 0; j < st->indio_dev->scan_count + 1; j++)
		spi_message_add_tail(&xfers[j], &msg);
	ret =  spi_sync(st->spi, &msg);
	kfree(xfers);

	return ret;
};

/* mainly used to test ring buffer style reads */
static ssize_t adis16350_scan_test(struct device *dev,
				   struct device_attribute *attr,
				   char *buf)
{

	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	int ret, i, j = 0, len = 0;
	s16 rxarray[st->indio_dev->scan_count];

	struct iio_scan_el *sm_el;
	/* Don't want the scan mode changing during a can */
	mutex_lock(&indio_dev->mlock);
	ret = adis16350_read_all(st, rxarray);
	if (ret) {
		mutex_unlock(&indio_dev->mlock);
		return ret;
	}
	mutex_unlock(&indio_dev->mlock);

/* array is null termintate so subtract 1 from size */
	for(i = 0; i < ARRAY_SIZE(adis16350_scan_el_attrs) - 1; i++)
		if(indio_dev->scan_mask & (1 << i)) {
			sm_el = to_iio_scan_el((struct device_attribute *)
					       (adis16350_scan_el_attrs[i]));
			switch (sm_el->bit_count) {
			case IIO_SIGNED(12):
				if((rxarray[j] & (1 << (12 - 1))))
					rxarray[j] |= 0xF000;
				else
					rxarray[j] &= ~0xF000;
				len += sprintf(buf + len, "%d ", rxarray[j]);
				break;
			case IIO_SIGNED(14):
				if((rxarray[j] & (1 << (14 - 1))))
					rxarray[j] |= 0xC000;
				else
					rxarray[j] &= ~0xC000;
				len += sprintf(buf + len, "%d ", rxarray[j]);
				break;
			case IIO_UNSIGNED(12):
				rxarray[j] &= 0x0FFF;
				len += sprintf( buf + len, "%u ", rxarray[j]);
				break;
			}
			j++;
		}
	len += sprintf(buf+len, "\n");

	return len;

}

static IIO_DEV_ATTR_SCAN(adis16350_scan_test);

static struct attribute *adis16350_attributes[] = {
	&iio_dev_attr_flash_write_count.dev_attr.attr,
	&iio_dev_attr_accel_x_offset.dev_attr.attr,
	&iio_dev_attr_accel_y_offset.dev_attr.attr,
	&iio_dev_attr_accel_z_offset.dev_attr.attr,
	&iio_dev_attr_gyro_x_offset.dev_attr.attr,
	&iio_dev_attr_gyro_y_offset.dev_attr.attr,
	&iio_dev_attr_gyro_z_offset.dev_attr.attr,
	&iio_dev_attr_accel_x.dev_attr.attr,
	&iio_dev_attr_accel_y.dev_attr.attr,
	&iio_dev_attr_accel_z.dev_attr.attr,
	&iio_dev_attr_gyro_x.dev_attr.attr,
	&iio_dev_attr_gyro_y.dev_attr.attr,
	&iio_dev_attr_gyro_z.dev_attr.attr,
	&iio_dev_attr_temp_gyro_x.dev_attr.attr,
	&iio_dev_attr_temp_gyro_y.dev_attr.attr,
	&iio_dev_attr_temp_gyro_z.dev_attr.attr,
	&iio_dev_attr_adc_0.dev_attr.attr,
	&iio_dev_attr_supply_voltage.dev_attr.attr,
	&iio_dev_attr_sampling_frequency.dev_attr.attr,

	&iio_dev_attr_dac_0.dev_attr.attr,
	&iio_dev_attr_gyro_range.dev_attr.attr,
	&iio_const_attr_gyro_range_available.dev_attr.attr,
	&iio_dev_attr_filter_taps.dev_attr.attr,
/* should probably be in a config command directory? */
	/* going to need specific documentation - nonobvious!*/
	&iio_dev_attr_auto_null.dev_attr.attr,
	&iio_dev_attr_auto_null_precision.dev_attr.attr,
	&iio_dev_attr_flash_store.dev_attr.attr,
	&iio_dev_attr_factory_reset.dev_attr.attr,
	&iio_dev_attr_gyro_bias_comp.dev_attr.attr,
	&iio_dev_attr_accel_origin_align.dev_attr.attr,
	&iio_dev_attr_scan.dev_attr.attr,
	NULL,
};

static const struct attribute_group adis16350_attribute_group = {
	.attrs = adis16350_attributes,
};
/* SHARES A LOT WITH SINGLE BIT SET */
static ssize_t adis16350_read_data_ready_config(struct device *dev,
						struct device_attribute *attr,
						char *buf)
{
	ssize_t len;
	int ret, set;
	u16 val;

	ret = __adis16350_read_reg_16(dev, ADIS16350_MSC_CTRL_LSB_ADDR, &val);
	if (ret)
		return ret;

	set = (val & ADIS16350_MSC_CTRL_LSB_DATA_RDY_ENABLE) ? 1 : 0;
	len = sprintf(buf, "%d\n", set);
	return len;
}

static int __adis16350_write_data_ready_config(struct device *dev,
					       struct iio_event_handler_list
					       *list,
					       bool toset)
{
	int ret;
	u16 val;
	const u16 mask = ADIS16350_MSC_CTRL_LSB_DATA_RDY_ENABLE;

	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	int set;

	ret = __adis16350_read_reg_16(dev, ADIS16350_MSC_CTRL_LSB_ADDR, &val);
	if (ret)
		goto error_ret;
	set = (val & mask);

	if(!(toset) && (set)) {
		ret = __adis16350_write_reg_8(dev, ADIS16350_MSC_CTRL_LSB_ADDR,
					      val & ~mask);
		/* As this is the only use of this interrupt line, no more
		   interrupts can be generated - make sure we've processed
		   any that have been scheduled
		*/
		flush_scheduled_work();
		ret = iio_remove_event_from_list(list);


		printk("disabling data ready? %u\n", val & ~mask);

	}
	else if (toset && !(set)) {
		ret = __adis16350_write_reg_8(dev, ADIS16350_MSC_CTRL_LSB_ADDR,
					      val | mask);
		printk("enabling data ready? %u\n", val | mask);
		ret = iio_add_event_to_list(list,
					    &indio_dev->interrupts[0]->ev_list);
	}
error_ret:
	return ret;

}
static ssize_t adis16350_write_data_ready_config(struct device *dev,
						 struct device_attribute *attr,
						 const char *buf,
						 size_t len)
{
	int ret;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct iio_event_attr *this_attr = to_iio_event_attr(attr);
	bool toset;
	if (indio_dev->currentmode == INDIO_RING_TRIGGERED)
		return len;
	toset = (buf[0] == '0') ? 0 : 1;

	ret = __adis16350_write_data_ready_config(dev, this_attr->listel,
						  toset);
	if (ret)
		goto error_ret;
	return len;
error_ret:
	return ret;
}
/* Device does not latch the interrupt! */

static int adis16350_data_rdy_event_th(struct iio_dev *dev_info,
				       int index,
				       s64 timestamp,
				       int no_test)
{
	struct adis16350_state *st = dev_info->dev_data;
	st->last_timestamp_datardy = timestamp;
	schedule_work(&st->work_data_rdy_event);
	return IRQ_HANDLED;
}

static int adis16350_data_rdy_ring_th(struct iio_dev *dev_info,
				      int index,
				      s64 timestamp,
				      int no_test)
{
	struct adis16350_state *st = dev_info->dev_data;
	st->last_timestamp_datardy = timestamp;
	schedule_work(&st->work_data_rdy_ring);
	return IRQ_HANDLED;
}

static void adis16350_data_rdy_bh_to_event(struct work_struct *work_s)
{
	struct adis16350_state *st
		= container_of(work_s, struct adis16350_state,
			       work_data_rdy_event);

	iio_push_event(st->indio_dev, 0, IIO_EVENT_CODE_DATA_RDY,
		       st->last_timestamp_datardy);

	enable_irq(st->pdata->irq_datardy);
}

static void adis16350_data_rdy_bh_to_ring(struct work_struct *work_s)
{
	struct adis16350_state *st
		= container_of(work_s, struct adis16350_state,
			       work_data_rdy_ring);
	u16 rxarray[st->indio_dev->scan_count];
	int ret;

	ret = adis16350_read_all(st, rxarray);
	if (ret)
		return;

	st->indio_dev->ring_access.store_to(st->indio_dev->ring,
					    (u8 *)(rxarray),
					    st->last_timestamp_datardy);
	enable_irq(st->pdata->irq_datardy);
}

static void adis16350_alarm_bh_no_check(struct work_struct *work_s)
{
	struct iio_work_cont *wc = to_iio_work_cont_no_check(work_s);
	struct adis16350_state *st = wc->st;
	if (st->current_alms[0])
		iio_push_event(st->indio_dev, 1,
			       adis16350_alarms[st->current_alms[0]]
			       .event_code,
			       st->last_timestamp_event);
	else
		iio_push_event(st->indio_dev, 1,
			       adis16350_alarms[st->current_alms[1]]
			       .event_code,
			       st->last_timestamp_event);

	enable_irq(st->pdata->irq_event);
}
static void adis16350_alarm_bh_check(struct work_struct *work_s)
{
	struct iio_work_cont *wc = to_iio_work_cont_check(work_s);
	struct adis16350_state *st = wc->st;
/* Establish which alarms have been triggered */
	u16 val = 0;

	__adis16350_read_reg_16(st->indio_dev->dev,
				ADIS16350_STATUS_LSB_ADDR,
				      &val);
	if (val & ADIS16350_STATUS_ALM_1)
		iio_push_event(st->indio_dev, 1,
			       adis16350_alarms[st->current_alms[0]]
			       .event_code,
			       st->last_timestamp_event);
	if (val & ADIS16350_STATUS_ALM_2)
		iio_push_event(st->indio_dev, 1,
			       adis16350_alarms[st->current_alms[1]]
			       .event_code,
			       st->last_timestamp_event);

	enable_irq(st->pdata->irq_event);
}

IIO_EVENT_ATTR_DATA_RDY(adis16350_read_data_ready_config,
			  adis16350_write_data_ready_config,
			  0,
			  &adis16350_data_rdy_event_th);


static int adis16350_alarm_event_th(struct iio_dev *dev_info,
				    int index,
				    s64 timestamp,
				    int no_test)
{
	struct adis16350_state *st = dev_info->dev_data;
	st->last_timestamp_event = timestamp;
	if(no_test)
		schedule_work(&st->work_alarm_event.ws_nocheck);
	else
		schedule_work(&st->work_alarm_event.ws);

	return IRQ_HANDLED;
}

IIO_EVENT_SH(alarms, adis16350_alarm_event_th);

/* why the heck do I need this, I'm maintaining state - don't care for now!*/
static int __adis16350_alarm_estab_location(struct adis16350_state *st,
					    enum adis16350_alarm_mode mod)
{
	if (st->current_alms[0] == mod)
		return 1;
	if (st->current_alms[1] == mod)
		return 2;
	return 0;
}

static ssize_t adis16350_alarm_show(struct device *dev,
				    struct device_attribute *attr,
				    char *buf)
{

	int ret;
	struct iio_event_attr *this_attr = to_iio_event_attr(attr);
	enum adis16350_alarm_mode mod = this_attr->mask;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;

	ret = __adis16350_alarm_estab_location(st, mod);
	if (ret < 0)
		goto error_ret;
	return sprintf(buf, "%d\n", ret);
error_ret:
	return ret;
}

static int __adis16350_enable_alarm(struct device *dev,
				    int location,
				    struct adis16350_state *st,
				    enum adis16350_alarm_mode mod)
{
	int ret;
	u16 val;
	struct spi_message msg;

	u8 txbuf[12];
	int i;
	struct spi_transfer xfers[] = {{
			.tx_buf = txbuf,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}, {
			.tx_buf = txbuf + 2,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}, {
			.tx_buf = txbuf + 4,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}, {
			.tx_buf = txbuf + 6,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}, {
			.tx_buf = txbuf + 8,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		}, {
			.tx_buf = txbuf + 10,
			.bits_per_word = 8,
			.len = 2,
			.speed_hz = st->max_speed_hz,
			.cs_change = 1,
		},
	};

	ret = __adis16350_read_reg_16(dev, ADIS16350_ALM_CTRL_LSB_ADDR, &val);
	if (ret)
		goto error_ret;

	st->current_alms[location - 1] = mod;
	st->next_alarm = (st->next_alarm > 1) ? 1 : 2;
	/* mask out alarm source selection */
	val &= ~((ADIS16350_ALM_CTRL_SEL_TO_ALM_NO(location,
						   ADIS16350_ALM_SEL_MASK)
		  << 8) | ADIS16350_ALM_CTRL_LSB_ROC(location));

	val |= (ADIS16350_ALM_CTRL_SEL_TO_ALM_NO(location,
						 adis16350_alarms[mod].mask)
		<< 8);

	val |= (adis16350_alarms[mod].roc)
		? ADIS16350_ALM_CTRL_LSB_ROC(location) : 0;
	/* enable alarms etc*/
	/* FIXME - not here */
	val |= ADIS16350_ALM_CTRL_LSB_ALM_LINE_2;

	/* disbable the buffer*/
	txbuf[0] = ADIS16350_WRITE(ADIS16350_ALM_CTRL_LSB_ADDR);
	txbuf[1] = val & ~ADIS16350_ALM_CTRL_LSB_ENABLE;

	txbuf[2] = ADIS16350_WRITE(ADIS16350_ALM_SAMP_THRESH_ADDR(location));
	txbuf[3] = st->alarm_val[mod];

	txbuf[4] = ADIS16350_WRITE
		(ADIS16350_ALM_SAMP_THRESH_ADDR(location) + 1);
	txbuf[5] = ((st->alarm_val[mod] >> 8) & 0x3F)
		| (adis16350_alarms[mod].pol << 7);

	txbuf[6] = ADIS16350_WRITE(ADIS16350_ALM_SAMP_SIZE_ADDR(location));
	txbuf[7] = st->alarm_samp[mod];
/* setup alarm */
	txbuf[8] = ADIS16350_WRITE(ADIS16350_ALM_CTRL_LSB_ADDR + 1);
	txbuf[9] = val >> 8;
/* enable alarm */
	txbuf[10] = ADIS16350_WRITE(ADIS16350_ALM_CTRL_LSB_ADDR);
	txbuf[11] = val | ADIS16350_ALM_CTRL_LSB_ENABLE;

	spi_message_init(&msg);
	for( i = 0; i < ARRAY_SIZE(xfers); i++)
		spi_message_add_tail(&xfers[i], &msg);
	return spi_sync(st->spi, &msg);
	return 0;
error_ret:
	return ret;

}

/* This came out annoying complex, and is still a little more restrictive
 * that the device (which allows both alarms to be on the same thing, just
 * with different thresholds */
static ssize_t adis16350_alarm_store(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf,
				     size_t len)
{
	/* if 0 is written we need to check if the alarm is currently enabled
	 * or not */
	int location, ret;
	u16 val;/*value alm ctrl reg */
	u8 msb;

	struct iio_event_attr *this_attr = to_iio_event_attr(attr);
	enum adis16350_alarm_mode mod = this_attr->mask;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	/* if disable requested */
	if (buf[0]== '0') {
		location = __adis16350_alarm_estab_location(st, mod);
		if (location < 0) {
			ret = location;
			goto error_ret;
		}
		if (location == 0)
			goto done;
		ret = __adis16350_read_reg_16(dev, ADIS16350_ALM_CTRL_LSB_ADDR,
					      &val);
		if (ret)
			goto error_ret;

		msb = val >> 8;
		/* need to disable the alarm at location */
		msb &= ~(ADIS16350_ALM_CTRL_SEL_TO_ALM_NO
			 (location, ADIS16350_ALM_SEL_MASK));
		ret = __adis16350_write_reg_8(dev,
					      ADIS16350_ALM_CTRL_MSB_ADDR,
					      msb);
		if (ret)
			goto error_ret;
		st->next_alarm = location;
		st->current_alms[location - 1] = adis16350_alm_disabled;
		ret = iio_remove_event_from_list(this_attr->listel);
		if (ret)
			goto error_ret;
		goto done;
	}


	location = __adis16350_alarm_estab_location(st, mod);
	if (location < 0) {
		ret = location;
		goto error_ret;
	}
	if (location != 0)
		goto done;

	if (st->next_alarm > 1)
		location = 2;
	else
		location = 1;

	if (st->current_alms[location - 1] == adis16350_alm_disabled) {
		/* Make sure handler is in place before enabling.
		 * This is only necessary if we aren't simply replacing
		 * another alarm.*/
		/* Technically as there is only one handler we could add it
		 * the start then ignore, however in the future we may
		 * want to allow datardy to share the line */
		ret = iio_add_event_to_list(this_attr->listel,
					    &indio_dev->interrupts[1]->ev_list);
		if (ret)
			goto error_ret;
	}
	ret = __adis16350_enable_alarm(dev, location, st, mod);
	if (ret)
		goto error_ret;
done:
	return len;

error_ret:
	return ret;
}
/*As these are all similar in form, a macro shortens the code condsiderably!*/
#define ADIS16350_EV_ATTR(_name, _index)				\
	IIO_EVENT_ATTR_##_name##_SH(iio_event_alarms,			\
					  adis16350_alarm_show,		\
					  adis16350_alarm_store,	\
					  _index)

ADIS16350_EV_ATTR(ACCEL_X_HIGH, adis16350_alm_accel_x_gt_static);
ADIS16350_EV_ATTR(ACCEL_X_LOW, adis16350_alm_accel_x_lt_static);
ADIS16350_EV_ATTR(ACCEL_Y_HIGH, adis16350_alm_accel_y_gt_static);
ADIS16350_EV_ATTR(ACCEL_Y_LOW, adis16350_alm_accel_y_lt_static);
ADIS16350_EV_ATTR(ACCEL_Z_HIGH, adis16350_alm_accel_z_gt_static);
ADIS16350_EV_ATTR(ACCEL_Z_LOW, adis16350_alm_accel_z_lt_static);

ADIS16350_EV_ATTR(ACCEL_X_ROC_HIGH, adis16350_alm_accel_x_gt_roc);
ADIS16350_EV_ATTR(ACCEL_X_ROC_LOW, adis16350_alm_accel_x_lt_roc);
ADIS16350_EV_ATTR(ACCEL_Y_ROC_HIGH, adis16350_alm_accel_y_gt_roc);
ADIS16350_EV_ATTR(ACCEL_Y_ROC_LOW, adis16350_alm_accel_y_lt_roc);
ADIS16350_EV_ATTR(ACCEL_Z_ROC_HIGH, adis16350_alm_accel_z_gt_roc);
ADIS16350_EV_ATTR(ACCEL_Z_ROC_LOW, adis16350_alm_accel_z_lt_roc);

ADIS16350_EV_ATTR(GYRO_X_HIGH, adis16350_alm_gyro_x_gt_static);
ADIS16350_EV_ATTR(GYRO_X_LOW, adis16350_alm_gyro_x_lt_static);
ADIS16350_EV_ATTR(GYRO_Y_HIGH, adis16350_alm_gyro_y_gt_static);
ADIS16350_EV_ATTR(GYRO_Y_LOW, adis16350_alm_gyro_y_lt_static);
ADIS16350_EV_ATTR(GYRO_Z_HIGH, adis16350_alm_gyro_z_gt_static);
ADIS16350_EV_ATTR(GYRO_Z_LOW, adis16350_alm_gyro_z_lt_static);

ADIS16350_EV_ATTR(GYRO_X_ROC_HIGH, adis16350_alm_accel_x_gt_roc);
ADIS16350_EV_ATTR(GYRO_X_ROC_LOW, adis16350_alm_accel_x_lt_roc);
ADIS16350_EV_ATTR(GYRO_Y_ROC_HIGH, adis16350_alm_accel_y_gt_roc);
ADIS16350_EV_ATTR(GYRO_Y_ROC_LOW, adis16350_alm_accel_y_lt_roc);
ADIS16350_EV_ATTR(GYRO_Z_ROC_HIGH, adis16350_alm_accel_z_gt_roc);
ADIS16350_EV_ATTR(GYRO_Z_ROC_LOW, adis16350_alm_accel_z_lt_roc);

static struct attribute *adis16350_event_attributes_1[] = {
	&iio_event_attr_data_rdy.dev_attr.attr,
	NULL,
};

static struct attribute *adis16350_event_attributes_2[] = {
	&iio_event_attr_accel_x_high.dev_attr.attr,
	&iio_event_attr_accel_x_low.dev_attr.attr,
	&iio_event_attr_accel_y_high.dev_attr.attr,
	&iio_event_attr_accel_y_low.dev_attr.attr,
	&iio_event_attr_accel_z_high.dev_attr.attr,
	&iio_event_attr_accel_z_low.dev_attr.attr,

	&iio_event_attr_gyro_x_high.dev_attr.attr,
	&iio_event_attr_gyro_x_low.dev_attr.attr,
	&iio_event_attr_gyro_y_high.dev_attr.attr,
	&iio_event_attr_gyro_y_low.dev_attr.attr,
	&iio_event_attr_gyro_z_high.dev_attr.attr,
	&iio_event_attr_gyro_z_low.dev_attr.attr,

	&iio_event_attr_accel_x_roc_high.dev_attr.attr,
	&iio_event_attr_accel_x_roc_low.dev_attr.attr,
	&iio_event_attr_accel_y_roc_high.dev_attr.attr,
	&iio_event_attr_accel_y_roc_low.dev_attr.attr,
	&iio_event_attr_accel_z_roc_high.dev_attr.attr,
	&iio_event_attr_accel_z_roc_low.dev_attr.attr,

	&iio_event_attr_gyro_x_roc_high.dev_attr.attr,
	&iio_event_attr_gyro_x_roc_low.dev_attr.attr,
	&iio_event_attr_gyro_y_roc_high.dev_attr.attr,
	&iio_event_attr_gyro_y_roc_low.dev_attr.attr,
	&iio_event_attr_gyro_z_roc_high.dev_attr.attr,
	&iio_event_attr_gyro_z_roc_low.dev_attr.attr,

	NULL,
};

static ssize_t adis16350_show_alm_samp(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;

	return sprintf(buf, "%u\n", st->alarm_samp[this_attr->address]);
}

static ssize_t adis16350_store_alm_samp(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf,
				       size_t len)
{

	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	unsigned long val;
	int ret;
	ret = strict_strtoul(buf, 10, &val);
	if (ret)
		return ret;
	st->alarm_samp[this_attr->address] = val;
	return len;
}

static ssize_t adis16350_show_alm_mag(struct device *dev,
				      struct device_attribute *attr,
				      char *buf)
{
	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;

	return sprintf(buf, "%d\n", st->alarm_val[this_attr->address]);
}


static ssize_t adis16350_store_alm_mag(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf,
				       size_t len)
{

	struct iio_dev_attr *this_attr = to_iio_dev_attr(attr);
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct adis16350_state *st = indio_dev->dev_data;
	long val;
	int ret;
	ret = strict_strtol(buf, 10, &val);
	if (ret)
		return ret;
	st->alarm_val[this_attr->address] = val;

	return len;
}
#define ADIS16350_ALARM_VAL(_name, _count)				\
	IIO_DEVICE_ATTR(_name##_val, S_IRUGO | S_IWUSR,			\
			adis16350_show_alm_mag,				\
			adis16350_store_alm_mag, _count);

#define ADIS16350_ALARM_SAMP(_name, _count)				\
	IIO_DEVICE_ATTR(_name##_samp, S_IRUGO | S_IWUSR,		\
			adis16350_show_alm_samp,			\
			adis16350_store_alm_samp, _count);

/*TODO lots more to add here - temp, adc and power supply */
ADIS16350_ALARM_VAL(accel_x_high,	adis16350_alm_accel_x_gt_static);
ADIS16350_ALARM_VAL(accel_x_low,	adis16350_alm_accel_x_lt_static);
ADIS16350_ALARM_VAL(accel_y_high,	adis16350_alm_accel_y_gt_static);
ADIS16350_ALARM_VAL(accel_y_low,	adis16350_alm_accel_y_lt_static);
ADIS16350_ALARM_VAL(accel_z_high,	adis16350_alm_accel_z_gt_static);
ADIS16350_ALARM_VAL(accel_z_low,	adis16350_alm_accel_z_lt_static);

ADIS16350_ALARM_VAL(gyro_x_high,	adis16350_alm_gyro_x_gt_static);
ADIS16350_ALARM_VAL(gyro_x_low,		adis16350_alm_gyro_x_lt_static);
ADIS16350_ALARM_VAL(gyro_y_high,	adis16350_alm_gyro_y_gt_static);
ADIS16350_ALARM_VAL(gyro_y_low,		adis16350_alm_gyro_y_lt_static);
ADIS16350_ALARM_VAL(gyro_z_high,	adis16350_alm_gyro_z_gt_static);
ADIS16350_ALARM_VAL(gyro_z_low,		adis16350_alm_gyro_z_lt_static);

ADIS16350_ALARM_VAL(accel_x_roc_high,	adis16350_alm_accel_x_gt_roc);
ADIS16350_ALARM_VAL(accel_x_roc_low,	adis16350_alm_accel_x_lt_roc);
ADIS16350_ALARM_VAL(accel_y_roc_high,	adis16350_alm_accel_y_gt_roc);
ADIS16350_ALARM_VAL(accel_y_roc_low,	adis16350_alm_accel_y_lt_roc);
ADIS16350_ALARM_VAL(accel_z_roc_high,	adis16350_alm_accel_z_gt_roc);
ADIS16350_ALARM_VAL(accel_z_roc_low,	adis16350_alm_accel_z_lt_roc);

ADIS16350_ALARM_VAL(gyro_x_roc_high,	adis16350_alm_gyro_x_gt_roc);
ADIS16350_ALARM_VAL(gyro_x_roc_low,	adis16350_alm_gyro_x_lt_roc);
ADIS16350_ALARM_VAL(gyro_y_roc_high,	adis16350_alm_gyro_y_gt_roc);
ADIS16350_ALARM_VAL(gyro_y_roc_low,	adis16350_alm_gyro_y_lt_roc);
ADIS16350_ALARM_VAL(gyro_z_roc_high,	adis16350_alm_gyro_z_gt_roc);
ADIS16350_ALARM_VAL(gyro_z_roc_low,	adis16350_alm_gyro_z_lt_roc);


ADIS16350_ALARM_SAMP(accel_x_high,	adis16350_alm_accel_x_gt_static);
ADIS16350_ALARM_SAMP(accel_x_low,	adis16350_alm_accel_x_lt_static);
ADIS16350_ALARM_SAMP(accel_y_high,	adis16350_alm_accel_y_gt_static);
ADIS16350_ALARM_SAMP(accel_y_low,	adis16350_alm_accel_y_lt_static);
ADIS16350_ALARM_SAMP(accel_z_high,	adis16350_alm_accel_y_gt_static);
ADIS16350_ALARM_SAMP(accel_z_low,	adis16350_alm_accel_y_lt_static);

ADIS16350_ALARM_SAMP(gyro_x_high,	adis16350_alm_gyro_x_gt_static);
ADIS16350_ALARM_SAMP(gyro_x_low,	adis16350_alm_gyro_x_lt_static);
ADIS16350_ALARM_SAMP(gyro_y_high,	adis16350_alm_gyro_y_gt_static);
ADIS16350_ALARM_SAMP(gyro_y_low,	adis16350_alm_gyro_y_lt_static);
ADIS16350_ALARM_SAMP(gyro_z_high,	adis16350_alm_gyro_z_gt_static);
ADIS16350_ALARM_SAMP(gyro_z_low,	adis16350_alm_gyro_z_lt_static);

ADIS16350_ALARM_SAMP(accel_x_roc_high,	adis16350_alm_accel_x_gt_static);
ADIS16350_ALARM_SAMP(accel_x_roc_low,	adis16350_alm_accel_x_lt_static);
ADIS16350_ALARM_SAMP(accel_y_roc_high,	adis16350_alm_accel_y_gt_static);
ADIS16350_ALARM_SAMP(accel_y_roc_low,	adis16350_alm_accel_y_lt_static);
ADIS16350_ALARM_SAMP(accel_z_roc_high,	adis16350_alm_accel_y_gt_static);
ADIS16350_ALARM_SAMP(accel_z_roc_low,	adis16350_alm_accel_y_lt_static);

ADIS16350_ALARM_SAMP(gyro_x_roc_high,	adis16350_alm_gyro_x_gt_roc);
ADIS16350_ALARM_SAMP(gyro_x_roc_low,	adis16350_alm_gyro_x_lt_roc);
ADIS16350_ALARM_SAMP(gyro_y_roc_high,	adis16350_alm_gyro_y_gt_roc);
ADIS16350_ALARM_SAMP(gyro_y_roc_low,	adis16350_alm_gyro_y_lt_roc);
ADIS16350_ALARM_SAMP(gyro_z_roc_high,	adis16350_alm_gyro_z_gt_roc);
ADIS16350_ALARM_SAMP(gyro_z_roc_low,	adis16350_alm_gyro_z_lt_roc);

static struct attribute *adis16350_event_config_attributes_2[] = {
	&iio_dev_attr_accel_x_high_val.dev_attr.attr,
	&iio_dev_attr_accel_x_low_val.dev_attr.attr,
	&iio_dev_attr_accel_y_high_val.dev_attr.attr,
	&iio_dev_attr_accel_y_low_val.dev_attr.attr,
	&iio_dev_attr_accel_z_high_val.dev_attr.attr,
	&iio_dev_attr_accel_z_low_val.dev_attr.attr,

	&iio_dev_attr_gyro_x_high_val.dev_attr.attr,
	&iio_dev_attr_gyro_x_low_val.dev_attr.attr,
	&iio_dev_attr_gyro_y_high_val.dev_attr.attr,
	&iio_dev_attr_gyro_y_low_val.dev_attr.attr,
	&iio_dev_attr_gyro_z_high_val.dev_attr.attr,
	&iio_dev_attr_gyro_z_low_val.dev_attr.attr,

	&iio_dev_attr_accel_x_roc_high_val.dev_attr.attr,
	&iio_dev_attr_accel_x_roc_low_val.dev_attr.attr,
	&iio_dev_attr_accel_y_roc_high_val.dev_attr.attr,
	&iio_dev_attr_accel_y_roc_low_val.dev_attr.attr,
	&iio_dev_attr_accel_z_roc_high_val.dev_attr.attr,
	&iio_dev_attr_accel_z_roc_low_val.dev_attr.attr,

	&iio_dev_attr_gyro_x_roc_high_val.dev_attr.attr,
	&iio_dev_attr_gyro_x_roc_low_val.dev_attr.attr,
	&iio_dev_attr_gyro_y_roc_high_val.dev_attr.attr,
	&iio_dev_attr_gyro_y_roc_low_val.dev_attr.attr,
	&iio_dev_attr_gyro_z_roc_high_val.dev_attr.attr,
	&iio_dev_attr_gyro_z_roc_low_val.dev_attr.attr,

	&iio_dev_attr_accel_x_high_samp.dev_attr.attr,
	&iio_dev_attr_accel_x_low_samp.dev_attr.attr,
	&iio_dev_attr_accel_y_high_samp.dev_attr.attr,
	&iio_dev_attr_accel_y_low_samp.dev_attr.attr,
	&iio_dev_attr_accel_z_high_samp.dev_attr.attr,
	&iio_dev_attr_accel_z_low_samp.dev_attr.attr,

	&iio_dev_attr_gyro_x_high_samp.dev_attr.attr,
	&iio_dev_attr_gyro_x_low_samp.dev_attr.attr,
	&iio_dev_attr_gyro_y_high_samp.dev_attr.attr,
	&iio_dev_attr_gyro_y_low_samp.dev_attr.attr,
	&iio_dev_attr_gyro_z_high_samp.dev_attr.attr,
	&iio_dev_attr_gyro_z_low_samp.dev_attr.attr,

	&iio_dev_attr_accel_x_roc_high_samp.dev_attr.attr,
	&iio_dev_attr_accel_x_roc_low_samp.dev_attr.attr,
	&iio_dev_attr_accel_y_roc_high_samp.dev_attr.attr,
	&iio_dev_attr_accel_y_roc_low_samp.dev_attr.attr,
	&iio_dev_attr_accel_z_roc_high_samp.dev_attr.attr,
	&iio_dev_attr_accel_z_roc_low_samp.dev_attr.attr,

	&iio_dev_attr_gyro_x_roc_high_samp.dev_attr.attr,
	&iio_dev_attr_gyro_x_roc_low_samp.dev_attr.attr,
	&iio_dev_attr_gyro_y_roc_high_samp.dev_attr.attr,
	&iio_dev_attr_gyro_y_roc_low_samp.dev_attr.attr,
	&iio_dev_attr_gyro_z_roc_high_samp.dev_attr.attr,
	&iio_dev_attr_gyro_z_roc_low_samp.dev_attr.attr,
	NULL,
};

static struct attribute_group adis16350_event_config_attribute_groups[] = {{
	}, {
		.attrs = adis16350_event_config_attributes_2,
	},
};

static struct attribute_group adis16350_event_attribute_groups [] = {{
		.attrs = adis16350_event_attributes_1,
	}, {
		.attrs = adis16350_event_attributes_2,
	},
};

static int adis16350_ring_preenable(struct iio_dev *indio_dev)
{
	struct adis16350_state *st = indio_dev->dev_data;
	int bpd ;
	if(indio_dev->ring_access.set_bpd) {
		bpd = sizeof(s64) + indio_dev->scan_count * sizeof(u16);
		indio_dev->ring_access.set_bpd(indio_dev->ring, bpd);
	}
	return __adis16350_write_data_ready_config(&st->spi->dev,
						   &iio_event_data_rdy,
						   0);
}
IIO_EVENT_SH(sw_ring_enable, &adis16350_data_rdy_ring_th);
static int adis16350_ring_postenable(struct iio_dev *indio_dev)
{
	struct adis16350_state *st = indio_dev->dev_data;
	return __adis16350_write_data_ready_config(&st->spi->dev,
						   &iio_event_sw_ring_enable,
						   1);
}

static int adis16350_ring_predisable(struct iio_dev *indio_dev)
{
	struct adis16350_state *st = indio_dev->dev_data;
	return __adis16350_write_data_ready_config(&st->spi->dev,
						   &iio_event_sw_ring_enable,
						   0);
}


static int adis16350_initialize_device(struct adis16350_state *st)
{
	int ret;
	u16 val;
	/* Find out if we can operate in fast mode?*/
	/* N.B. this must occur after all registrations */
	if (st->spi->max_speed_hz > 200000)
		st->max_speed_hz = 200000;
	ret = __adis16350_read_reg_16(&st->spi->dev,
				      ADIS16350_SAMP_PERIOD_LSB_ADDR,
				      &val);
	if (ret)
		goto error_ret;

	st->max_speed_hz = ( val > 0x09) && (st->spi->max_speed_hz > 2000000)
		? 2000000 : 0;

	return 0;

error_ret:
	return ret;
}

static int adis16350_stop_device(struct adis16350_state *st)
{
	int ret;
	u16 val;
	const u16 drmask = ADIS16350_MSC_CTRL_LSB_DATA_RDY_ENABLE;
	/* Stop interrupts being sent */
	/* Data rdy */

	ret = __adis16350_read_reg_16(&st->spi->dev,
				      ADIS16350_MSC_CTRL_LSB_ADDR, &val);
	if (ret)
		return ret;
	if (val & drmask)
		ret = __adis16350_write_reg_8(&st->spi->dev,
					      ADIS16350_MSC_CTRL_LSB_ADDR,
					      val & ~drmask);
	if (ret)
		return ret;
	/* Alarms */
	/* Set each alarm to disabled rather than disabling alarms altogether */
	if (st->current_alms[0] || st->current_alms[1])
		ret = __adis16350_write_reg_8(&st->spi->dev,
					      ADIS16350_ALM_CTRL_MSB_ADDR,
					      0);
	return ret;
}
static int __devinit adis16350_probe(struct spi_device *spi)
{
	int ret;
	struct adis16350_state *st;

	st = kzalloc(sizeof(struct adis16350_state), GFP_KERNEL);
	if (st == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}

	st->pdata = spi->dev.platform_data;
	st->indio_dev = kzalloc(sizeof(struct iio_dev), GFP_KERNEL);
	if (st->indio_dev == NULL) {
		ret = -ENOMEM;
		goto error_free_device_state;
	}

	st->spi = spi;
	st->spi->mode = SPI_MODE_3;
	spi_setup(st->spi);

	st->indio_dev->dev = &spi->dev;
	st->indio_dev->num_interrupt_lines = 2;
	/* initially turn on accels and gyros but not temp or adc etc */
	st->indio_dev->scan_mask
		= iio_scan_el_accel_x.mask | iio_scan_el_accel_y.mask
		| iio_scan_el_accel_z.mask | iio_scan_el_gyro_x.mask
		| iio_scan_el_gyro_y.mask | iio_scan_el_gyro_z.mask;
	st->indio_dev->scan_count = 6;
	st->indio_dev->scan_el_attrs = &adis16350_scan_el_group;
	st->indio_dev->event_attrs = adis16350_event_attribute_groups;
	st->indio_dev->event_conf_attrs
		= adis16350_event_config_attribute_groups;
	st->indio_dev->attrs = &adis16350_attribute_group;
	st->indio_dev->dev_data = (void *)(st);
	st->indio_dev->modes = INDIO_DIRECT_MODE | INDIO_RING_TRIGGERED;

	st->indio_dev->ring_preenable = &adis16350_ring_preenable;
	st->indio_dev->ring_postenable = &adis16350_ring_postenable;
	st->indio_dev->ring_predisable = &adis16350_ring_predisable;

	ret = iio_device_register(st->indio_dev);
	if (ret)
		goto error_free_indiodev;

	INIT_WORK(&st->work_data_rdy_event, adis16350_data_rdy_bh_to_event);
	INIT_WORK(&st->work_data_rdy_ring, adis16350_data_rdy_bh_to_ring);

	iio_init_work_cont(&st->work_alarm_event, adis16350_alarm_bh_check,
			   adis16350_alarm_bh_no_check,
			   ADIS16350_STATUS_LSB_ADDR, 0, st);

	ret = iio_register_interrupt_line(st->pdata->irq_datardy,
					  st->indio_dev,
					  0,
					  st->pdata->irq_datardy_type,
					  "adis16350datardy");
	if (ret)
		goto error_unregister_device;
	ret = iio_register_interrupt_line(st->pdata->irq_event,
					  st->indio_dev,
					  1,
					  st->pdata->irq_event_type,
					  "adis16350ev");
	if (ret)
		goto error_unregister_datardy_line;
	ret = adis16350_initialize_device(st);
	if (ret)
		goto error_unregister_event_line;;

	if (ret)
		goto error_unregister_event_line;
	return 0;
error_unregister_event_line:
	iio_unregister_interrupt_line(st->indio_dev,1);
error_unregister_datardy_line:
	iio_unregister_interrupt_line(st->indio_dev,0);
error_unregister_device:
	iio_device_unregister(st->indio_dev);
error_free_indiodev:
	kfree(st->indio_dev);
error_free_device_state:
	kfree(st);
error_ret:
	return ret;
}

static int adis16350_remove(struct spi_device *spi)
{
	struct iio_dev *indio_dev = spi_get_drvdata(spi);
	struct adis16350_state *st = indio_dev->dev_data;

	adis16350_stop_device(st);
	flush_scheduled_work();

	iio_unregister_interrupt_line(st->indio_dev,1);
	iio_unregister_interrupt_line(st->indio_dev,0);
	iio_device_unregister(indio_dev);
	kfree(st->indio_dev);
	kfree(st);

	return 0;
}

static struct spi_driver adis16350_driver = {
	.driver = {
		.name = "adis16350",
		.owner = THIS_MODULE,
	},
	.probe = adis16350_probe,
	.remove = __devexit_p(adis16350_remove),
};

static __init int adis16350_init(void)
{
	return spi_register_driver(&adis16350_driver);
}

static __exit void adis16350_exit(void)
{
	spi_unregister_driver(&adis16350_driver);
}

module_init(adis16350_init);
module_exit(adis16350_exit);

MODULE_AUTHOR("Jonathan Cameron <jic23@cam.ac.uk>");
MODULE_DESCRIPTION("Analog Devices ADIS16350 IMU driver");
MODULE_LICENSE("GPL v2");
