/* The industrial I/O core
 *
 * Copyright (c) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Based on elements of hwmon and input subsystems.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/idr.h>
#include <linux/kdev_t.h>
#include <linux/err.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/poll.h>

#include <linux/industrialio/iio.h>
#include <linux/industrialio/trigger.h>


#define IIO_ID_PREFIX "iio_device"
#define IIO_ID_FORMAT IIO_ID_PREFIX "%d"

/* Temporary - I'll request one when we nearer submission*/

#define IIO_MAJOR 244

/* Integer id  - used to assign each registered device a unique id*/
static DEFINE_IDR(iio_idr);
static DEFINE_SPINLOCK(iio_idr_lock);

struct class iio_class = {
	.name = "industrialio",
};

/**
 * struct __iio_state - internal iio subsystem state information.
 * @fhs:	file operations for the various chrdevs. These are
 *		all intially set to null in init.
 **/
struct __iio_state {
	struct iio_handler *fhs[256];
};

static struct __iio_state iio_state;
static DEFINE_SPINLOCK(iio_state_lock);

ssize_t iio_scan_el_show(struct device *dev,
			 struct device_attribute *attr,
			 char *buf)
{
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct iio_scan_el *this_el = to_iio_scan_el(attr);

	return sprintf(buf, "%d\n",
		       !!(indio_dev->scan_mask & this_el->mask));
}
EXPORT_SYMBOL(iio_scan_el_show);

ssize_t iio_scan_el_store(struct device *dev,
				 struct device_attribute *attr,
				 const char *buf,
				 size_t len)
{
	int ret = 0;
	bool state;
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct iio_scan_el *this_el = to_iio_scan_el(attr);

	state = !(buf[0] == '0');
	mutex_lock(&indio_dev->mlock);
	if (indio_dev->currentmode == INDIO_RING_TRIGGERED) {
		ret = -EBUSY;
		goto error_ret;
	}

	if (!state && (indio_dev->scan_mask & this_el->mask)) {
		indio_dev->scan_mask &= ~this_el->mask;
		indio_dev->scan_count--;
	} else if (state && !(indio_dev->scan_mask & this_el->mask)) {
		indio_dev->scan_mask |= this_el->mask;
		indio_dev->scan_count++;
	}
	if (this_el->set_state)
		ret = this_el->set_state(this_el, indio_dev, state);
	mutex_unlock(&indio_dev->mlock);
error_ret:
	if(ret)
		return ret;
	return len;

}
EXPORT_SYMBOL(iio_scan_el_store);

/* TODO Check if these need locking */
void __iio_change_event(struct iio_detected_event_list *ev,
			int ev_code,
			s64 timestamp)
{
	ev->ev.id = ev_code;
	ev->ev.timestamp = timestamp;
}
EXPORT_SYMBOL(__iio_change_event);

/* Used both in the interrupt line put events and the ring buffer ones */

/* Note that in it's current form someone has to be listening before events
 * are queued. Hence a client MUST open the chrdev before the ring buffer is
 * switched on.
 */
 int __iio_push_event(struct iio_event_interface *ev_int,
		     int ev_code,
		     s64 timestamp,
		     struct iio_shared_ev_pointer*
		     shared_pointer_p)
{
	struct iio_detected_event_list *ev;
	int ret = 0;

	/* Does anyone care? */
	if (test_bit(IIO_BUSY_BIT_POS, &ev_int->handler.flags)) {
		if (ev_int->current_events == ev_int->max_events)
			return 0;
		ev = kmalloc(sizeof(struct iio_detected_event_list),
			     GFP_KERNEL);
		if (ev == NULL) {
			ret = -ENOMEM;
			goto error_ret;
		}
		ev->ev.id = ev_code;
		ev->ev.timestamp = timestamp;
		if (shared_pointer_p != NULL) {
			ev->shared_pointer = shared_pointer_p;
			shared_pointer_p->ev_p = ev;
		} else
			ev->shared_pointer = NULL;

		mutex_lock(&ev_int->event_list_lock);
		list_add_tail(&ev->list, &ev_int->det_events.list);
		mutex_unlock(&ev_int->event_list_lock);

		ev_int->current_events++;
		wake_up_interruptible(&ev_int->wait);
	}

error_ret:
	return ret;
}
EXPORT_SYMBOL(__iio_push_event);

int iio_push_event(struct iio_dev *dev_info,
		   int ev_line,
		   int ev_code,
		   s64 timestamp)
{
	return __iio_push_event(&dev_info->event_interfaces[ev_line],
			       ev_code, timestamp, NULL);
}
EXPORT_SYMBOL(iio_push_event);

/* Generic interrupt line interrupt handler */
irqreturn_t iio_interrupt_handler(int irq, void *_int_info)
{
	struct iio_interrupt *int_info = _int_info;
	struct iio_dev *dev_info = int_info->dev_info;
	struct iio_event_handler_list *p;
	s64 time_ns;

	if (list_empty(&int_info->ev_list.list))
		return IRQ_NONE;
//	printk("interrupt\n");
	time_ns = iio_get_time_ns();
	/* detect single element list*/
	if (int_info->ev_list.list.next->next == &int_info->ev_list.list) {
		disable_irq_nosync(irq);
		p = list_first_entry(&int_info->ev_list.list,
				     struct iio_event_handler_list,
				     list);
		/* single event handler - maybe shared */
		p->handler(dev_info, 1, time_ns, !(p->refcount > 1));
	} else
		list_for_each_entry(p, &int_info->ev_list.list, list) {
			disable_irq_nosync(irq);
			p->handler(dev_info, 1, time_ns, 0);
		}
	return IRQ_HANDLED;
}

/* Confirming the validity of supplied irq is left to drivers.*/
int iio_register_interrupt_line(unsigned int irq,
				struct iio_dev *dev_info,
				int line_number,
				unsigned long type,
				const char *name)
{
	int ret;

	dev_info->interrupts[line_number] =
		kmalloc(sizeof(struct iio_interrupt), GFP_KERNEL);
	if (dev_info->interrupts[line_number] == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}
	printk(KERN_INFO  "initializing interrupt line %d\n", line_number);
	INIT_LIST_HEAD(&dev_info->interrupts[line_number]->ev_list.list);
	dev_info->interrupts[line_number]->line_number = line_number;
	dev_info->interrupts[line_number]->irq = irq;
	dev_info->interrupts[line_number]->dev_info = dev_info;

	/* Possibly only request on demand?
	 * Can see this may complicate the handling of interrupts.
	 * However, with this approach we might end up handling lots of
	 * events no-one cares about.*/
	ret = request_irq(irq,
			  &iio_interrupt_handler,
			  type,
			  name,
			  dev_info->interrupts[line_number]);
	if (ret < 0)
		goto error_ret;

	return 0;

error_ret:
	return ret;
}
EXPORT_SYMBOL(iio_register_interrupt_line);

/* This turns up an awful lot */
ssize_t iio_read_const_attr(struct device *dev,
			    struct device_attribute *attr,
			    char *buf)
{
	return sprintf(buf, "%s\n", to_iio_const_attr(attr)->string);
}
EXPORT_SYMBOL(iio_read_const_attr);

/* Before this runs the interrupt generator must have been disabled */
void iio_unregister_interrupt_line(struct iio_dev *dev_info,
					    int line_number)
{
	/* make sure the interrupt handlers are all done */
	flush_scheduled_work();
	free_irq(dev_info->interrupts[line_number]->irq,
		 dev_info->interrupts[line_number]);
	kfree(dev_info->interrupts[line_number]);
}
EXPORT_SYMBOL(iio_unregister_interrupt_line);

/* Reference counted add and remove */
int iio_add_event_to_list(struct iio_event_handler_list *el,
			  struct iio_event_handler_list *list)
{
	if (el->refcount == 0)
		list_add(&el->list, &list->list);
	el->refcount++;
	return 0;
}
EXPORT_SYMBOL(iio_add_event_to_list);

int iio_remove_event_from_list(struct iio_event_handler_list *el)
{
	el->refcount--;
	if (el->refcount == 0)
		list_del_init(&el->list);
	return 0;
}
EXPORT_SYMBOL(iio_remove_event_from_list);

int iio_allocate_chrdev(struct iio_handler *handler)
{
	int id;

	spin_lock(&iio_state_lock);
	for (id = 0; id <= 256; id++)
		if (iio_state.fhs[id] == NULL)
			break;
	if (id == 256) {
		spin_unlock(&iio_state_lock);
		return -ENOMEM;
	}
	iio_state.fhs[id] = handler;
	spin_unlock(&iio_state_lock);
	handler->id = id;

	return 0;
}

void iio_deallocate_chrdev(struct iio_handler *handler)
{
	spin_lock(&iio_state_lock);
	iio_state.fhs[handler->id] = NULL;
	spin_unlock(&iio_state_lock);
}

/* Upon open, switch in the correct file ops.
 * Lifted directly from input subsystem (more or less)
 */
static int iio_open_file(struct inode *inode, struct file *file)
{
	struct iio_handler *handler;
	const struct file_operations *old_fops, *new_fops = NULL;
	int err;

	/* This lock needed as  we are dynamically allocating chrdevs */
	spin_lock(&iio_state_lock);
	handler = iio_state.fhs[iminor(inode)];
	spin_unlock(&iio_state_lock);

	if (!handler) {
		fops_put(file->f_op);
		return -ENODEV;
	}
	new_fops = fops_get(handler->fops);
	if (new_fops == NULL) {
		fops_put(file->f_op);
		return -ENODEV;
	}

	/* cribbed from lp.c */
	if (test_and_set_bit(IIO_BUSY_BIT_POS, &handler->flags)) {
		fops_put(file->f_op);
		return -EBUSY;
	}

	if (!new_fops->open) {
		fops_put(new_fops);
		return -ENODEV;
	}
	old_fops = file->f_op;
	file->f_op = new_fops;
	/* use the private data pointer in file to give access to device
	 * specific stuff */
	file->private_data = handler->private;
	err = new_fops->open(inode, file);

	if (err) {
		fops_put(file->f_op);
		file->f_op = fops_get(old_fops);
	}
	fops_put(old_fops);

	return err;
}

static const struct file_operations iio_fops = {
	.owner = THIS_MODULE,
	.open = iio_open_file,
};

/* File ops for event character device files */
ssize_t iio_event_chrdev_read(struct file *filep,
			      char *buf,
			      size_t count,
			      loff_t *f_ps)
{
	struct iio_event_interface *ev_int = filep->private_data;
	struct iio_detected_event_list *el;
	int ret;

	mutex_lock(&ev_int->event_list_lock);
	if (list_empty(&ev_int->det_events.list)) {
		if (filep->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			goto error_mutex_unlock;
		}
		mutex_unlock(&ev_int->event_list_lock);
		/* Blocking on device; waiting for something to be there */
		ret = wait_event_interruptible(ev_int->wait,
					       !list_empty(&ev_int
							   ->det_events.list));
		if (ret)
			goto error_ret;
		/* Single access device so noone else can get the data */
		mutex_lock(&ev_int->event_list_lock);
	}

	el = list_first_entry(&ev_int->det_events.list,
			      struct iio_detected_event_list,
			      list);

	if (copy_to_user(buf, &(el->ev),
			 sizeof(struct iio_event_data))) {
		ret = -EFAULT;
		goto error_mutex_unlock;
	}

	list_del(&el->list);
	ev_int->current_events--;
	mutex_unlock(&ev_int->event_list_lock);

	spin_lock(&el->shared_pointer->lock);
	if (el->shared_pointer)
		(el->shared_pointer->ev_p) = NULL;
	spin_unlock(&el->shared_pointer->lock);

	kfree(el);

	return sizeof(struct iio_event_data);

error_mutex_unlock:
	mutex_unlock(&ev_int->event_list_lock);
error_ret:

	return ret;
}

int iio_event_chrdev_release(struct inode *inode, struct file *filep)
{
	struct iio_event_interface *ev_int = filep->private_data;

	module_put(ev_int->owner);
	clear_bit(IIO_BUSY_BIT_POS, &ev_int->handler.flags);

	return 0;
}

int iio_event_chrdev_open(struct inode *inode, struct file *filep)
{
	struct iio_event_interface *ev_int = filep->private_data;

	try_module_get(ev_int->owner);

	return 0;
}

static const struct file_operations iio_event_chrdev_fileops = {
	.read =  iio_event_chrdev_read,
	.release = iio_event_chrdev_release,
	.open = iio_event_chrdev_open,
	.owner = THIS_MODULE,
};

ssize_t iio_show_attr_minor(struct device *dev,
			    struct device_attribute *attr,
			    char *buf)
{
	struct iio_chrdev_minor_attr *_attr
		= to_iio_chrdev_minor_attr(attr);
	return sprintf(buf, "%d\n", _attr->minor);
}

void __init_iio_chrdev_minor_attr(struct iio_chrdev_minor_attr *minor_attr,
				const char *name,
				struct module *owner,
				int id)
{
	minor_attr->dev_attr.attr.name = name;
	minor_attr->dev_attr.attr.owner = owner;
	minor_attr->dev_attr.attr.mode = S_IRUGO;
	minor_attr->minor = id;
	minor_attr->dev_attr.show = &iio_show_attr_minor;
}

int iio_setup_ev_int(struct iio_event_interface *ev_int,
		     const char *name,
		     struct module *owner,
		     struct device *dev)
{
	int ret;

	mutex_init(&ev_int->event_list_lock);
	/* discussion point - make this variable? */
	ev_int->max_events = 10;
	ev_int->current_events = 0;
	INIT_LIST_HEAD(&ev_int->det_events.list);
	init_waitqueue_head(&ev_int->wait);
	ev_int->handler.fops = &iio_event_chrdev_fileops;
	ev_int->handler.private = ev_int;
	ev_int->handler.flags = 0;
	ret = iio_allocate_chrdev(&ev_int->handler);
	if (ret)
		goto error_ret;
	__init_iio_chrdev_minor_attr(&ev_int->attr,
				     (const char *)(name),
				     owner,
				     ev_int->handler.id);
	ret = sysfs_create_file(&dev->kobj, &ev_int->attr.dev_attr.attr);
	if (ret)
		goto error_deallocate_chrdev;

	return 0;
error_deallocate_chrdev:
	iio_deallocate_chrdev(&ev_int->handler);
error_ret:
	return ret;
}

void iio_free_ev_int(struct iio_event_interface *ev_int, struct device *dev)
{
	sysfs_remove_file(&dev->kobj, &ev_int->attr.dev_attr.attr);
	iio_deallocate_chrdev(&ev_int->handler);
}

static int __init iio_init(void)
{
	int ret;

	memset(iio_state.fhs,
	       sizeof(struct iio_handler *)*256,
	       0);

	/* Create sysfs class */
	ret  = class_register(&iio_class);
	if (ret < 0) {
		printk(KERN_ERR
		       "industrialio.c: could not create sysfs class\n");
		goto error_nothing;
	}

	ret = register_chrdev(IIO_MAJOR, "industrialio", &iio_fops);
	if (ret) {
		printk(KERN_ERR
		       "industrialio: unable to register a char major %d",
		       IIO_MAJOR);
		goto error_unregister_class;
	}
	return 0;

error_unregister_class:
	class_unregister(&iio_class);
error_nothing:
	return ret;
}

static void __exit iio_exit(void)
{
	unregister_chrdev(IIO_MAJOR, "industrialio");
	class_unregister(&iio_class);
}

static int iio_device_register_sysfs(struct iio_dev *dev_info)
{
	int ret;

	dev_info->sysfs_dev = device_create(&iio_class,
					    dev_info->dev,
					    MKDEV(0, 0),
					    NULL,
					    IIO_ID_FORMAT,
					    dev_info->id);

	if (IS_ERR(dev_info->sysfs_dev)) {
		/* what would correct error here be?*/
		ret = -EINVAL;
		goto error_ret;
	}
	/* register attributes */
	ret = sysfs_create_group(&dev_info->dev->kobj, dev_info->attrs);
	if (ret) {
		dev_err(dev_info->dev, "Failed to register sysfs hooks\n");
		goto error_free_sysfs_device;
	}

	if (dev_info->scan_el_attrs)
		ret = sysfs_create_group(&dev_info->dev->kobj, dev_info->scan_el_attrs);

	return ret;

error_free_sysfs_device:
	device_unregister(dev_info->dev);
error_ret:
	return ret;
}

static void iio_device_unregister_sysfs(struct iio_dev *dev_info)
{
	if (dev_info->scan_el_attrs)
		sysfs_remove_group(&dev_info->dev->kobj, dev_info->scan_el_attrs);

	sysfs_remove_group(&dev_info->dev->kobj, dev_info->attrs);
	device_unregister(dev_info->sysfs_dev);
}

static int iio_device_register_id(struct iio_dev *dev_info)
{
	int ret;

idr_again:
	if (unlikely(idr_pre_get(&iio_idr, GFP_KERNEL) == 0))
		return -ENOMEM;

	spin_lock(&iio_idr_lock);
	ret = idr_get_new(&iio_idr, NULL, &dev_info->id);
	spin_unlock(&iio_idr_lock);
	if (unlikely(ret == -EAGAIN))
		goto idr_again;
	else if (unlikely(ret))
		return ret;
	dev_info->id = dev_info->id & MAX_ID_MASK;

	return 0;
}

static void iio_device_unregister_id(struct iio_dev *dev_info)
{
	/* Can I use the save id? */
	int id;

	if (likely(sscanf(dev_info->sysfs_dev->bus_id,
			  IIO_ID_FORMAT, &id) == 1)) {
		spin_lock(&iio_idr_lock);
		idr_remove(&iio_idr, id);
		spin_unlock(&iio_idr_lock);
	} else
		dev_dbg(dev_info->dev->parent,
			"indio_device_unregister() failed: bad class ID!\n");
}

static inline int __iio_add_event_config_attrs(struct iio_dev *dev_info, int i)
{
	int ret;
	/*p for adding, q for removing */
	struct attribute **attrp, **attrq;

	if (dev_info->event_conf_attrs && dev_info->event_conf_attrs[i].attrs) {
		attrp = dev_info->event_conf_attrs[i].attrs;
		while (*attrp) {
			ret =  sysfs_add_file_to_group(&dev_info->dev->kobj,
						       *attrp,
						       dev_info
						       ->event_attrs[i].name);
			if(ret)
				goto error_ret;
			attrp++;
		}
	}
	return 0;

error_ret:
	attrq = dev_info->event_conf_attrs[i].attrs;
	while(attrq != attrp) {
		sysfs_remove_file_from_group(&dev_info->dev->kobj,
					     *attrq,
					     dev_info->event_attrs[i].name);
		attrq++;
	}

	return ret;
}

static inline int __iio_remove_event_config_attrs(struct iio_dev *dev_info,
						  int i)
{
	struct attribute **attrq;

	if (dev_info->event_conf_attrs
		&& dev_info->event_conf_attrs[i].attrs) {
		attrq = dev_info->event_conf_attrs[i].attrs;
		while(*attrq) {
			sysfs_remove_file_from_group(&dev_info->dev->kobj,
						     *attrq,
						     dev_info
						     ->event_attrs[i].name);
			attrq++;
		}
	}

	return 0;
}

static int iio_device_register_eventset(struct iio_dev *dev_info)
{
	int ret, i, j;

	if (dev_info->num_interrupt_lines == 0)
		return 0;

	dev_info->event_interfaces = (struct iio_event_interface *)
		(kzalloc(sizeof(struct iio_event_interface)
			 *dev_info->num_interrupt_lines,
			 GFP_KERNEL));
	if (dev_info->event_interfaces == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}
	/* assign id's to the event_interface elements */
	for (i = 0; i < dev_info->num_interrupt_lines; i++) {
		dev_info->event_interfaces[i].id = i;
		dev_info->event_interfaces[i].owner = dev_info->driver_module;
	}
	dev_info->interrupts
		= kzalloc(sizeof(struct iio_interrupt *)
			  *dev_info->num_interrupt_lines,
			  GFP_KERNEL);
	if (dev_info->interrupts == NULL) {
		dev_err(dev_info->dev,
			"Failed to register sysfs hooks for events attributes");
		ret = -ENOMEM;
		goto error_free_event_interfaces;
	}

	for (i = 0; i < dev_info->num_interrupt_lines; i++) {

		snprintf(dev_info->event_interfaces[i]._name, 20,
			"event_line%d_minor", i);
		ret = iio_setup_ev_int(&dev_info->event_interfaces[i],
				       (const char *)(dev_info
						      ->event_interfaces[i]
						      ._name),
				       dev_info->driver_module,
				       dev_info->dev);
		if (ret) {
			dev_err(dev_info->dev,
				"Could not get chrdev interface\n");
			goto error_free_setup_ev_ints;
		}
	}
	for (i = 0; i < dev_info->num_interrupt_lines; i++) {
		snprintf(dev_info->event_interfaces[i]._attrname, 20,
			"event_line%d_sources", i);
		dev_info->event_attrs[i].name
			= (const char *)
			(dev_info->event_interfaces[i]._attrname);
		ret = sysfs_create_group(&dev_info->dev->kobj,
					 &dev_info->event_attrs[i]);
		if (ret) {
			dev_err(dev_info->dev,
				"Failed to register sysfs for event attrs");
			goto error_remove_sysfs_interfaces;
		}
	}
	for (i = 0; i < dev_info->num_interrupt_lines; i++) {
		ret = __iio_add_event_config_attrs(dev_info, i);
		if (ret)
			goto error_unregister_config_attrs;
	}



	return 0;

error_unregister_config_attrs:
	for (j = 0; j < i; j++)
		__iio_remove_event_config_attrs(dev_info, i);
	i = dev_info->num_interrupt_lines - 1;
error_remove_sysfs_interfaces:
	for (j = 0; j < i; j++)
		sysfs_remove_group(&dev_info->dev->kobj,
				   &dev_info->event_attrs[j]);

	i = dev_info->num_interrupt_lines - 1;
error_free_setup_ev_ints:
	for (j = 0; j < i; j++)
		iio_free_ev_int(&dev_info->event_interfaces[j],
				dev_info->dev);
	kfree(dev_info->interrupts);
error_free_event_interfaces:
	kfree(dev_info->event_interfaces);
error_ret:

	return ret;
}

static void iio_device_unregister_eventset(struct iio_dev *dev_info)
{
	int i;
	if (dev_info->num_interrupt_lines == 0)
		return;
	for( i = 0; i < dev_info->num_interrupt_lines; i++)
		sysfs_remove_group(&dev_info->dev->kobj,
				   &dev_info->event_attrs[i]);

	for (i = 0; i < dev_info->num_interrupt_lines; i++)
		iio_free_ev_int(&dev_info->event_interfaces[i],
					 dev_info->dev);
	kfree(dev_info->event_interfaces);
}


int iio_device_register(struct iio_dev *dev_info)
{
	int ret;
	mutex_init(&dev_info->mlock);
	dev_set_drvdata(dev_info->dev, (void *)(dev_info));

	printk("d %p %p\n", dev_get_drvdata(dev_info->dev), dev_info);

/*Get a unique id */
	ret = iio_device_register_id(dev_info);
	if (ret)
		goto error_nothing;

/* Create sysfs device */
	ret = iio_device_register_sysfs(dev_info);
	if (ret)
		goto error_free_idr;
	printk("c %p p %p\n", dev_get_drvdata(dev_info->dev), dev_info);
/* Interrupt triggered events setup */
	ret = iio_device_register_eventset(dev_info);
	if (ret)
		goto error_free_sysfs;


/* Ring buffer init if relevant */
	if (dev_info->modes & (INDIO_RING_TRIGGERED
			       | INDIO_RING_HARDWARE_BUFFER)) {
		ret = iio_device_register_ring(dev_info, 0);
		if (ret)
			goto error_free_eventset;
	}
/* Trigger registration */
	printk("b %p %p\n", dev_get_drvdata(dev_info->dev), dev_info);
/* Up to devices to register their own trigger */

	/* Register ability to use triggers */
	iio_device_register_trigger_consumer(dev_info);
	printk("a %p %p\n", dev_get_drvdata(dev_info->dev), dev_info);
	return 0;

error_free_eventset:
	iio_device_unregister_eventset(dev_info);
error_free_sysfs:
	iio_device_unregister_sysfs(dev_info);
error_free_idr:
	iio_device_unregister_id(dev_info);
error_nothing:

	return ret;
}
EXPORT_SYMBOL(iio_device_register);

void iio_device_unregister(struct iio_dev *dev_info)
{
  //	if (dev_info->modes & INDIO_RING_POLLED)
  //		iio_device_unregister_ptimer(dev_info);
	/*FIXME:ADD if*/
	iio_device_unregister_trigger_consumer(dev_info);
	if (dev_info->ring)
		iio_device_unregister_ring(dev_info);
	iio_device_unregister_eventset(dev_info);
	iio_device_unregister_sysfs(dev_info);
	iio_device_unregister_id(dev_info);

}
EXPORT_SYMBOL(iio_device_unregister);



subsys_initcall(iio_init);
module_exit(iio_exit);

MODULE_AUTHOR("Jonathan Cameron <jic23@cam.ac.uk>");
MODULE_DESCRIPTION("Industrial I/O core");
MODULE_LICENSE("GPL");
