/* The industrial I/O core
 *
 * Copyright (c) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Handling of ring allocation / resizing.
 *
 *
 * Things to look at here.
 * - Better memory allocation techniques?
 * - Alternative access techniques?
 */
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/industrialio/iio.h>
#include <linux/industrialio/ring_generic.h>

/* Allow for 3 digit id */
#define IIO_RING_AC_MINOR_N_LEN 22
#define IIO_RING_AC_MINOR_N "ring_buf%d_acc_minor"
#define IIO_RING_EV_MINOR_N_LEN 21
#define IIO_RING_EV_MINOR_N "ring_buf%d_ev_minor"

int iio_push_ring_event(struct iio_ring_buffer *ring_buf,
		       int event_code,
		       s64 timestamp)
{
	return __iio_push_event(&ring_buf->ev_int,
			       event_code,
			       timestamp,
			       &ring_buf->shared_ev_pointer);
}
EXPORT_SYMBOL(iio_push_ring_event);

int iio_push_or_escallate_ring_event(struct iio_ring_buffer *ring_buf,
				    int event_code,
				    s64 timestamp)
{
	if (ring_buf->shared_ev_pointer.ev_p)
		__iio_change_event(ring_buf->shared_ev_pointer.ev_p,
				   event_code,
				   timestamp);
	else
		return iio_push_ring_event(ring_buf,
					  event_code,
					  timestamp);
	return 0;
}
EXPORT_SYMBOL(iio_push_or_escallate_ring_event);

/* Next two absolutely rely on first element of all instances of ring buffers
   being an iio_ring_buffer */
int iio_ring_open(struct inode *inode, struct file *filp)
{
	void *r = filp->private_data;
	struct iio_ring_buffer *rb = r;

	if(rb->ring_access->mark_in_use)
		rb->ring_access->mark_in_use(r);
	try_module_get(rb->access_minor_attr.dev_attr.attr.owner);

	return 0;
}

int iio_ring_release(struct inode *inode, struct file *filp)
{
	void *r = filp->private_data;
	struct iio_ring_buffer *rb = r;

	module_put(rb->access_minor_attr.dev_attr.attr.owner);
	clear_bit(IIO_BUSY_BIT_POS, &rb->access_handler.flags);
	if(rb->ring_access->unmark_in_use)
		rb->ring_access->unmark_in_use(r);

	return 0;
}

/* Can only occur currently when the ring buffer is marked
   - from userspace call */
ssize_t iio_ring_rip_outer(struct file *filp,
			   char *buf,
			   size_t count,
			   loff_t *f_ps)
{
	void *r = filp->private_data;
	struct iio_ring_buffer *rb = r;
	int ret, dead_offset, copied;
	u8 *data;
	/* rip lots must exist. */
	if(!rb->ring_access->rip_lots)
		return -EINVAL; 
	copied = rb->ring_access->rip_lots(r, count, &data, &dead_offset);
	if (copied < 0) {
		ret = copied;
		goto error_ret;
	}
	if (copy_to_user(buf, data + dead_offset, copied))  {
		ret =  -EFAULT;
		goto error_free_data_cpy;
	}
	/* In clever ring buffer designs this may not need to be freed.
	 * When such a design exists I'll add this to ring access funcs. */
	kfree(data);

	return copied;

error_free_data_cpy:
	kfree(data);
error_ret:
	return ret;
}

static const struct file_operations iio_ring_fileops = {
	.read = iio_ring_rip_outer,
	.release = iio_ring_release,
	.open = iio_ring_open,
	.owner = THIS_MODULE,
};

static inline int
__iio_request_ring_buffer_event_chrdev(struct iio_ring_buffer *buf,
				       int id,
				       struct module *owner,
				       struct device *dev)
{
	int ret;

/* Create and register the event character device */
	buf->event_minor_name = kmalloc(IIO_RING_EV_MINOR_N_LEN, GFP_KERNEL);
	if (buf->event_minor_name == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}
	sprintf(buf->event_minor_name, IIO_RING_EV_MINOR_N, id);
	ret = iio_setup_ev_int(&(buf->ev_int),
			       (const char *)(buf->event_minor_name),
			       owner,
			       dev);
	if (ret)
		goto error_free_event_minor_name;

	return 0;

error_free_event_minor_name:
	kfree(buf->event_minor_name);
error_ret:
	return ret;
}

static inline void
__iio_free_ring_buffer_event_chrdev(struct iio_ring_buffer *buf,
				    struct device *dev)
{
	iio_free_ev_int(&(buf->ev_int), dev);
	kfree(buf->event_minor_name);
}

static inline int
__iio_request_ring_buffer_access_chrdev(struct iio_ring_buffer *buf,
					int id,
					struct module *owner,
					struct device *dev,
					const struct file_operations *fops)
{
	int ret;
/* Create and register the access character device */
	buf->access_minor_name = kmalloc(IIO_RING_AC_MINOR_N_LEN,
					 GFP_KERNEL);
	if (buf->access_minor_name == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}
	sprintf(buf->access_minor_name, IIO_RING_AC_MINOR_N, id);

	ret = iio_allocate_chrdev(&buf->access_handler);
	if (ret)
		goto error_free_access_minor_name;
	buf->access_handler.fops = fops;
	buf->access_handler.flags = 0;

	__init_iio_chrdev_minor_attr(&buf->access_minor_attr,
				   (const char *)(buf->access_minor_name),
				   owner,
				   buf->access_handler.id);

	ret = sysfs_create_file(&dev->kobj,
				&(buf->access_minor_attr.dev_attr.attr));
	if (ret)
		goto error_deallocate_chrdev;
	return 0;

error_deallocate_chrdev:
	iio_deallocate_chrdev(&buf->access_handler);
error_free_access_minor_name:
	kfree(buf->access_minor_name);
error_ret:
	return ret;
}

static inline void __iio_free_ring_buffer_access_chrdev(struct iio_ring_buffer *buf,
							struct device *dev)
{
	sysfs_remove_file(&dev->kobj,
			  &buf->access_minor_attr.dev_attr.attr);
	iio_deallocate_chrdev(&buf->access_handler);
	kfree(buf->access_minor_name);
}

static int iio_request_ring_buffer(struct iio_dev *dev_info,
				   int id,
				   struct module *owner,
				   struct device *dev)
{
	int ret;
	struct iio_ring_buffer *rb;
	ret = dev_info->ring_access._create(&dev_info->ring);
	if (ret)
		goto error_ret;
	rb = dev_info->ring;
	/* Not ideal */
	rb->ring_access = &dev_info->ring_access;
	/* Initially mark for update of ring parameters */
	if(dev_info->ring_access.mark_param_change)
		dev_info->ring_access.mark_param_change(dev_info->ring);

	ret = __iio_request_ring_buffer_event_chrdev(rb, id, owner, dev);
	if (ret)
		goto error_free_ring;

	ret = __iio_request_ring_buffer_access_chrdev(rb,
						      id,
						      owner,
						      dev,
						      &iio_ring_fileops);
	if (ret)
		goto error_free_ring_buffer_event_chrdev;

	rb->ev_int.private = rb;
	rb->access_handler.private = rb;

	return 0;
error_free_ring_buffer_event_chrdev:
	__iio_free_ring_buffer_event_chrdev(rb, dev);
error_free_ring:
	if(dev_info->ring_access._free)
		dev_info->ring_access._free(dev_info->ring);
error_ret:
	return ret;
}

static void iio_unrequest_ring_buffer(struct iio_dev *dev_info)
{

	__iio_free_ring_buffer_event_chrdev(dev_info->ring, dev_info->dev);
	if(dev_info->ring_access._free)
		dev_info->ring_access._free(dev_info->ring);
}

static ssize_t iio_read_ring_length(struct device *dev,
				    struct device_attribute *attr,
				    char *buf)
{
	int len = 0;
	struct iio_dev *dev_info = dev_get_drvdata(dev);

	if (dev_info->ring_access.get_length)
		len = sprintf(buf, "%d\n",
			      dev_info->ring_access.get_length(dev_info->ring));

	return len;
}

static ssize_t iio_write_ring_length(struct device *dev,
					      struct device_attribute *attr,
					      const char *buf,
					      size_t len)
{
	int ret;
	ulong val;
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	struct iio_ring_access_funcs *ra = &dev_info->ring_access;
	ret = strict_strtoul(buf, 10, &val);
	if (ret)
		return ret;

	if (dev_info->ring_access.get_length)
		if(val == ra->get_length(dev_info->ring))
			return len;

	if (ra->set_length) {
		ra->set_length(dev_info->ring, val);
		if (ra->mark_param_change)
			ra->mark_param_change(dev_info->ring);
	}

	return len;
}

static ssize_t iio_read_ring_bps(struct device *dev,
					  struct device_attribute *attr,
					  char *buf)
{
	int len = 0;
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	struct iio_ring_access_funcs *ra = &dev_info->ring_access;

	if (ra->get_bpd)
		len = sprintf(buf, "%d\n",
			      ra->get_bpd(dev_info->ring));

	return len;
}


static DEVICE_ATTR(length, S_IRUGO | S_IWUSR,
	    iio_read_ring_length,
	    iio_write_ring_length);
/* The software ring buffers aren't currently capable of changing the
 * storage accuracy so this is read only. Note bps can change due to
 * scan mode changes.
 */
static DEVICE_ATTR(bps, S_IRUGO, iio_read_ring_bps, NULL);

static ssize_t iio_store_ring_enable(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf,
				     size_t len)
{
	int ret;
	bool requested_state, current_state;
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	struct iio_ring_access_funcs *ra = &dev_info->ring_access;

	mutex_lock(&dev_info->mlock);
	requested_state = (buf[0] == '0') ? 0 : 1;
	current_state = !!(dev_info->currentmode & INDIO_ALL_RING_MODES);
	if (current_state == requested_state)
		goto done;
	if (requested_state) {
		if (dev_info->ring_preenable) {
			ret = dev_info->ring_preenable(dev_info);
			if (ret)
				goto error_ret;
		}
		if (ra->request_update) {
			ret = ra->request_update(dev_info->ring);
			if (ret)
				goto error_ret;
		}
		if (ra->mark_in_use)
			ra->mark_in_use(dev_info->ring);
		/* Definitely possible for devices to support both of these.*/
		if (dev_info->modes & INDIO_RING_TRIGGERED)
			dev_info->currentmode = INDIO_RING_TRIGGERED;
		else if (dev_info->modes & INDIO_RING_HARDWARE_BUFFER)
			dev_info->currentmode = INDIO_RING_HARDWARE_BUFFER;
		else { /* should never be reached */
			ret = -EINVAL;
			goto error_ret;
		}

		if (dev_info->ring_postenable) {
			ret = dev_info->ring_postenable(dev_info);
			if (ret)
				goto error_ret;
		}
	} else {
		if (dev_info->ring_predisable) {
			ret = dev_info->ring_predisable(dev_info);
			if (ret)
				goto error_ret;
		}
		if(ra->unmark_in_use)
			ra->unmark_in_use(dev_info->ring);
		dev_info->currentmode = INDIO_DIRECT_MODE;
		if (dev_info->ring_postdisable) {
			ret = dev_info->ring_postdisable(dev_info);
			if (ret)
				goto error_ret;
		}
	}
done:
	mutex_unlock(&dev_info->mlock);
	return len;

error_ret:
	mutex_unlock(&dev_info->mlock);
	return ret;
}

static ssize_t iio_show_ring_enable(struct device *dev,
				    struct device_attribute *attr,
				    char *buf)
{
	struct iio_dev *indio_dev = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", !!(indio_dev->currentmode
				       & INDIO_ALL_RING_MODES));
}

DEVICE_ATTR(ring_enable, S_IRUGO | S_IWUSR,
	    iio_show_ring_enable,
	    iio_store_ring_enable);

/* Standard set of ring buffer attributes */
static struct attribute *iio_ring_attributes[] = {
	&dev_attr_length.attr,
	&dev_attr_bps.attr,
	&dev_attr_ring_enable.attr,
	NULL,
};

static const struct attribute_group iio_ring_attribute_group = {
	.name = "ring_buffer",
	.attrs = iio_ring_attributes,
};

int iio_device_register_ring(struct iio_dev *dev_info, int id)
{
	int ret = 0;
	struct iio_ring_access_funcs *ra = &dev_info->ring_access;

	ret = iio_request_ring_buffer(dev_info, id, dev_info->driver_module,
				      dev_info->dev);
	if (ret)
		goto error_ret;
	if(ra->_init) {
		ret = ra->_init(dev_info->ring, dev_info);
		if (ret)
			goto error_unrequest_ring;
	}
	ret = sysfs_create_group(&dev_info->dev->kobj,
				 &iio_ring_attribute_group);
	if (ret)
		goto error_exit_ring;

	return 0;

error_exit_ring:
	if (ra->_exit)
		ra->_exit(dev_info->ring);
error_unrequest_ring:
	iio_unrequest_ring_buffer(dev_info);
error_ret:
	return ret;
}

void iio_device_unregister_ring(struct iio_dev *dev_info)
{
	struct iio_ring_access_funcs *ra = &dev_info->ring_access;
	struct iio_ring_buffer *rb = dev_info->ring;;
	if(dev_info->ring) {
		sysfs_remove_group(&dev_info->dev->kobj,
				   &iio_ring_attribute_group);
		__iio_free_ring_buffer_access_chrdev(rb, dev_info->dev);
		__iio_free_ring_buffer_event_chrdev(rb, dev_info->dev);
		if (ra->_exit)
			ra->_exit(dev_info->ring);
		iio_unrequest_ring_buffer(dev_info);
	}
}


