/* The industrial I/O core, trigger handling functions
 *
 * Copyright (c) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/idr.h>
#include <linux/kdev_t.h>
#include <linux/err.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/list.h>
#include <linux/platform_device.h>
#include <linux/industrialio/iio.h>
#include <linux/industrialio/trigger.h>

/* RFC - Question of approach
 * Make the common case (single sensor single trigger)
 * simple by starting trigger capture from when first sensors
 * is added.
 *
 * Complex simultaneous start requires use of 'hold' functionality
 * of the trigger. (not implemented)
 *
 * Any other suggestions?
 */


static DEFINE_IDR(iio_trigger_idr);
static DEFINE_SPINLOCK(iio_trigger_idr_lock);

LIST_HEAD(iio_trigger_list);
DEFINE_MUTEX(iio_trigger_list_lock);

/* Wrap in function? */
extern struct class iio_class;

static int iio_trigger_register_sysfs(struct iio_trigger *trig_info)
{
	int ret = 0;

	trig_info->sysfs_dev = device_create(&iio_class,
					     trig_info->dev,
					     MKDEV(0, 0),
					     NULL,
					     IIO_TRIGGER_ID_FORMAT,
					     trig_info->id);
	if (IS_ERR(trig_info->sysfs_dev))
		return -EINVAL;

	if (trig_info->control_attrs)
		ret = sysfs_create_group(&trig_info->dev->kobj,
					 trig_info->control_attrs);

	return ret;
}

static void iio_trigger_unregister_sysfs(struct iio_trigger *trig_info)
{
	if (trig_info->control_attrs)
		sysfs_remove_group(&trig_info->dev->kobj,
				   trig_info->control_attrs);
	device_unregister(trig_info->sysfs_dev);
}

static int iio_trigger_register_id(struct iio_trigger *trig_info)
{
	int ret = 0;

idr_again:
	if (unlikely(idr_pre_get(&iio_trigger_idr, GFP_KERNEL) == 0))
		return -ENOMEM;

	spin_lock(&iio_trigger_idr_lock);
	ret = idr_get_new(&iio_trigger_idr, NULL, &trig_info->id);
	spin_unlock(&iio_trigger_idr_lock);
	if (unlikely(ret == -EAGAIN))
		goto idr_again;
	else if (likely(!ret))
		trig_info->id = trig_info->id & MAX_ID_MASK;

	return ret;
}

static void iio_trigger_unregister_id(struct iio_trigger *trig_info)
{
	int id;
	if (likely(sscanf(trig_info->sysfs_dev->bus_id,
			  IIO_TRIGGER_ID_FORMAT, &id) == 1)) {
		spin_lock(&iio_trigger_idr_lock);
		idr_remove(&iio_trigger_idr, id);
		spin_unlock(&iio_trigger_idr_lock);
	} else
		dev_dbg(trig_info->dev->parent,
			"indio_device_unregister() failed: bad class ID!\n");
}

int iio_trigger_register(struct iio_trigger *trig_info,
			 struct device *parent,
			 int id)
{
	int ret;
	struct platform_device *subdev;

	subdev = platform_device_alloc(trig_info->name, id);
	if (!subdev) {
		ret = -ENOMEM;
		goto error_ret;
	}
	subdev->dev.parent = parent;
	ret = platform_device_add(subdev);
	if (ret)
		goto error_free_platform_device;
	trig_info->dev = &subdev->dev;
	dev_set_drvdata(trig_info->dev, (void *)(trig_info));
	ret = iio_trigger_register_id(trig_info);
	if (ret)
		goto error_free_platform_device;
	ret = iio_trigger_register_sysfs(trig_info);
	if (ret)
		goto error_unregister_id;

	mutex_lock(&iio_trigger_list_lock);
	list_add_tail(&trig_info->list, &iio_trigger_list);
	mutex_unlock(&iio_trigger_list_lock);

	return 0;

error_unregister_id:
	iio_trigger_unregister_id(trig_info);
error_free_platform_device:
	platform_device_put(subdev);
error_ret:
	return ret;
}
EXPORT_SYMBOL(iio_trigger_register);

void iio_trigger_unregister(struct iio_trigger *trig_info)
{
	struct iio_trigger *cursor;
	struct platform_device *subdev;

	mutex_lock(&iio_trigger_list_lock);
	list_for_each_entry(cursor, &iio_trigger_list, list)
		if (cursor == trig_info) {
			list_del(&cursor->list);
			break;
		}
	mutex_unlock(&iio_trigger_list_lock);

	iio_trigger_unregister_sysfs(trig_info);
	iio_trigger_unregister_id(trig_info);
	subdev = to_platform_device(trig_info->dev);
	platform_device_unregister(subdev);
}
EXPORT_SYMBOL(iio_trigger_unregister);

struct iio_trigger *iio_trigger_find_by_name(const char *name, size_t len)
{
	struct iio_trigger *trig;
	bool found = false;

	mutex_lock(&iio_trigger_list_lock);
	list_for_each_entry(trig, &iio_trigger_list, list) {
		if(strncmp(trig->name, name, len)==0){
			found = true;
			break;
		}
	}
	mutex_unlock(&iio_trigger_list_lock);

	return found ? trig : NULL;
};
EXPORT_SYMBOL(iio_trigger_find_by_name);


/* See explanation of poll_func structure in industrialio_trigger for
 * justification of double list approach
 */
void iio_trigger_poll(struct iio_trigger *trig)
{
	struct iio_poll_func *pf_cursor;

	list_for_each_entry(pf_cursor, &trig->pollfunc_list, list){
		if (pf_cursor->poll_func_immediate)
			pf_cursor->poll_func_immediate(pf_cursor->private_data);
	}
	list_for_each_entry(pf_cursor, &trig->pollfunc_list, list) {
		if (pf_cursor->poll_func_main)
			pf_cursor->poll_func_main(pf_cursor->private_data);
	}
}
EXPORT_SYMBOL(iio_trigger_poll);

ssize_t iio_trigger_read_name(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	struct iio_trigger *trig = dev_get_drvdata(dev);
	return sprintf(buf,"%s\n", trig->name);
}
EXPORT_SYMBOL(iio_trigger_read_name);

/* Trigger Consumer related functions */

/* Complexity in here.  With certain triggers (datardy) an acknowledgement
 * may be needed if the pollfuncs do not include the data read for the
 * triggering device.
 * This is not currently handled.  Alternative of not enabling trigger unless
 * the relevant function is in there may be the best option.
 */
/* Worth protecting against double additions?*/
int iio_trigger_attach_poll_func(struct iio_trigger *trig,
				 struct iio_poll_func *pf)
{
	int ret = 0;
	unsigned long flags;

	spin_lock_irqsave(trig->pollfunc_list_lock, flags);
	list_add_tail(&pf->list, &trig->pollfunc_list);
	spin_unlock_irqrestore(trig->pollfunc_list_lock, flags);

	if (trig->set_trigger_state)
		ret = trig->set_trigger_state(trig, true);

	return ret;
}
EXPORT_SYMBOL(iio_trigger_attach_poll_func);

int iio_trigger_dettach_poll_func(struct iio_trigger *trig,
				  struct iio_poll_func *pf)
{
	struct iio_poll_func *pf_cursor;
	unsigned long flags;
	int succeeded = 0;

	spin_lock_irqsave(trig->pollfunc_list_lock, flags);
	list_for_each_entry(pf_cursor, &trig->pollfunc_list, list)
		if (pf_cursor == pf) {
			succeeded = 1;
			break;
		}
	if (single_list(&trig->pollfunc_list) && trig->set_trigger_state) {
		spin_unlock_irqrestore(trig->pollfunc_list_lock, flags);
		/* May require interrupts */
		trig->set_trigger_state(trig, false);
		spin_lock_irqsave(trig->pollfunc_list_lock, flags);
	}
	/* Now we can delete safe the the knowledge that no interrupts
	 * can be generated if nothing is there to handle them. */
	if (succeeded)
		list_del(&pf_cursor->list);
	spin_unlock_irqrestore(trig->pollfunc_list_lock, flags);

	return succeeded ? 0 : -EINVAL;
}
EXPORT_SYMBOL(iio_trigger_dettach_poll_func);

static ssize_t iio_trigger_read_current(struct device *dev,
					 struct device_attribute *attr,
					 char *buf)
{
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	int len = 0;
	if (dev_info->trig)
		len = snprintf(buf,
			       IIO_TRIGGER_NAME_LENGTH,
			       "%s\n",
			       dev_info->trig->name);
	return len;
}

static ssize_t iio_trigger_write_current(struct device *dev,
					 struct device_attribute *attr,
					 const char *buf,
					 size_t len)
{
	struct iio_dev *dev_info = dev_get_drvdata(dev);
	struct iio_trigger *oldtrig = dev_info->trig;
	mutex_lock(&dev_info->mlock);
	if (dev_info->currentmode == INDIO_RING_TRIGGERED) {
		mutex_unlock(&dev_info->mlock);
		return -EBUSY;
	}
	mutex_unlock(&dev_info->mlock);

	len = len < IIO_TRIGGER_NAME_LENGTH ? len : IIO_TRIGGER_NAME_LENGTH;

	dev_info->trig = iio_trigger_find_by_name(buf, len);
	if (dev_info->trig != oldtrig)
		module_put(oldtrig->owner);
	else if (dev_info->trig)
		try_module_get(dev_info->trig->owner);

	return len;
}

DEVICE_ATTR(current_trigger, S_IRUGO | S_IWUSR,
	    iio_trigger_read_current,
	    iio_trigger_write_current);

static struct attribute *iio_trigger_consumer_attrs[] = {
	&dev_attr_current_trigger.attr,
	NULL,
};

static const struct attribute_group iio_trigger_consumer_attr_group = {
	.name = "trigger",
	.attrs = iio_trigger_consumer_attrs,
};

int iio_device_register_trigger_consumer(struct iio_dev *dev_info)
{
	int ret;
	ret = sysfs_create_group(&dev_info->dev->kobj,
				 &iio_trigger_consumer_attr_group);
	return ret;
}
EXPORT_SYMBOL(iio_device_register_trigger_consumer);

int iio_device_unregister_trigger_consumer(struct iio_dev *dev_info)
{
	sysfs_remove_group(&dev_info->dev->kobj,
			   &iio_trigger_consumer_attr_group);
	return 0;
}
EXPORT_SYMBOL(iio_device_unregister_trigger_consumer);


