#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/spinlock.h>
/*FIXME 0 shouldn't need this lot */
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
//#include <linux/rtc.h>

#include <linux/industrialio/ring_sw.h>

MODULE_DESCRIPTION("Industrialio I/O software ring buffer");
MODULE_LICENSE("GPL");

static inline int __iio_init_sw_ring_buffer(struct iio_sw_ring_buffer *ring,
					    int bytes_per_datum, int length)
{
	__iio_init_ring_buffer(&ring->buf, bytes_per_datum, length);
	ring->read_p = 0;
	ring->write_p = 0;
	ring->last_written_p = 0;
	ring->data = kmalloc(length*ring->buf.bpd, GFP_KERNEL);
	if (ring->data == NULL)
		return -ENOMEM;
	ring->use_count = 0;
	ring->use_lock = __SPIN_LOCK_UNLOCKED((ring)->use_lock);
	return 0;
}


static inline void __iio_free_sw_ring_buffer(struct iio_sw_ring_buffer *ring)
{
	kfree(ring->data);
}
void iio_mark_sw_rb_in_use(void *r)
{
	struct iio_sw_ring_buffer *ring = r;
	spin_lock(&ring->use_lock);
	ring->use_count++;
	spin_unlock(&ring->use_lock);
}
EXPORT_SYMBOL(iio_mark_sw_rb_in_use);

void iio_unmark_sw_rb_in_use(void *r)
{
	struct iio_sw_ring_buffer *ring = r;
	spin_lock(&ring->use_lock);
	ring->use_count--;
	spin_unlock(&ring->use_lock);
}
EXPORT_SYMBOL(iio_unmark_sw_rb_in_use);


/* Ring buffer related functionality */
/* Store to ring is typically called in the bh of a data ready interrupt handler
 * in the device driver */
/* Lock always held if their is a chance this may be called */
int iio_store_to_sw_ring(struct iio_sw_ring_buffer *ring,
			 unsigned char *data,
			 s64 timestamp)
{
	bool init_read = true;
	int ret;
	int code;

	/* initial store */
	if (unlikely(ring->write_p == 0)) {
		ring->write_p = ring->data;
		/* doesn't actually matter if this is out of the set */
		ring->half_p = ring->data - ring->buf.length*ring->buf.bpd/2;
		init_read = false;
	}
	memcpy(ring->write_p, data, ring->buf.bpd);
	barrier();
	ring->last_written_p = ring->write_p;
	barrier();
	ring->write_p += ring->buf.bpd;
	/* End of ring, back to the beginning */
	if (ring->write_p == ring->data + ring->buf.length*ring->buf.bpd) {
		ring->write_p = ring->data;
		ring->buf.loopcount++;
	}
	if (ring->read_p == 0)
		ring->read_p = ring->data;
	/* Buffer full - move the read pointer and create / escalate
	 * ring event */
	else if (ring->write_p == ring->read_p) {
		ring->read_p += ring->buf.bpd;
		if (ring->read_p
		    == ring->data + ring->buf.length*ring->buf.bpd)
			ring->read_p = ring->data;

		spin_lock(&ring->buf.shared_ev_pointer.lock);
		if (ring->buf.shared_ev_pointer.ev_p) {
			/* Event escalation - probably quicker to let this
			   keep running than check if it is necessary */
			code = IIO_EVENT_CODE_RING_100_FULL;
			__iio_change_event(ring
					   ->buf.shared_ev_pointer.ev_p,
					   code,
					   timestamp);
		} else {
			code = IIO_EVENT_CODE_RING_100_FULL;
			ret = __iio_push_event(&ring->buf.ev_int,
					       code,
					       timestamp,
					       &ring
					       ->buf.shared_ev_pointer);
			if (ret) {
				spin_unlock(&ring->buf.shared_ev_pointer.lock);
				goto error_ret;
			}
		}
		spin_unlock(&ring->buf.shared_ev_pointer.lock);
	}
	/* investigate if our event barrier has been passed */
	/* There are definite 'issues' with this and chances of
	 * simultaneous read */
	/* Also need to use loop count to ensure this only happens once */
	ring->half_p += ring->buf.bpd;
	if (ring->half_p == ring->data + ring->buf.length*ring->buf.bpd)
		ring->half_p = ring->data;
	if (ring->half_p == ring->read_p) {
		spin_lock(&ring->buf.shared_ev_pointer.lock);
		code = IIO_EVENT_CODE_RING_50_FULL;
		ret = __iio_push_event(&ring->buf.ev_int,
				       code,
				       timestamp,
				       &ring->buf.shared_ev_pointer);
		spin_unlock(&ring->buf.shared_ev_pointer.lock);

		if (ret)
			goto error_ret;
	}
	return 0;
error_ret:
	return ret;
}

int iio_rip_sw_rb(void *r,  size_t count, u8 **data, int *dead_offset)
{
	struct iio_sw_ring_buffer *ring = r;

	u8 *initial_read_p, *initial_write_p, *current_read_p, *end_read_p;

	int ret, max_copied;
	/* Round down to nearest datum boundary */
	int bytes_to_rip = count - count % ring->buf.bpd;

	/* Limit size to whole of ring buffer */
	bytes_to_rip = min(ring->buf.bpd*ring->buf.length, bytes_to_rip);

	*data = kmalloc(bytes_to_rip, GFP_KERNEL);
	if (*data == NULL) {
		ret = -ENOMEM;
		goto error_ret;
	}

	/* build local copy */
	initial_read_p = ring->read_p;
	if (unlikely(initial_read_p == 0)) { /* No data here as yet */
		ret = 0;
		goto error_free_data_cpy;
	}

	initial_write_p = ring->write_p;

	/* Need a consistent pair */
	while ((initial_read_p != ring->read_p)
	       || (initial_write_p != ring->write_p)) {
		initial_read_p = ring->read_p;
		initial_write_p = ring->write_p;
	}
	if (initial_write_p == initial_read_p) {
		/* No new data available.*/
		ret = 0;
		goto error_free_data_cpy;
	}

	if (initial_write_p > initial_read_p + bytes_to_rip) {
		/* write_p is greater than necessary, all is easy */
		max_copied = bytes_to_rip;
		memcpy(*data, initial_read_p, max_copied);
		end_read_p = initial_read_p + max_copied;
	} else if (initial_write_p > initial_read_p) {
		/*not enough data to cpy */
		max_copied = initial_write_p - initial_read_p;
		memcpy(*data, initial_read_p, max_copied);
		end_read_p = initial_write_p;
	} else {
		/* going through 'end' of ring buffer */
		max_copied = ring->data
			+ ring->buf.length*ring->buf.bpd - initial_read_p;
		memcpy(*data, initial_read_p, max_copied);
		if (initial_write_p > ring->data + bytes_to_rip - max_copied) {
			/* enough data to finish */
			memcpy(*data + max_copied, ring->data,
			       bytes_to_rip - max_copied);
			max_copied = bytes_to_rip;
			end_read_p = ring->data + (bytes_to_rip - max_copied);
		} else {  /* not enough data */
			memcpy(*data + max_copied, ring->data,
			       initial_write_p - ring->data);
			max_copied += initial_write_p - ring->data;
			end_read_p = initial_write_p;
		}
	}
	/* Now to verify which section was cleanly copied - i.e. how far
	 * read pointer has been pushed */
	current_read_p = ring->read_p;

	if (initial_read_p <= current_read_p)
		*dead_offset = current_read_p - initial_read_p;
	else
		*dead_offset = ring->buf.length*ring->buf.bpd
			- (initial_read_p - current_read_p);

	/* possible issue if the initial write has been lapped or indeed
	 * the point we were reading to has been passed */
	/* No valid data read.
	 * In this case the read pointer is already correct having been
	 * pushed further than we would look. */
	if (max_copied - *dead_offset < 0) {
		ret = 0;
		goto error_free_data_cpy;
	}

	/* setup the next read position */
	ring->read_p = end_read_p;
	return max_copied - *dead_offset;

error_free_data_cpy:
	kfree(data);
error_ret:
	return 0;
}
EXPORT_SYMBOL(iio_rip_sw_rb);
//EXPORT_SYMBOL_GPL(iio_store_to_sw_ring);
int iio_store_to_sw_rb(void *r, u8 *data, s64 timestamp)
{
	struct iio_sw_ring_buffer *ring = r;
	return iio_store_to_sw_ring(ring, data, timestamp);
}
EXPORT_SYMBOL(iio_store_to_sw_rb);
/* FIXME - combine with below */
int iio_read_last_from_sw_ring(struct iio_sw_ring_buffer *ring,
			       unsigned char *data)
{
	int loopcount_copy;
	unsigned char *last_written_p_copy;

	iio_mark_sw_rb_in_use(ring);
again:
	loopcount_copy = ring->buf.loopcount;
	barrier();
	last_written_p_copy = ring->last_written_p;
	barrier(); /*unnessecary? */
	/* Check there is anything here */
	if (last_written_p_copy == 0)
		return -EAGAIN;
	memcpy(data, last_written_p_copy, ring->buf.bpd);

	if (unlikely(loopcount_copy != ring->buf.loopcount)) {
		if (unlikely(ring->last_written_p >= last_written_p_copy))
			goto again;
	}
	iio_unmark_sw_rb_in_use(ring);
	return 0;
}

/*doesn't currently read the timestamp */
/* For software ring buffers this function is needed to get the latest
 * reading without preventing it from ending up in the ring buffer.
*/
int iio_read_last_from_sw_rb(void *r,
			     unsigned char *data)
{
	struct iio_sw_ring_buffer *ring = r;
	return iio_read_last_from_sw_ring(ring, data);
}
EXPORT_SYMBOL(iio_read_last_from_sw_rb);

int iio_create_sw_rb(void **r)
{
	*r = kzalloc(sizeof(struct iio_sw_ring_buffer),
			GFP_KERNEL);
	if (*r == NULL)
		return -ENOMEM;
	return 0;
}
EXPORT_SYMBOL(iio_create_sw_rb);

int iio_init_sw_rb(void *r, void *indio_dev)
{
	return 0;
}
EXPORT_SYMBOL(iio_init_sw_rb);

void iio_exit_sw_rb(void *r)
{
	struct iio_sw_ring_buffer *ring = r;
	__iio_free_sw_ring_buffer(ring);

}
EXPORT_SYMBOL(iio_exit_sw_rb);


void iio_free_sw_rb(void *r)
{
	kfree(r);
}
EXPORT_SYMBOL(iio_free_sw_rb);


/* Still needs work!*/
int iio_request_update_sw_rb(void *r)
{
	int ret;
	struct iio_sw_ring_buffer *ring = r;
	printk("iio_request swrb\n");
	spin_lock(&ring->use_lock);
	if (ring->use_count || !ring->update_needed) {
		printk("error\n");
		ret = -EAGAIN;
		goto error_ret;
	}
	__iio_free_sw_ring_buffer(ring);
	ret = __iio_init_sw_ring_buffer(ring, ring->buf.bpd, ring->buf.length);
error_ret:
	spin_unlock(&ring->use_lock);
	return ret;
}
EXPORT_SYMBOL(iio_request_update_sw_rb);

int iio_get_bpd_sw_rb(void *r)
{
	struct iio_sw_ring_buffer *ring = r;
	return ring->buf.bpd;
}
EXPORT_SYMBOL(iio_get_bpd_sw_rb);

int iio_set_bpd_sw_rb(void *r, int bpd)
{
	struct iio_sw_ring_buffer *ring = r;
	if (ring->buf.bpd != bpd) {
		ring->buf.bpd = bpd;
		if(ring->buf.ring_access->mark_param_change)
			ring->buf.ring_access->mark_param_change(r);
	}
	return 0;
}
EXPORT_SYMBOL(iio_set_bpd_sw_rb);

int iio_get_length_sw_rb(void *r)
{
	struct iio_sw_ring_buffer *ring = r;
	return ring->buf.length;
}
EXPORT_SYMBOL(iio_get_length_sw_rb);

int iio_set_length_sw_rb(void *r, int length)
{
	struct iio_sw_ring_buffer *ring = r;
	if(ring->buf.length != length) {
		ring->buf.length = length;
		if(ring->buf.ring_access->mark_param_change)
			ring->buf.ring_access->mark_param_change(r);
	}
	return 0;
}
EXPORT_SYMBOL(iio_set_length_sw_rb);

int iio_mark_update_needed_sw_rb(void *r)
{
	struct iio_sw_ring_buffer *ring = r;
	ring->update_needed = true;
	return 0;
}
EXPORT_SYMBOL(iio_mark_update_needed_sw_rb);

/* no init or exit as this is basically a library module */
