
/*
 * sht15.c -- support for SHT15 Temperature and Humidity  Sensor
 *
 * Copyright (c) 2008 Jonathan Cameron (General Purpose GPIO + port to
 *                                      2.6.25)
 *
 * Copyright (c) 2007 Wouter Horre (Original driver)
 *
 *This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/irq.h>
#include <asm/uaccess.h> /* for copy_to_user */
#include <mach/hardware.h>
#include <mach/pxa-regs.h>  /* for the register access macro's */
#include <mach/irqs.h> /* interrupts */
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/sht15.h>


MODULE_LICENSE("GPL");
MODULE_AUTHOR("wouter.horre@cs.kuleuven.be");
MODULE_DESCRIPTION("SHT15 humidity/temperature sensor");

// parameter that can be changed at load time
int sht15_major = SHT15_MAJOR;
module_param(sht15_major, int, S_IRUGO);

static struct platform_driver driver = {
	.driver = {
		.name = "sht15",
		.owner = THIS_MODULE,
	},
	
};

static sht15_state sht15_dev_state;

// File operations
static int sht15_read(struct file *filp, 
		      char *buffer, 
		      size_t count, 
		      loff_t *ppos);
static int sht15_open(struct inode * inode, 
		      struct file * file);
static int sht15_release(struct inode * inode, 
			 struct file * file);
static int sht15_ioctl(struct inode * inode, 
		       struct file * file, 
		       unsigned int cmd, 
		       unsigned long arg);

struct file_operations sht15_fops = {
owner:          THIS_MODULE,
read:           sht15_read,
open:           sht15_open,
release:        sht15_release,
ioctl:          sht15_ioctl,
};

// Tasklet
static void sht15_read_measurement_tasklet(unsigned long);
DECLARE_TASKLET(sht15_read_measurement, sht15_read_measurement_tasklet, 0);

static void sht15_connection_reset(void)
{
	int i;
	gpio_direction_output(sht15_dev_state.pdata->gpio_data, 1);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
	for(i=0; i<9; ++i) {
		gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
		ndelay(2);
		gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0); 
	}
}

static void sht15_transmission_start(void)
{
	gpio_direction_output(sht15_dev_state.pdata->gpio_data, 1);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_data, 0);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_data, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
}

static void sht15_end_transmission(void)
{
	gpio_direction_output(sht15_dev_state.pdata->gpio_data, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
}

static void sht15_ack(void)
{
	gpio_direction_output(sht15_dev_state.pdata->gpio_data, 0);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
	gpio_direction_input(sht15_dev_state.pdata->gpio_data);
	gpio_set_value(sht15_dev_state.pdata->gpio_data, 1);
}

static void sht15_enable_interrupt(void)
{
	gpio_direction_input(sht15_dev_state.pdata->gpio_data);
	gpio_set_value(sht15_dev_state.pdata->gpio_data, 1);
	set_irq_type(IRQ_GPIO(sht15_dev_state.pdata->gpio_data),IRQ_TYPE_EDGE_FALLING);
}

static void sht15_disable_interrupt(void)
{
	set_irq_type(IRQ_GPIO(sht15_dev_state.pdata->gpio_data),IRQ_TYPE_NONE);
}

static void sht15_init_pins(void)
{
	gpio_direction_output(sht15_dev_state.pdata->gpio_sck, 0);
	gpio_direction_input(sht15_dev_state.pdata->gpio_data);
	gpio_set_value(sht15_dev_state.pdata->gpio_data, 1);
	//disable interrupt
	sht15_disable_interrupt();
}

static void sht15_send_byte(uint8_t byte)
{
	int i;
	printk (KERN_DEBUG "STH15: Sending byte %x\n",byte);
	for(i=0; i < 8; ++i){
		if(byte & 0x80){
			gpio_set_value(sht15_dev_state.pdata->gpio_data, 1);
		} else {
			gpio_set_value(sht15_dev_state.pdata->gpio_data, 0);
		}
		byte = byte << 1;
		gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
		ndelay(2);
		ndelay(2);
		gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
	}
}

static int sht15_wait_for_response(void)
{
	gpio_direction_input(sht15_dev_state.pdata->gpio_data);
	gpio_set_value(sht15_dev_state.pdata->gpio_data, 1);
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
	ndelay(2);
	if(gpio_get_value(sht15_dev_state.pdata->gpio_data)) { 
		// sht15 did not pull data low
		sht15_connection_reset();
		return FAIL;
	}
	ndelay(2);
	gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
	return SUCCESS;
}

static int sht15_send_cmd(void)
{
	sht15_init_pins();
	sht15_connection_reset();
	sht15_transmission_start();
	sht15_send_byte(sht15_dev_state.cmd&0x1F);
	return sht15_wait_for_response();
}

static irqreturn_t sht15_interrupt_fired(int this_irq, void * dev_id)
{
	printk (KERN_DEBUG "SHT15: Interrupt fired\n");
	//disable interrupt
	sht15_disable_interrupt();
	tasklet_schedule(&sht15_read_measurement);
	return IRQ_HANDLED;
}

static void sht15_read_measurement_tasklet(unsigned long unused)
{
	int i;
	uint16_t data = 0;
	printk (KERN_DEBUG "SHT15: Read measurement tasklet executing\n");
	
	gpio_direction_input(sht15_dev_state.pdata->gpio_data);
	for(i=0; i < 16; ++i){
		data = data << 1;
		gpio_set_value(sht15_dev_state.pdata->gpio_sck, 1);
		ndelay(2);
		data |=  gpio_get_value(sht15_dev_state.pdata->gpio_data) 
			? 1 : 0;
		gpio_set_value(sht15_dev_state.pdata->gpio_sck, 0);
		if(i%8 == 7) {
			sht15_ack();
			gpio_direction_input(sht15_dev_state.pdata
					     ->gpio_data);
		}
	}
	sht15_end_transmission();
	sht15_dev_state.reading = data;
	sht15_dev_state.flag = TRUE;
	wake_up(&sht15_dev_state.sht15_wait_queue);
}


	
static __init int sht15_probe(struct platform_device *pdev)
{
	int result;

	printk(KERN_INFO "SHT15 initializing\n");
	sht15_dev_state.pdata = pdev->dev.platform_data;
	gpio_request(sht15_dev_state.pdata->gpio_sck,  "SHT15 SCK");
	gpio_request(sht15_dev_state.pdata->gpio_data, "SHT15 Data");
//Initialize pins
	sht15_init_pins();
	
	// register device node
	
	sht15_dev_state.busy = FALSE;
	sht15_dev_state.fops = &sht15_fops;
	sht15_dev_state.cmd = SHT15_CMD_NULL;
	sht15_dev_state.flag = FALSE;
	sht15_dev_state.reading = 0;
	init_waitqueue_head(&sht15_dev_state.sht15_wait_queue);
	if(sht15_major){
		sht15_dev_state.devno = MKDEV(sht15_major, SHT15_TEMP_MINOR);
		result = register_chrdev_region(sht15_dev_state.devno, 
						SHT15_SENSOR_COUNT, 
						"sht15");
	} else {
		result = alloc_chrdev_region(&sht15_dev_state.devno,
					     SHT15_TEMP_MINOR,
					     SHT15_SENSOR_COUNT,
					     "sht15");
	}
	if(result < 0){
		printk (KERN_WARNING 
			"SHT15: failed -- could not allocate chrdev region\n");
		goto fail_alloc_chrdev;
	}
	printk (KERN_INFO "SHT15: major: %d, first minor %d\n", 
		MAJOR(sht15_dev_state.devno), 
		MINOR(sht15_dev_state.devno));
	
	cdev_init(&sht15_dev_state.dev, &sht15_fops);
	sht15_dev_state.dev.owner = THIS_MODULE;
	sht15_dev_state.dev.ops = &sht15_fops;
	
	if(cdev_add(&sht15_dev_state.dev,sht15_dev_state.devno, 
		    SHT15_SENSOR_COUNT)){
		printk (KERN_WARNING "SHT15: failed -- could not add cdev\n");
		goto fail_add_cdev;
	}
	
	// request IRQ
	if(request_irq(IRQ_GPIO(sht15_dev_state.pdata->gpio_data), 
		       sht15_interrupt_fired, 
		       IRQF_DISABLED, 
		       "SHT15 DATA", 
		       NULL)){
		printk (KERN_WARNING "SHT15: failed -- irq %d not free\n",
			IRQ_GPIO(sht15_dev_state.pdata->gpio_data));
		goto fail_irq_data;
	}
	
	return 0;
	
fail_irq_data:
	cdev_del(&sht15_dev_state.dev);
fail_add_cdev:
	unregister_chrdev_region(sht15_dev_state.devno, SHT15_SENSOR_COUNT);
fail_alloc_chrdev:
	return -EIO;
}
static __init int sht15_init(void)
{
	return platform_driver_probe(&driver, sht15_probe);
}

static void __exit sht15_exit(void)
{
	printk(KERN_INFO "SHT15 cleaning up\n");
	gpio_free(sht15_dev_state.pdata->gpio_sck);
	gpio_free(sht15_dev_state.pdata->gpio_data);
	free_irq(IRQ_GPIO(sht15_dev_state.pdata->gpio_data),NULL);
	cdev_del(&sht15_dev_state.dev);
	unregister_chrdev_region(sht15_dev_state.devno, SHT15_SENSOR_COUNT);
	platform_driver_unregister(&driver);
}

static int sht15_read(struct file *filp, 
		      char *buffer, 
		      size_t count, 
		      loff_t *ppos)
{
	int minor;
	printk (KERN_DEBUG "SHT15: read called\n");
	
	if((*ppos) > 0){
		// We have already sent a reading -> return 0 (EOF)
		return 0;
	}
	
	minor = iminor(filp->f_dentry->d_inode);
	switch(minor) {
	case SHT15_RH_MINOR:
		sht15_dev_state.cmd = SHT15_MEASURE_RH;
		break;
	case SHT15_TEMP_MINOR:
	default: // Default = measure temperature
		sht15_dev_state.cmd = SHT15_MEASURE_TEMP;
		break;
	}
	if(sht15_send_cmd()!=SUCCESS){
		printk( KERN_WARNING 
			"SHT15: could not send command to device\n");
		return -EIO;
	}
	//enable interrupts
	sht15_enable_interrupt();
	//put calling process asleep
	wait_event(sht15_dev_state.sht15_wait_queue, 
		   sht15_dev_state.flag == TRUE);
	//read data
	if(sht15_dev_state.flag==TRUE){
		sht15_dev_state.flag=FALSE;
		if(copy_to_user(buffer,&sht15_dev_state.reading,2)){
			return -EFAULT;
		} else {
			*ppos += 2;
			return 2;
		}
	} else {
		return -EIO;
	}
}

static int sht15_open(struct inode * inode, struct file * file)
{
	printk (KERN_DEBUG "SHT15: open called\n");
	if(sht15_dev_state.busy == TRUE){
		return -EBUSY;
	} else if(file->f_flags & O_NONBLOCK) {
		// We cannot do this
		return -EAGAIN;
	} else {
		sht15_dev_state.busy = TRUE;
		return 0;
	}
}

static int sht15_release(struct inode * inode, 
			 struct file * file)
{
	printk (KERN_DEBUG "STH15: release called\n");
	sht15_dev_state.busy = FALSE;
	return 0;
}

static int sht15_ioctl(struct inode * inode, 
		       struct file * file, 
		       unsigned int cmd, 
		       unsigned long arg)
{
	printk (KERN_DEBUG "SHT15: ioctl called\n");
	// TODO
	return 0;
}

module_init(sht15_init);
module_exit(sht15_exit);
