/*
 * linux/drivers/pcmcia/pxa2xx_stargate2.c
 *
 * Stargate 2 PCMCIA specific routines.
 *
 * Created:	December 6, 2005
 * Author:	Ed C. Epp
 * Copyright:	Intel Corp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>

#include <pcmcia/ss.h>

#include <asm/irq.h>
#include <asm/io.h>

#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/pxa2xx-regs.h>
#include <mach/stargate2.h>

#include "soc_common.h"

static struct pcmcia_irqs irqs[] = {
	{ 0, SG2_S0_CD_IRQ, "PCMCIA0 CD" },
 };

static int sg2_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
   	skt->irq = SG2_S0_IRQ;
	return soc_pcmcia_request_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

static void sg2_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
	soc_pcmcia_free_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

static unsigned long sg2_pcmcia_status[2];

static void sg2_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
				    struct pcmcia_state *state)
{
  	unsigned long status, flip;

  	status = SG2_PCMCIA0 ;
  	flip = (status ^ sg2_pcmcia_status[skt->nr]) & SG2_PCMCIA_nSTSCHG_BVD1;
	state->detect = gpio_get_value(SG2_S0_GPIO_DETECT) ? 0 : 1;
	state->ready  = gpio_get_value(SG2_S0_GPIO_READY) ? 1 : 0;
	state->bvd1   = 0; /* not available - battery detect on card */
	state->bvd2   = 0; /* not available */
	state->vs_3v  = 1; /* not available - voltage detect for card */
	state->vs_Xv  = 0; /* not available */
	state->wrprot = 0; /* not available - write protect */
}

static int sg2_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
				       const socket_state_t *state)
{
        int detect;


        /* Enable card power */
        switch (state->Vcc) {
	case 0:
		//sets power ctl register high.
		gpio_set_value(SG2_S0_POWER_CTL, 1);
		break;
        case 33:
        case 50:
		//sets power control register low (clear)
		gpio_set_value(SG2_S0_POWER_CTL, 0);
		mdelay(100);
		//sets reset high
		gpio_set_value(SG2_S0_GPIO_RESET, 1);
		udelay(15);
		//clears reset.
		gpio_set_value(SG2_S0_GPIO_RESET, 0);
		mdelay(25);
		break;
        default:
		printk(KERN_ERR "%s(): bad Vcc %u\n",
		       __FUNCTION__, state->Vcc);
	  return -1;
	}

        // reset
	if (state->flags&SS_RESET)
	{
          mdelay(25);
	  //set reset
	  gpio_set_value(SG2_S0_GPIO_RESET, 1);
          udelay(15);               // 10 usec min
	  //clear reset.
	  gpio_set_value(SG2_S0_GPIO_RESET, 0);
          mdelay(25);
	}


	//turns on irq handling for this guy. why here?
        soc_pcmcia_enable_irqs(skt, irqs, ARRAY_SIZE(irqs));
	//read detection gpio is high or not.
	/* suspect unecessary? */
	detect = gpio_get_value(SG2_S0_GPIO_DETECT) ? 0 : 1;

        if (detect) {
            MECR |= MECR_CIT;
	} else {
            MECR &= ~MECR_CIT;
	}
	return 0;
}

static void sg2_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{

}

static void sg2_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{

}

static struct pcmcia_low_level sg2_pcmcia_ops = {
	.owner			= THIS_MODULE,
	.hw_init		= sg2_pcmcia_hw_init,
	.hw_shutdown		= sg2_pcmcia_hw_shutdown,
	.socket_state		= sg2_pcmcia_socket_state,
	.configure_socket	= sg2_pcmcia_configure_socket,
	.socket_init		= sg2_pcmcia_socket_init,
	.socket_suspend		= sg2_pcmcia_socket_suspend,
	.nr			= 1,
};

static struct platform_device *sg2_pcmcia_device;

static int __init sg2_pcmcia_init(void)
{
	int ret;
	sg2_pcmcia_device = platform_device_alloc("pxa2xx-pcmcia", -1);
	if (!sg2_pcmcia_device)
		return -ENOMEM;
	sg2_pcmcia_device->dev.platform_data = &sg2_pcmcia_ops;
	ret = platform_device_add(sg2_pcmcia_device);
	if (ret)
	  platform_device_put(sg2_pcmcia_device);
	return ret;
}

static void __exit sg2_pcmcia_exit(void)
{
	platform_device_unregister(sg2_pcmcia_device);
}

fs_initcall(sg2_pcmcia_init);
module_exit(sg2_pcmcia_exit);

MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:pxa2xx-pcmcia");
