/*
 * Real Time Clock interface for periodic timers on the XScale PXA27x
 *
 * Copyright (c) 2008 Jonathan Cameron
 *
 * Based on rtc-sa1100.c by Nils Faerber and others
 * and rtc-cmos.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * Uses the periodic functionality of operating system timers 5-11.
 * OSC 4 is used to allow checking of whether an interrupt has been missed.
 *
 * Purpose of this driver is to make available 7 periodic timers for use
 * by other drivers. In the fullness of time, this will be replaced by a
 * more generic timers interface such as that described by David Brownell.
 */

#include <linux/platform_device.h>
#include <linux/module.h>
#include <linux/rtc.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/interrupt.h>

#include <mach/hardware.h>
#include <mach/pxa-regs.h>


/* FIXME - move to pxa headers - awaiting pxa tree clearing out
 * as per request from RMK */

#define OMCR_C				(1 << 7)

/*Channel continues incrementing after detecting a match */
#define OMCR_PERIODIC			(1 << 6)

/* Reset OSCRx on rising edge of ext_sync 0 */
#define OMCR_EXT_SYNC0			((0 << 5) | (1 << 4))
#define OMCR_EXT_SYNC1			((1 << 5) | (0 << 4))
#define OMCR_OSCR_RESET_ON_MATCH	(1 << 3)
#define OMCR_FREQ_32768HZ		0x01
#define OMCR_FREQ_1KHZ			0x02
#define OMCR_FREQ_1HZ			0x03
#define OMCR_FREQ_1MHZ			0x04
#define OMCR_FREQ_EXT			0x05

#define OIER_E(x) (1 << (x))
#define OSSR_M(x) (1 << (x))

/* Only applicable for x > 4 - generalize these?*/
#define OSCRx(x) __REG2(0x40A00030, ((x) << 2))

#define OMCRx(x) __REG2(0x40A000B0, ((x) << 2))

#define OSMRx(x) __REG2(0x40A00070, ((x) << 2))

/* Will contain more info at some stage in future */
struct pxa27x_periodic_counter_info {
	int num;
};

static const struct pxa27x_periodic_counter_info pxa27x_pcinfo[] = {
	{
		.num = 5,
	}, {
		.num = 6,
	}, {
		.num = 7,
	}, {
		.num = 8,
	}, {
		.num = 9,
	}, {
		.num = 10,
	}, {
		.num = 11,
	},
};

#define PXA27X_IRQ_SET_STATE(_n)					\
	static int pxa27x_irq_set_state_##_n(struct device *dev, int enabled) \
	{								\
		return pxa27x_irq_set_state_counted(dev, enabled, _n);	\
	}

#define PXA27X_IRQ_SET_FREQ(_n)						\
	static int pxa27x_irq_set_freq_##_n(struct device *dev, int freq) \
	{								\
		return pxa27x_irq_set_freq_counted(dev, freq, _n);	\
	}

#define PXA27X_RTC_PROC(_n)						\
	static int pxa27x_rtc_proc_##_n(struct device *dev,		\
					struct seq_file *seq)		\
	{								\
		return pxa27x_rtc_proc_counted(dev, seq, _n);		\
	}

#define PXA27X_RTC_IOCTL(_n)						\
	static int pxa27x_rtc_ioctl_##_n(struct device *dev, unsigned int cmd, \
					 unsigned long arg) {		\
		return pxa27x_rtc_ioctl_counted(dev, cmd, arg, _n);	\
	}

#define PXA27X_PERIODIC_RTC_FUNCS(_n)		\
	PXA27X_IRQ_SET_STATE(_n);		\
	PXA27X_RTC_PROC(_n);			\
	PXA27X_RTC_IOCTL(_n);			\
	PXA27X_IRQ_SET_FREQ(_n);		\
	static const struct rtc_class_ops pxa27x_periodic_rtc_ops_##_n = { \
		.ioctl = pxa27x_rtc_ioctl_##_n,				\
		.proc = pxa27x_rtc_proc_##_n,				\
		.irq_set_freq = pxa27x_irq_set_freq_##_n,		\
		.irq_set_state = pxa27x_irq_set_state_##_n,		\
	}

#define PXA27X_RTC_PERIODIC_FREQ 32768

#define PXA27X_NUM_PERIODIC_RTCS 7
static unsigned long rtc_freq[PXA27X_NUM_PERIODIC_RTCS];
static DEFINE_SPINLOCK(pxa27x_rtc_lock);

static unsigned int last_int_time[PXA27X_NUM_PERIODIC_RTCS];

/* Based on docs, it is possible that more than one interrupt can be generated
 * before any are handled thus leading to this hander running more often than
 * strictly necessary. Should still work despite this. Not yet observed in the
 * wild.
 */
static irqreturn_t pxa27x_periodic_interrupt(int irq, void *dev_id)
{
	int i;
	struct platform_device *pdev = to_platform_device(dev_id);
	struct rtc_device **rtc = platform_get_drvdata(pdev);
	int num_interrupts, period;
	/* Must be longer than 32 bits!*/
	unsigned int current_time = OSCRx(4);
	int state = OSSR;

	for (i = 0; i < PXA27X_NUM_PERIODIC_RTCS; i++)
		if (state & OSSR_M(pxa27x_pcinfo[i].num)) {

			period = (PXA27X_RTC_PERIODIC_FREQ/rtc_freq[i]);
			num_interrupts = (current_time
					  - last_int_time[i]) / period;
			/*deal with wrap around in osc 4 */
			if (current_time < last_int_time[i])
				num_interrupts
					= (0xFFFFFFFF - last_int_time[i]
					   + current_time)/period;
			last_int_time[i] = current_time;
			rtc_update_irq(rtc[i],
				       num_interrupts, RTC_PF | RTC_IRQF);
			OSSR = OSSR_M(pxa27x_pcinfo[i].num); /* clear match */
	}

	return IRQ_HANDLED;
}


static int pxa27x_rtc_ioctl_counted(struct device *dev, unsigned int cmd,
			    unsigned long arg, int id)
{
	switch (cmd) {
	case RTC_PIE_OFF:
		spin_lock_irq(&pxa27x_rtc_lock);
		OIER &= ~OIER_E(pxa27x_pcinfo[id].num);
		spin_unlock_irq(&pxa27x_rtc_lock);
		return 0;

	case RTC_PIE_ON:
		spin_lock_irq(&pxa27x_rtc_lock);
		OIER |= OIER_E(pxa27x_pcinfo[id].num);
		OSMRx(pxa27x_pcinfo[id].num)
			= PXA27X_RTC_PERIODIC_FREQ/rtc_freq[id];
		last_int_time[id] = OSCRx(4);
		OSCRx(pxa27x_pcinfo[id].num) = 0;
		spin_unlock_irq(&pxa27x_rtc_lock);
		return 0;

	case RTC_IRQP_READ:
		return put_user(rtc_freq[id], (unsigned long *)arg);

	case RTC_IRQP_SET:
		if (arg < 1 || arg > PXA27X_RTC_PERIODIC_FREQ)
			return -EINVAL;
		rtc_freq[id] = arg;
		return 0;
	}

	return -ENOIOCTLCMD;
}

static int pxa27x_rtc_proc_counted(struct device *dev, struct seq_file *seq,
				   int id)
{
	seq_printf(seq, "periodic_IRQ\t: %s\n",
		   (OIER & OIER_E(pxa27x_pcinfo[id].num)) ? "yes" : "no");
	seq_printf(seq, "periodic_freq\t: %ld\n", rtc_freq[id]);

	return 0;
}

static int pxa27x_irq_set_freq_counted(struct device *dev, int freq, int id)
{
	unsigned long flags;

	spin_lock_irqsave(&pxa27x_rtc_lock, flags);
	rtc_freq[id] = freq;
	spin_unlock_irqrestore(&pxa27x_rtc_lock, flags);

	return 0;
}

/* How can this know which one it is refering to? */
static int pxa27x_irq_set_state_counted(struct device *dev, int enabled, int id)
{
	unsigned long flags;

	spin_lock_irqsave(&pxa27x_rtc_lock, flags);
	if (enabled) {
		/* Enable interrupt generation */
		OIER |= OIER_E(pxa27x_pcinfo[id].num);
		/* Set the match register */
		/* FIXME - should not be hard coded!*/
		OSMRx(pxa27x_pcinfo[id].num)
			= PXA27X_RTC_PERIODIC_FREQ/rtc_freq[id];
		/* Clear previous match */
		/* Possible corner case - counter loops and hits the match again
		 * before reset? - Very unlikely*/
		OSSR = OSSR_M(pxa27x_pcinfo[id].num);
		/* Start primary counter - used to detect missed interrupts */
		last_int_time[id] = OSCRx(4);

		/* Reset periodic counter */
		OSCRx(pxa27x_pcinfo[id].num) = 0;
		/* May have already missed the match? */
	} else
		OIER &= ~OIER_E(pxa27x_pcinfo[id].num);
/* what if an interrupt triggers in here?*/
	spin_unlock_irqrestore(&pxa27x_rtc_lock, flags);

return 0;
}

PXA27X_PERIODIC_RTC_FUNCS(0);
PXA27X_PERIODIC_RTC_FUNCS(1);
PXA27X_PERIODIC_RTC_FUNCS(2);
PXA27X_PERIODIC_RTC_FUNCS(3);
PXA27X_PERIODIC_RTC_FUNCS(4);
PXA27X_PERIODIC_RTC_FUNCS(5);
PXA27X_PERIODIC_RTC_FUNCS(6);

static int pxa27x_rtc_probe(struct platform_device *pdev)
{
	struct rtc_device **rtc;
	int ret;
	int i;

	rtc = kzalloc(sizeof(struct rtc_device *)*PXA27X_NUM_PERIODIC_RTCS,
		      GFP_KERNEL);
	for (i = 0; i < PXA27X_NUM_PERIODIC_RTCS; i++)
		last_int_time[i] = 0;

	device_init_wakeup(&pdev->dev, 1);

	/* Enable primary time counter - used to check for missed interrupts */
	OMCRx(4) = OMCR_C | OMCR_FREQ_32768HZ | OMCR_PERIODIC;
	OSCRx(4) = 0;
	for (i = 0; i < PXA27X_NUM_PERIODIC_RTCS; i++)
		OMCRx(pxa27x_pcinfo[i].num) = OMCR_C | OMCR_PERIODIC
			| OMCR_OSCR_RESET_ON_MATCH | OMCR_FREQ_32768HZ;

	rtc[0] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_0,
				     THIS_MODULE);
	if (IS_ERR(rtc[0])) {
		ret = PTR_ERR(rtc[0]);
		goto error_free_rtc;
	}

	rtc[1] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_1,
				     THIS_MODULE);
	if (IS_ERR(rtc[1])) {
		ret = PTR_ERR(rtc[1]);
		goto error_unreg_0;
	}
	rtc[2] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_2,
				     THIS_MODULE);
	if (IS_ERR(rtc[2])) {
		ret = PTR_ERR(rtc[2]);
		goto error_unreg_1;
	}
	rtc[3] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_3,
				     THIS_MODULE);
	if (IS_ERR(rtc[3])) {
		ret = PTR_ERR(rtc[3]);
		goto error_unreg_2;
	}
	rtc[4] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_4,
				     THIS_MODULE);
	if (IS_ERR(rtc[4])) {
		ret = PTR_ERR(rtc[4]);
		goto error_unreg_3;
	}
	rtc[5] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_5,
				     THIS_MODULE);
	if (IS_ERR(rtc[5])) {
		ret = PTR_ERR(rtc[5]);
		goto error_unreg_4;
	}
	rtc[6] = rtc_device_register(pdev->name,
				     &pdev->dev, &pxa27x_periodic_rtc_ops_6,
				     THIS_MODULE);
	if (IS_ERR(rtc[6])) {
		ret = PTR_ERR(rtc[6]);
		goto error_unreg_5;
	}

	platform_set_drvdata(pdev, rtc);

	/* Now configure the interrupt we are going to be using.*/
	ret = request_irq(IRQ_OST_4_11, pxa27x_periodic_interrupt, 0,
			  "px27x_periodic_timer", &pdev->dev);
	if (ret) {
		dev_err(&pdev->dev, "IRQ %d already in use.\n", IRQ_OST1);
		goto error_unreg_6;
	}

	return 0;
error_unreg_6:
	rtc_device_unregister(rtc[6]);
error_unreg_5:
	rtc_device_unregister(rtc[5]);
error_unreg_4:
	rtc_device_unregister(rtc[4]);
error_unreg_3:
	rtc_device_unregister(rtc[3]);
error_unreg_2:
	rtc_device_unregister(rtc[2]);
error_unreg_1:
	rtc_device_unregister(rtc[1]);
error_unreg_0:
	rtc_device_unregister(rtc[0]);
error_free_rtc:
	kfree(rtc);
	return ret;
}

static int pxa27x_rtc_remove(struct platform_device *pdev)
{
	struct rtc_device **rtc = platform_get_drvdata(pdev);
	int i;
	if (rtc)
		for (i = 0; i < PXA27X_NUM_PERIODIC_RTCS; i++)
			rtc_device_unregister(rtc[i]);

	free_irq(IRQ_OST_4_11, &pdev->dev);
	/* irq's all now being ignored.
	 * Time to stop everything we might have started */

	/* Clear RTC status register */
	RTSR = 0;

	/* Clear match for periodic counter */
	OSSR = OSSR_M(5) | OSSR_M(6) | OSSR_M(7) | OSSR_M(8)
		| OSSR_M(9) | OSSR_M(10) | OSSR_M(11);
	kfree(rtc);

	return 0;
}


static struct platform_driver pxa27x_rtc_driver = {
	.probe		= pxa27x_rtc_probe,
	.remove		= pxa27x_rtc_remove,
	.driver		= {
		.name		= "pxa27x-rtc",
	},
};

static int __init pxa27x_rtc_init(void)
{
	return platform_driver_register(&pxa27x_rtc_driver);
}

static void __exit pxa27x_rtc_exit(void)
{
	platform_driver_unregister(&pxa27x_rtc_driver);
}

module_init(pxa27x_rtc_init);
module_exit(pxa27x_rtc_exit);

MODULE_AUTHOR("Jonathan Cameron <jic23@cam.ac.uk>");
MODULE_DESCRIPTION("PXA27x Supplementary Realtime Clock Driver (RTC)");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:pxa27x-rtc");
