#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/fs.h>
#include <linux/sysfs.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/irqs.h>

#include "superbird2.h"

MODULE_AUTHOR("Brian Mayton");
MODULE_DESCRIPTION("Driver for the CSE466 SuperBird board");
MODULE_VERSION("2.0");
MODULE_LICENSE("GPL");

struct i2c_client *superbird_client = NULL;

struct superbird_data {
	struct i2c_client *client;
	dev_t chrdev;
};

static int superbird_probe(struct i2c_client *client, const struct i2c_device_id *id);
static int superbird_remove(struct i2c_client *client);
static int superbird_detect(struct i2c_client *client, int kind, struct i2c_board_info *info);
static uint8_t switches = 0;

static struct work_struct superbird_read_status_work;

static unsigned short normal_i2c[] = { SUPERBIRD_I2C_ADDR, I2C_CLIENT_END };
I2C_CLIENT_INSMOD_1(superbird);

static struct i2c_device_id superbird_id_table[] = {
	{ "superbird", 0 },
	{ }
};

MODULE_DEVICE_TABLE(i2c, superbird_id_table);

static struct i2c_driver superbird_driver = {
	.driver = {
		.name = "superbird",
		.owner = THIS_MODULE,
	},
	
	.id_table = superbird_id_table,
	.probe = superbird_probe,
	.remove = __devexit_p(superbird_remove),
	.class = I2C_CLASS_ALL,
	.address_data = &addr_data,
	.detect = superbird_detect,
};

int superbird_audio_pwr_en(int enable) {
    if(superbird_client != NULL) {
        superbird_write_reg(superbird_client, 0x21, enable);
		return 0;
   	} else {
        return -ENODEV;
	}
}

EXPORT_SYMBOL(superbird_audio_pwr_en);

static ssize_t superbird_audio_pwr_show(struct device *dev, struct device_attribute *attr, char *buf) {
	struct i2c_client *client;
	uint8_t pwr_status;
	client = to_i2c_client(dev);
	pwr_status = superbird_read_reg(client, 0x21);
	return sprintf(buf, "%d\n", pwr_status);
}

static ssize_t superbird_audio_pwr_store(struct device *dev, struct device_attribute *attr, const char *buf,
	size_t count) {
	struct i2c_client *client;
	int pwr_status, ret;
	client = to_i2c_client(dev);
	ret = sscanf(buf, "%d", &pwr_status);
	superbird_write_reg(client, 0x21, pwr_status);
	return count;
}

DEVICE_ATTR(audio_pwr, S_IRUGO | S_IWUSR , superbird_audio_pwr_show, superbird_audio_pwr_store);

static ssize_t superbird_led_color_show(struct device *dev, struct device_attribute *attr, char *buf) {
	struct i2c_client *client;
	int color;
	uint8_t i2cbuf[3];
	client = to_i2c_client(dev);
	i2cbuf[0] = 0x10;
	i2c_master_send(client, i2cbuf, 1);
	i2c_master_recv(client, i2cbuf, 3);
	color = ((int)i2cbuf[0] << 16) | ((int)i2cbuf[1] << 8) | ((int)i2cbuf[2]);
	return sprintf(buf, "%06x\n", color);
}

static ssize_t superbird_led_color_store(struct device *dev, struct device_attribute *attr, const char *buf,
	size_t count) {
	struct i2c_client *client;
	int color;
	uint8_t i2cbuf[4];
	client = to_i2c_client(dev);
	sscanf(buf, "%x", &color);
	i2cbuf[0] = 0x10;
	i2cbuf[1] = (color >> 16) & 0xFF;
	i2cbuf[2] = (color >> 8) & 0xFF;
	i2cbuf[3] = (color) & 0xFF;
	i2c_master_send(client, i2cbuf, 4);
	return count;
}

DEVICE_ATTR(led_color, S_IRUGO | S_IWUSR , superbird_led_color_show, superbird_led_color_store);

// For some reason, the interrupt-driven switch stuff works really well for a while, then
// just borks itself.  For now, we'll query the device every time the switches are read...
static ssize_t superbird_switches_show(struct device *dev, struct device_attribute *attr, char *buf) {
	struct i2c_client *client;
	int switches_local;
	client = to_i2c_client(dev);
	switches_local = superbird_read_reg(client, 0x30);
	return sprintf(buf, "%d\n", switches_local);
}

DEVICE_ATTR(switches, S_IRUGO, superbird_switches_show, NULL);

static ssize_t superbird_power_status_show(struct device *dev, struct device_attribute *attr, char *buf) {
	struct i2c_client *client;
	int powerstat;
	client = to_i2c_client(dev);
	powerstat = superbird_read_reg(client, 0x31);
	return sprintf(buf, "%d\n", powerstat);
}

DEVICE_ATTR(power_status, S_IRUGO, superbird_power_status_show, NULL);

static ssize_t superbird_voltage_selected_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct i2c_client *client;
    int voltage;
    client = to_i2c_client(dev);
    voltage = superbird_read_reg(client, 0x40);
    voltage = (voltage * 6000) >> 8;
    return sprintf(buf, "%d\n", voltage);
}

DEVICE_ATTR(voltage_selected, S_IRUGO, superbird_voltage_selected_show, NULL);

static ssize_t superbird_voltage_bat1_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct i2c_client *client;
    int voltage;
    client = to_i2c_client(dev);
    voltage = superbird_read_reg(client, 0x41);
    voltage = (voltage * 6000) >> 8;
    return sprintf(buf, "%d\n", voltage);
}

DEVICE_ATTR(voltage_battery1, S_IRUGO, superbird_voltage_bat1_show, NULL);

static ssize_t superbird_voltage_bat2_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct i2c_client *client;
    int voltage;
    client = to_i2c_client(dev);
    voltage = superbird_read_reg(client, 0x42);
    voltage = (voltage * 6000) >> 8;
    return sprintf(buf, "%d\n", voltage);
}

DEVICE_ATTR(voltage_battery2, S_IRUGO, superbird_voltage_bat2_show, NULL);

static int superbird_detect(struct i2c_client *client, int kind, struct i2c_board_info *info) {
	char buf[10];
	printk(KERN_INFO "Detect called for %d.\n", info->addr);
	buf[0] = 0x00;
	i2c_master_send(client, buf, 1);
	i2c_master_recv(client, buf, 6);
	buf[6] = '\0';
	if(strncmp("BIRD02", buf, 6)) return -ENODEV;
	printk(KERN_INFO "Detected SuperBird with signature: %s\n", buf);
	strlcpy(info->type, "superbird", I2C_NAME_SIZE);
	return 0;
}

static void superbird_read_status(struct work_struct *work) {
	if(superbird_client != NULL)
		switches = superbird_read_reg(superbird_client, 0x30);
	printk(KERN_DEBUG "SuperBird IRQ, switches = 0x%02x\n", switches);
}

static irqreturn_t superbird_irq(int irq, void *dev_id) {
	schedule_work(&superbird_read_status_work);
	GEDR0 = (1<<10);
	return IRQ_RETVAL(IRQ_HANDLED);
}

static int superbird_probe(struct i2c_client *client, const struct i2c_device_id *id) {
	struct superbird_data *data;
	int ret;
	
	printk(KERN_INFO "Probe called.\n");
	data = kzalloc(sizeof(struct superbird_data), GFP_KERNEL);
	if(data == NULL) return -ENODEV;
	i2c_set_clientdata(client, data);
	data->client = client;
	
	ret = alloc_chrdev_region(&data->chrdev, 0, 1, "superbird");
	if(ret < 0) goto exit_error;
	
	ret = device_create_file(&client->dev, &dev_attr_audio_pwr);
	if(ret < 0) goto exit_error;
	ret = device_create_file(&client->dev, &dev_attr_led_color);
	if(ret < 0) goto exit_error;
	ret = device_create_file(&client->dev, &dev_attr_switches);
	if(ret < 0) goto exit_error;
	ret = device_create_file(&client->dev, &dev_attr_power_status);
	if(ret < 0) goto exit_error;
    ret = device_create_file(&client->dev, &dev_attr_voltage_selected);
	if(ret < 0) goto exit_error;
	ret = device_create_file(&client->dev, &dev_attr_voltage_battery1);
	if(ret < 0) goto exit_error;
	ret = device_create_file(&client->dev, &dev_attr_voltage_battery2);
	if(ret < 0) goto exit_error;

    if(superbird_client == NULL) {
		superbird_client = client;
		/*
		ret = request_irq(IRQ_GPIO(10), superbird_irq, 0, "superbird", NULL);
		if (ret < 0) {
			printk(KERN_WARNING "superbird: Unable to register IRQ! (%d)\n", ret);
			goto exit_error;
		}

		local_irq_disable();
		GPDR0 &= ~(1 << 10);    // Set GPIO10 (AT_INT) to be an input
		GAFR0_L &= ~0x00300000; // Set GPIO10 as a GPIO pin
		GRER0 |= (1 << 10);    // Enable rising edge interrupt on GPIO10
		GFER0 &= ~(1 << 10);    // Diable falling edge interrupt on GPIO10 (just in case)
		ICMR |= 0x00000400;	// Enable GPIO interrupts
		GEDR0 = (1<<10);
		local_irq_enable();
		*/
	}
	return 0;
	
exit_error:
	kfree(data);
	return ret;
};

static int __devexit superbird_remove(struct i2c_client *client) {
	struct superbird_data *data;
	printk(KERN_INFO "Remove called.\n");
    if(superbird_client == client)
        superbird_client = NULL;
	data = i2c_get_clientdata(client);
	unregister_chrdev_region(data->chrdev, 1);
	kfree(data);
	return 0;
};

int superbird_read_reg(struct i2c_client *client, uint8_t reg) {
	uint8_t data;
	i2c_master_send(client, &reg, 1);
	i2c_master_recv(client, &data, 1);
	return data;
}

int superbird_write_reg(struct i2c_client *client, uint8_t reg, uint8_t data) {
	char buf[2];
	buf[0] = reg;
	buf[1] = data;
	return i2c_master_send(client, buf, 2);
}

static int __init superbird_init(void) {
	printk(KERN_INFO "Loading SuperBird module 2.\n");
	INIT_WORK(&superbird_read_status_work, superbird_read_status);
	return i2c_add_driver(&superbird_driver);
}

static void __exit superbird_exit(void) {
	printk(KERN_INFO "Unloading SuperBird module.\n");
	//local_irq_disable();
	//GRER0 &= ~(1 << 10);
	//local_irq_enable();
	//free_irq(IRQ_GPIO(10), NULL);
	i2c_del_driver(&superbird_driver);
}

module_init(superbird_init);
module_exit(superbird_exit);

