/* The industrial I/O core
 *
 * Copyright (c) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#ifndef _INDUSTRIAL_IO_H_
#define _INDUSTRIAL_IO_H_

#include <linux/device.h>
#include <linux/industrialio/sysfs.h>
#include <linux/industrialio/ring_generic.h>
#include <linux/industrialio/chrdev.h>

/* IIO TODO LIST */
/* Static device specific elements (conversion factors etc)
 * should be exported via sysfs
 *
 * Provide alternative to high resolution timers if not available.
 *
 */

/* Event interface flags */
#define IIO_BUSY_BIT_POS 1

struct iio_dev;

/* Requires high resolution timers */
static inline s64 iio_get_time_ns(void)
{
	struct timespec ts;
	ktime_get_ts(&ts);
	return timespec_to_ns(&ts);
}

/**
 * struct iio_event_handler_list - element in list handlers for events
 *				(interrupt handlers for device events)
 * @list:		list header
 * @handler:		event handler function - called on event if this
 *			event_handler is enabled.
 * @refcount:		as the handler may be shared between multiple device
 *			side events, reference counting ensures clean removal
 *
 * Each device has one list of these per interrupt line
 **/
struct iio_event_handler_list {
	struct list_head list;
	int (*handler)(struct iio_dev *dev_io, int index, s64 timestamp,
		       int no_test);
	/* This element may be shared */
	int refcount;
};

/**
 * iio_add_event_to_list() - Wraps adding to lists and does reference
 * counting to allow shared handlers.
 *
 * @list:	the list associated with the event handler being used.
 * @el:		the list element of the event to be handled.
 */
int iio_add_event_to_list(struct iio_event_handler_list *list,
			  struct iio_event_handler_list *el);

int iio_remove_event_from_list(struct iio_event_handler_list *el);


/* Device operating modes */
#define INDIO_DIRECT_MODE		0x01
#define INDIO_RING_TRIGGERED		0x02
#define INDIO_RING_HARDWARE_BUFFER	0x08

#define INDIO_ALL_RING_MODES (INDIO_RING_TRIGGERED | INDIO_RING_HARDWARE_BUFFER)

/* Vast majority of this is set by the industrialio subsystem on a
 * call to iio_device_register. */

/**
 * struct iio_dev - industrial I/O device
 * @id:			[INTERN] used to identify device internally
 * @dev_data:		[DRIVER] device specific data
 * @modes:		[DRIVER] operating modes supported by device
 * @currentmode:	[DRIVER] current operating mode
 * @sysfs_dev:		[INTERN] device created by sysfs needed for management
 * @dev:		[DRIVER] the actual device structure
 * @attrs:		[DRIVER] device attributes
 * @driver_module:	[DRIVER] module structure used to ensure correct
 *			ownership of chrdevs etc
 * @scan_el_attrs:	[DRIVER] control of scan elements if that scan mode
 *			control method is used
 * @num_interrupt_lines:[DRIVER] number of physical interrupt lines from device
 * @interrupts:		[INTERN] interrupt line specific event lists etc
 * @event_attrs:	[DRIVER] event control attributes
 * @event_conf_attrs:	[DRIVER] event configuration attributes
 * @event_interfaces:	[INTERN] event chrdevs associated with interrupt lines
 * @ring_bytes_per_datum:[DRIVER] number of bytes per datum in ring buffer
 * @ring_length:	[DRIVER] number of datums in ring buffer
 * @ring_attrs:		[INTERN] ring related attributes
 * @ring_prenable:	[DRIVER] function to run prior to marking ring enabled
 * @ring_postenable:	[DRIVER] function to run after marking ring enabled
 * @ring_predisable:	[DRIVER] function to run prior to marking ring disabled
 * @ring_postdisable:	[DRIVER] function to run after marking ring disabled
 * @mlock:		[INTERN] lock used to prevent simultaneous device state
 *			changes
 * @scan_count:	        [INTERN] the number of elements in the current scan mode
 * @scan_mask:		[INTERN] bitmask used in masking scan mode elements
 * @trig:		[INTERN] current device trigger (ring buffer modes)
 * @pollfunc:		[DRIVER] function run on trigger being recieved
 **/

struct iio_dev {
	int				id;
	void				*dev_data;
	int				modes;
	int				currentmode;
	struct device			*sysfs_dev;
	struct device			*dev;
	const struct attribute_group	*attrs;
	struct module			*driver_module;

	struct attribute_group		*scan_el_attrs;
	int				num_interrupt_lines;
	struct iio_interrupt		**interrupts;
	struct attribute_group		*event_attrs;
	struct attribute_group		*event_conf_attrs;

	struct iio_event_interface	*event_interfaces;
	struct iio_ring_access_funcs	ring_access;

	void				*ring;
	struct attribute_group		*ring_attrs_group;
	struct iio_ring_attr		*ring_attrs;

	int				(*ring_preenable)(struct iio_dev *);
	int				(*ring_postenable)(struct iio_dev *);
	int				(*ring_predisable)(struct iio_dev *);
	int				(*ring_postdisable)(struct iio_dev *);

	struct mutex			mlock;

	int				scan_count;
	u16				scan_mask;
	//Only one trigger used for now per device?
	struct iio_trigger		*trig;
	struct iio_poll_func		*pollfunc;
};

/**
 * iio_device_register() - register a device with the IIO subsystem
 * @dev_info:		Device structure filled by the device driver
 **/
int iio_device_register(struct iio_dev *dev_info);

/**
 * iio_device_unregister() - unregister a device from the IIO subsystem
 * @dev_info:		Device structure representing the device.
 **/
void iio_device_unregister(struct iio_dev *dev_info);

/**
 * struct iio_interrupt - wrapper used to allow easy handling of multiple
 *			physical interrupt lines
 * @dev_info:		the iio device for which the is an interrupt line
 * @linue_number:	associated line number
 * @irq:		associate interrupt number
 * @ev_list:		event handler list for associated events
 **/
struct iio_interrupt {
	struct iio_dev			*dev_info;
	int				line_number;
	int				irq;
	struct iio_event_handler_list	ev_list;
};

/**
 * iio_register_interrupt_line() - Tell IIO about interrupt lines
 *
 * @irq:		Typically provided via platform data.
 * @line_number:	Which interrupt line of the device is this?
 * @type:		Interrupt type (e.g. edge triggered etc)
 * @name:		Identifying name.
 **/
int iio_register_interrupt_line(unsigned int			irq,
				struct iio_dev	*dev_info,
				int				line_number,
				unsigned long			type,
				const char			*name);

void iio_unregister_interrupt_line(struct iio_dev *dev_info,
				   int line_number);



/**
 * iio_push_event() - try to add event to the list for userspace reading
 * @dev_info:		IIO device structure
 * @ev_line:		Which event line (hardware interrupt)
 * @ev_code:		What event
 * @timestamp:		When the event occured
 **/
int iio_push_event(struct iio_dev *dev_info,
		  int ev_line,
		  int ev_code,
		  s64 timestamp);

/**
 * struct iio_work_cont - convenient container for case where different
 *			functions are called depending on whether this event
 *			is the only one that may have generated the interrupt.
 * @ws:			[DEVICE]work_struct when not only possible event
 * @ws_nocheck:		[DEVICE]work_struct when only possible event
 * @address:		[DEVICE]associated register address
 * @mask:		[DEVICE]associated mask for identifying event source
 * @st:			[DEVICE]device specific state information
 **/
struct iio_work_cont {
	struct work_struct ws;
	struct work_struct ws_nocheck;
	int address;
	int mask;
	void *st;
};

#define to_iio_work_cont_check(_ws)			\
	container_of(_ws, struct iio_work_cont, ws)

#define to_iio_work_cont_no_check(_ws)				\
	container_of(_ws, struct iio_work_cont, ws_nocheck)

static inline void
iio_init_work_cont(struct iio_work_cont *cont,
		   void (*_checkfunc)(struct work_struct *),
		   void (*_nocheckfunc)(struct work_struct *),
		   int _add, int _mask, void* _st)
{
	INIT_WORK(&(cont)->ws, _checkfunc);
	INIT_WORK(&(cont)->ws_nocheck, _nocheckfunc);
	cont->address = _add;
	cont->mask = _mask;
	cont->st = _st;
}
/**
 * __iio_push_event() - tries to add an event to the list associated with a chrdev
 *
 * @ev_int:	the event interface associated to which we are pushing the event.
 **/
int __iio_push_event(struct iio_event_interface *ev_int,
		    int ev_code,
		    s64 timestamp,
		    struct iio_shared_ev_pointer*
		    shared_pointer_p);
void __iio_change_event(struct iio_detected_event_list *ev,
			int ev_code,
			s64 timestamp);
/**
 * iio_setup_ev_int() - Configure an event interface (chrdev)
 *
 * @ev_int:		Interface we are configuring
 * @owner:		Module that is responsible for registering this ev_int
 * @dev		Device whose ev_int this is
 **/
int iio_setup_ev_int(struct iio_event_interface *ev_int,
		     const char *name,
		     struct module *owner,
		     struct device *dev);

void iio_free_ev_int(struct iio_event_interface *ev_int,  struct device *dev);

/**
 * iio_allocate_chrdev() - Allocate a chrdev
 *
 * @hander:		Struct that contains relevant file handling for chrdev
 **/
int iio_allocate_chrdev(struct iio_handler *handler);
void iio_deallocate_chrdev(struct iio_handler *handler);

ssize_t iio_show_attr_minor(struct device *dev,
			    struct device_attribute *attr,
			    char *buf);

/* Used to distinguish between bipolar and unipolar scan elemenents */
#define IIO_SIGNED(a) -(a)
#define IIO_UNSIGNED(a) (a)
struct iio_dev;
/**
 * struct iio_scan_el - an individual element of a scan
 *
 * @dev_attr:		control attribute (if directly controllable)
 * @mask:		bitmask corresonding to this scan element
 * @bit_count:		number of bits in scan element
 * @label:		name of scan element
 **/
struct iio_scan_el {
	struct device_attribute dev_attr;
	unsigned int mask;
	int bit_count;
	/* typically the address of the register when that makes sense) */
	unsigned int label;
	int (* set_state)(struct iio_scan_el *scanel,
			  struct iio_dev *dev_info,
			  bool state);
};

#define to_iio_scan_el(_dev_attr)				\
	container_of(_dev_attr, struct iio_scan_el, dev_attr);

/**
 * iio_scan_el_store() - sysfs scan element selection interface.
 *
 * A generic function used to enable various scan elements.  In some
 * devices explicit read commands for each channel mean this is merely
 * a software switch.  In others this must actively disable the channel.
 * Complexities occur when this interacts with data ready type triggers
 * which may not reset unless every channel that is enabled is explicitly
 * read.
 **/
ssize_t iio_scan_el_store(struct device *dev, struct device_attribute *attr,
			  const char *buf, size_t len);
/**
 * iio_scal_el_show() -	sysfs interface to query whether a scan element is
 *			is enabled or not.
 **/
ssize_t iio_scan_el_show(struct device *dev, struct device_attribute *attr,
			 char *buf);
/**
 * IIO_SCAN_EL: - declare and initialize a scan element without control func
 * @_name:	identifying name. Resulting struct is iio_scan_el_##_name,
 *		sysfs element, scan_en_##_name.
 * @_number:	unique id number for the scan element.
 * @_bits:	number of bits in the scan element result (used in mixed bit
 *		length devices).
 * @_label:	indentification variable used by drivers.  Often a reg address.
 **/
#define IIO_SCAN_EL(_name, _number, _bits, _label)			\
	struct iio_scan_el iio_scan_el_##_name = {			\
		.dev_attr =	__ATTR(scan_en_##_name,			\
				       S_IRUGO | S_IWUSR,		\
				       iio_scan_el_show,		\
				       iio_scan_el_store),		\
		.mask = (1 << _number),					\
		.bit_count = _bits,					\
		.label = _label,					\
	}
/**
 * IIO_SCAN_EL_C: - declare and initialize a scan element with a control func
 *
 * @_controlfunc:	function used to notify hardware of whether state changes
 **/
#define IIO_SCAN_EL_C(_name, _number, _bits, _label, _controlfunc )	\
	struct iio_scan_el iio_scan_el_##_name = {			\
		.dev_attr =	__ATTR(xscan_en_##_name,		\
				       S_IRUGO | S_IWUSR,		\
				       iio_scan_el_show,		\
				       iio_scan_el_store),		\
		.mask = (1 << _number),					\
		.bit_count = _bits,					\
		.label = _label,					\
		.set_state = _controlfunc,				\
	}


/* Ring buffer related */
#ifdef CONFIG_IIO_RING_BUFFER

static inline bool iio_ring_enabled(struct iio_dev *dev_info)
{
	return dev_info->currentmode
		& (INDIO_RING_TRIGGERED
		   | INDIO_RING_HARDWARE_BUFFER);
};
int iio_device_register_ring(struct iio_dev *dev_info, int id);
void iio_device_unregister_ring(struct iio_dev *dev_info);

#else

static inline bool iio_ring_enabled(struct iio_dev *dev_info)
{
	return false;
};
static inline int iio_device_register_ring(struct iio_dev *devinfo, int id)
{
	return 0;
};
static inline void iio_device_unregister_ring(struct iio_dev *dev_info) {};


#endif /* CONFIG_IIO_RING_BUFFER */
static inline bool single_list(struct list_head *list)
{
	return list->next->next == list;
};
#endif /* _INDUSTRIAL_IO_H_ */
