/* The industrial I/O core - generic ring buffer interfaces.
 *
 * Copyright (c) 2008 Jonathan Cameron
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#ifndef _IIO_RING_GENERIC_H_
#define _IIO_RING_GENERIC_H_

#include <linux/industrialio/chrdev.h>

struct iio_ring_buffer;

#ifdef CONFIG_IIO_RING_BUFFER

/**
 * iio_push_ring_event() - ring buffer specific push to event chrdev
 * @ring_buf:		ring buffer that is the event source
 * @event_code:		event indentification code
 * @timestamp:		time of event
 **/
int iio_push_ring_event(struct iio_ring_buffer *ring_buf,
			int event_code,
			s64 timestamp);
/**
 * iio_push_or_escallate_ring_event() -	escallate or add as appropriate
 *
 * Typical usecase is to escallate a 50% ring full to 75% full if noone has yet
 * read the first event. Clearly the 50% full is no longer of interest in
 * typical use case.
 **/
int iio_push_or_escallate_ring_event(struct iio_ring_buffer *ring_buf,
				     int event_code,
				     s64 timestamp);

/**
 * struct iio_ring_access_funcs - access functions for ring buffers.
 * @create:		perform allocation
 * @init:		get ring buffer ready for use
 * @_exit:		reverse steps in init
 * @_free:		deallocate ring buffer
 * @mark_in_use:	reference counting, typically to prevent module removal
 * @unmark_in_use:	reduce reference count when no longer using ring buffer
 * @store_to:		actually store stuff to the ring buffer
 * @read_last:		get the last element stored
 * @rip_lots:		try to get a specified number of elements (must exist)
 * @mark_param_change:	notify ring that some relevant parameter has changed
 *			Often this means the underlying storage may need to change
 * @request_update:	if a parameter change has been marked, update underlying
 *			storage.
 * @get_bpd:		get current bytes per datum
 * @set_bpd:		set number of bytes per datum
 * @get_length:		get number of datums in ring
 * @set_length:		set number of datums in ring
 * @is_enabled:		query if ring is currently being used
 * @enable:		enable the ring
 *
 * The purpose of this structure is to make the ring buffer element
 * modular as event for a given driver, different usecases may require
 * different ring designs (space efficiency vs speed for example.
 *
 * It is worth noting that a given ring implementation may only support a small
 * proportion of these functions.  The core code 'should' cope fine with any of
 * them not existing.
 **/
struct iio_ring_access_funcs {
	int (*_create)(void **ring);
	int (*_init)(void *ring, void *indio_dev);
	void (*_exit)(void *ring);
	void (*_free)(void *ring);

	void (*mark_in_use)(void *ring);
	void (*unmark_in_use)(void *ring);

	int (*store_to)(void *ring, u8 *data, s64 timestamp);
	int (*read_last)(void *ring, u8 *data);
	int (*rip_lots)(void *ring, size_t count, u8 **data, int *dead_offset);

	int (*mark_param_change)(void *ring);
	int (*request_update)(void *ring);

	int (*get_bpd)(void *ring);
	int (*set_bpd)(void *ring, int bpd);
	int (*get_length)(void *ring);
	int (*set_length)(void *ring, int length);

	int (*is_enabled)(void *ring);
	int (*enable)(void *ring);
};

/**
 * struct iio_ring_buffer - general ring buffer structure
 * @length:		[DEVICE]number of datums in ring
 * @bpd:		[DEVICE]size of individual datum including timestamp
 * @loopcount:		[INTERN]number of times the ring has looped
 * @access_minor_name:	[INTERN]store of name of the access chrdev minor number
 *			sysfs attribute
 * @access_handler:	[INTERN]chrdev access handling
 * @event_minor_name:	[INTERN]store of name of the event chrdev minor number
 *			sysfs attribute
 * @ev_int:		[INTERN]chrdev interface for the event chrdev
 * @shared_ev_pointer:	[INTERN]the shared event pointer to allow escalation of
 *			events
 * @ring_access:	[DRIVER]ring access functions associated with the
 *			implementation.
 */
struct iio_ring_buffer {
	int				length;
	int				bpd;
	int				loopcount;
	char				*access_minor_name;
	struct iio_chrdev_minor_attr	access_minor_attr;
	struct iio_handler		access_handler;
	char				*event_minor_name;
	struct iio_event_interface	ev_int;
	struct iio_shared_ev_pointer	shared_ev_pointer;
	struct iio_ring_access_funcs    *ring_access;
};

/**
 * __iio_init_ring_buffer() - initialize common elements of ring buffers.
 **/
static inline void __iio_init_ring_buffer(struct iio_ring_buffer *ring,
				 int bytes_per_datum, int length)
{
	ring->bpd = bytes_per_datum;
	ring->length = length;
	ring->loopcount = 0;
	ring->shared_ev_pointer.ev_p = 0;
	ring->shared_ev_pointer.lock =
		__SPIN_LOCK_UNLOCKED(ring->shared_ev_pointer->loc);
}

#else

struct iio_ring_buffer {};
struct iio_ring_access_funcs {};
#endif /* CONFIG_IIO_RING_BUFFER */

#endif /* _IIO_RING_GENERIC_H_ */
