/*
 * ak4642.c  --  AK4642 ALSA SoC Audio driver
 *
 * Copyright 2008 University of Washington CSE
 *
 * Author: Brian Mayton <bmayton@cs.washington.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */


 #include <linux/module.h>
 #include <linux/moduleparam.h>
 #include <linux/version.h>
 #include <linux/kernel.h>
 #include <linux/init.h>
 #include <linux/delay.h>
 #include <linux/pm.h>
 #include <linux/i2c.h>
 #include <linux/platform_device.h>
 #include <sound/core.h>
 #include <sound/pcm.h>
 #include <sound/pcm_params.h>
 #include <sound/soc.h>
 #include <sound/soc-dapm.h>
 #include <sound/initval.h>
 #include <sound/tlv.h>
 #include <asm/div64.h>

 #include "ak4642.h"

 #define AUDIO_NAME "AK4642"
 #define AK4642_VERSION "0.1"

 /*
  * Debug
  */

 #define AK4642_DEBUG 1

 #ifdef AK4642_DEBUG
 #define dbg(format, arg...) \
 	printk(KERN_DEBUG AUDIO_NAME ": " format "\n" , ## arg)
 #else
 #define dbg(format, arg...) do {} while (0)
 #endif
 #define err(format, arg...) \
 	printk(KERN_ERR AUDIO_NAME ": " format "\n" , ## arg)
 #define info(format, arg...) \
 	printk(KERN_INFO AUDIO_NAME ": " format "\n" , ## arg)
 #define warn(format, arg...) \
 	printk(KERN_WARNING AUDIO_NAME ": " format "\n" , ## arg)

struct snd_soc_codec_device soc_codec_dev_ak4642;

/*
 * ak4642 register cache
 */
static const u16 ak4642_reg[AK4642_CACHEREGNUM] = {
    0x0000, 0x0000, 0x0001, 0x0000,
    0x0002, 0x0000, 0x0000, 0x0000,
    0x00E1, 0x00E1, 0x0018, 0x0000,
    0x00E1, 0x0018, 0x0011, 0x0008,
    0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000,
};

/*
 * read ak4642 register cache
 */
static inline unsigned int ak4642_read_reg_cache(struct snd_soc_codec *codec,
	unsigned int reg)
{
	u16 *cache = codec->reg_cache;
	if (reg >= AK4642_CACHEREGNUM)
		return -1;
	return cache[reg];
}

/*
 * write ak4642 register cache
 */
static inline void ak4642_write_reg_cache(struct snd_soc_codec *codec,
	u16 reg, unsigned int value)
{
	u16 *cache = codec->reg_cache;
	if (reg >= AK4642_CACHEREGNUM)
		return;
	cache[reg] = value;
}

/*
 * write to the ak4642 register space
 */
static int ak4642_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int value)
{
	u8 data[2];

	data[0] = reg & 0xFF;
	data[1] = value & 0xFF;

	ak4642_write_reg_cache (codec, reg, value);
	if (codec->hw_write(codec->control_data, data, 2) == 2)
		return 0;
	else
		return -EIO;
}

static void ak4642_reg_dump(struct snd_soc_codec *codec) {
	uint8_t data[17];
	int i;
	data[0] = 0x00;
	codec->hw_write(codec->control_data, data, 1);
	codec->hw_read(codec->control_data, data, 17);
	printk(KERN_DEBUG "regs: ");
	for(i=0; i<17; i++) {
		printk("%02x ", data[i]);
	}
	printk("\n");
}

#define AK4642_DAIFMT \
	(SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_LEFT_J | SND_SOC_DAIFMT_CBS_CFS | \
	SND_SOC_DAIFMT_NB_NF)
	
#define AK4642_FMTS ( SNDRV_PCM_FMTBIT_S16_LE )

#define AK4642_DIR \
	(SND_SOC_DAIDIR_PLAYBACK | SND_SOC_DAIDIR_CAPTURE)

#define AK4642_RATES \
	(SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 | SNDRV_PCM_RATE_16000 | \
	SNDRV_PCM_RATE_22050 | SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 | \
	SNDRV_PCM_RATE_48000)

static int ak4642_pcm_prepare(struct snd_pcm_substream *substream) {
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;
	int bfs;

	//ak4642_write(codec, AK4642_MODE1, mode);
	//ak4642_write(codec, AK4642_MODE2, mode2);
	
	printk(KERN_DEBUG "prepare\n");
	ak4642_reg_dump(codec);

	return 0;
}

static int ak4642_set_dai_fmt(struct snd_soc_dai *codec_dai,
		unsigned int fmt)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	u16 mode1;
    u16 pm2;
    
    mode1 = ak4642_read_reg_cache(codec, AK4642_MODE1) & ~(0x03);
    pm2 = ak4642_read_reg_cache(codec, AK4642_PM2) & ~(1<<3);

	/* set master/slave audio interface */

	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		pm2 |= (1<<3);
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		break;
	default:
		return -EINVAL;
	}

	/* interface format */
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		mode1 |= 0x0003;
		break;
	case SND_SOC_DAIFMT_RIGHT_J:
    	mode1 |= 0x0002;
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		mode1 |= 0x0001;
		break;
	default:
		return -EINVAL;
	}

	ak4642_write(codec, AK4642_MODE1, mode1);
	ak4642_write(codec, AK4642_PM2, pm2);
	printk(KERN_DEBUG "set_dai_fmt\n");
	ak4642_reg_dump(codec);
	return 0;
}

struct snd_soc_dai ak4642_dai = {
	.name = "AK4642",
	/* playback capabilities */
	.playback = {
		.stream_name = "Playback",
		.channels_min = 1,
		.channels_max = 2,
		.rates = AK4642_RATES,
		.formats = AK4642_FMTS,},
	/* capture capabilities */
	.capture = {
		.stream_name = "Capture",
		.channels_min = 1,
		.channels_max = 2,
		.rates = AK4642_RATES,
		.formats = AK4642_FMTS,},
	/* pcm operations - see section 4 below */
	.ops = {
		.prepare = ak4642_pcm_prepare,
	},
	.dai_ops = {
	    .set_fmt = ak4642_set_dai_fmt,
	},
};
EXPORT_SYMBOL_GPL(ak4642_dai);

static int ak4642_init(struct snd_soc_device *socdev)
{
	struct snd_soc_codec *codec = socdev->codec;
	int ret = 0;

	codec->name = "AK4642";
	codec->owner = THIS_MODULE;
	codec->read = ak4642_read_reg_cache;
	codec->write = ak4642_write;
	codec->dai = &ak4642_dai;
	codec->num_dai = 1;
	codec->reg_cache_size = ARRAY_SIZE(ak4642_reg);
	codec->reg_cache =
			kzalloc(sizeof(u16) * ARRAY_SIZE(ak4642_reg), GFP_KERNEL);
	if (codec->reg_cache == NULL)
		return -ENOMEM;
	memcpy(codec->reg_cache, ak4642_reg,
		sizeof(u16) * ARRAY_SIZE(ak4642_reg));
	codec->reg_cache_size = sizeof(u16) * ARRAY_SIZE(ak4642_reg);

	/* register pcms */
	ret = snd_soc_new_pcms(socdev, SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1);
	if (ret < 0) {
		kfree(codec->reg_cache);
		return ret;
	}

	//ak4642_add_controls(codec);
	//ak4642_add_widgets(codec);
	ret = snd_soc_register_card(socdev);
	if (ret < 0) {
		snd_soc_free_pcms(socdev);
		snd_soc_dapm_free(socdev);
	}

	return ret;
}

static struct snd_soc_device *ak4642_socdev;

#if defined (CONFIG_I2C) || defined (CONFIG_I2C_MODULE)

#define I2C_DRIVERID_AK4642 0xF00F /* liam -  need a proper id */

static unsigned short normal_i2c[] = { 0, I2C_CLIENT_END };

/* Magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static int ak4642_i2c_attach(struct i2c_adapter *adap);
static int ak4642_i2c_detach(struct i2c_client *client);

static struct i2c_driver ak4642_i2c_driver = {
	.driver = {
		.name = "AK4642 I2C Codec",
		.owner = THIS_MODULE,
	},
	.id =             I2C_DRIVERID_AK4642,
	.attach_adapter = ak4642_i2c_attach,
	.detach_client =  ak4642_i2c_detach,
	.command =        NULL,
};

static struct i2c_client client_template = {
	.name =   "AK4642",
	.driver = &ak4642_i2c_driver,
};

/* If the i2c layer weren't so broken, we could pass this kind of data
   around */
static int ak4642_codec_probe(struct i2c_adapter *adap, int addr, int kind)
{
	struct snd_soc_device *socdev = ak4642_socdev;
	struct ak4642_setup_data *setup = socdev->codec_data;
	struct snd_soc_codec *codec = socdev->codec;
	struct i2c_client *i2c;
	int ret;

	if (addr != setup->i2c_address)
		return -ENODEV;

	client_template.adapter = adap;
	client_template.addr = addr;

	i2c = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (i2c == NULL){
		kfree(codec);
		return -ENOMEM;
	}
	memcpy(i2c, &client_template, sizeof(struct i2c_client));
	i2c_set_clientdata(i2c, codec);
	codec->control_data = i2c;

	ret = i2c_attach_client(i2c);
	if (ret < 0) {
		printk(KERN_ERR "failed to attach codec at addr %x\n", addr);
		goto err;
	}

	ret = ak4642_init(socdev);
	if (ret < 0) {
		printk(KERN_ERR "failed to initialise AK4642\n");
		goto err;
	}
	
	codec->hw_write(codec->control_data, "\x00\x74\x00\xa1\x00\x03\x00\x28\x00\xe1\x91\x18\x00\x91\x18\x11\x08", 17);
	//ak4642_write(codec, AK4642_PM1, (1<<PMDAC) | (1<<PMSPK) | (1<<PMVCM));
    //ak4642_write(codec, AK4642_SIG1, (1<<DACS) | (1<<MGAIN0) | (1<<SPPSN));
	printk(KERN_DEBUG "codec probe\n");
	ak4642_reg_dump(codec);
	
	return ret;

err:
	kfree(codec);
	kfree(i2c);
	return ret;
}

static int ak4642_i2c_detach(struct i2c_client *client)
{
	struct snd_soc_codec* codec = i2c_get_clientdata(client);
	i2c_detach_client(client);
	kfree(codec->reg_cache);
	kfree(client);

	return 0;
}

static int ak4642_i2c_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, ak4642_codec_probe);
}

#endif

static int ak4642_probe(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct ak4642_setup_data *setup;
	struct snd_soc_codec* codec;
	int ret = 0;

	printk(KERN_INFO "AK4642 Audio Codec %s", AK4642_VERSION);

	setup = socdev->codec_data;
	codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL);
	if (codec == NULL)
		return -ENOMEM;

	socdev->codec = codec;
	mutex_init(&codec->mutex);
	INIT_LIST_HEAD(&codec->dapm_widgets);
	INIT_LIST_HEAD(&codec->dapm_paths);

	ak4642_socdev = socdev;
#if defined (CONFIG_I2C) || defined (CONFIG_I2C_MODULE)
	if (setup->i2c_address) {
		normal_i2c[0] = setup->i2c_address;
		codec->hw_write = (hw_write_t)i2c_master_send;
		codec->hw_read = (hw_read_t)i2c_master_recv;
		ret = i2c_add_driver(&ak4642_i2c_driver);
		if (ret != 0)
			printk(KERN_ERR "can't add i2c driver");
	}
#else
	/* Add other interfaces here */
#endif
	return ret;
}

struct snd_soc_codec_device soc_codec_dev_ak4642 = {
	.probe = 	ak4642_probe,
	//.remove = 	ak4642_remove,
	//.suspend = 	ak4642_suspend,
	//.resume =	ak4642_resume,
};

EXPORT_SYMBOL_GPL(soc_codec_dev_ak4642);


MODULE_DESCRIPTION("ASoC AK4642 driver");
MODULE_AUTHOR("Brian Mayton");
MODULE_LICENSE("GPL");
